//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package eg;

import la.la.Type;
import la.la.Value;
import static la.la.Value.*;
import la.maths.Vector;  // A data-set is a Vector of Values
import la.maths.Vector.* ;
import static la.maths.Vector.*;
import mml.*;            // For Model, UPModel, etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison
 *
 * Reference: 
 *   Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.
 *
 * This file is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL) as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this file. If not, see <http://www.gnu.org/licenses/>.
 */
// ----------------------------------------------------------------------------

/** Ducks is a <I>toy</I> application program to illustrate the
   "{@link NaiveBayes (Naive-)&nbsp;Bayes}"
    {@link UPFunctionModel function-model} in
    [<A HREF="https://doi.org/10.1007/978-3-319-76433-7" TARGET="BOOKA">the&nbsp;book</A>].
    Also see the
    [<A HREF="https://www.cantab.net/users/mmlist/MML/A/" TARGET="mmlist">home&nbsp;page</A>].   */
public class Ducks
 {
   /** x to 4 sig-figs. */
   private static String x2S(double x) { return la.util.Util.x2String(x, 4); }

   /** 't' and 'f', short for {@link Value#ttrue} and {@link Value#ttrue}. */
   public static Value.Bool t = la.la.Value.ttrue, f = la.la.Value.ffalse;

   /** Species = coot | duck | swan. */
   public static Type.Enum Species =
      new Type.Enum( "Species", new String []{"coot", "duck", "swan"} );

   /** coot, duck, swan : {@link #Species}.  */
   public static Value coot = Species.string2value("coot"),
                       duck = Species.string2value("duck"),
                       swan = Species.string2value("swan");

   /** Also see {@link #waddles waddles}, {@link #quacks quacks}
      and {@link #weights weights}. */
   public static Vector
      birds   = values( new Value[]  { coot, coot, duck, duck, swan, swan} );

   /** Also see {@link #birds birds}, {@link #quacks quacks}
       and {@link #weights weights}. */
   public static Vector
      waddles = values( new Value[]  { t,    f,    t,    f,    f,    t   } );

   /** Also see {@link #birds birds}, {@link #waddles waddles}
       and {@link #weights weights}. */
   public static Vector
      quacks  = values( new Value[]  { t,    f,    t,    f,    t,    f   } );

   /** 'weights' are frequencies of the observations; also see
       {@link #birds birds}, {@link #waddles waddles} and
       {@link #quacks quacks}. */
   public static double []
      weights =         new double[] { 5,   45,   27,    3,    2,   18   };

   /** {@link #waddles waddles}.zip({@link #quacks quacks}).
       The properties of observed {@link #birds}.  */
   public static Vector attributes = waddles.zip(quacks);

   /** A toy data-set from which to estimate a function-model.
       ds=({@link #attributes}.zip({@link #birds})).weight({@link #weights})
       :&nbsp;&lang;&lang;waddles?,quacks?&rang;,bird&rang;<SUP><B>*</B></SUP>,
       note <I>weighted</I> by weights.  */
   public static Vector ds = (attributes.zip(birds)).weight(weights);
   // -------------------------------------------------------------------------

   /** UnParameterised Model over Bool={{@link #f},{@link #t}}. */
   public static MultiState boolUPM = new MultiState(pair(f, t));  // 2-state

   /** UnParameterised Model over {@link #Species}={coot,duck,swan}. */
   public static MultiState birdUPM = new MultiState(pair(coot, swan)); // 3

   /** UnParameterised <I>Independent</I> Model over Bool&times;Bool.
       It is the use of this Independent in {@link #walk_N_talk} and
       thence in {@link #bird_N_walk_N_talk}
       that makes {@link #walk_N_talk_2_bird} "Naive".  */
   public static Independent walk_N_talk =
      new Independent(pair(boolUPM, boolUPM));

   /** UnParameterised function-model of Species&rarr;Bool&times;Bool.
       It is new {@link CPT CPT}(triple(coot,swan,{@link #walk_N_talk})). */
   public static UPFunctionModel bird_2_walk_N_talk =
      new CPT(triple(coot, swan, walk_N_talk));  // Species->Bool^2

   /** UnParameterised Model of Species&times;(Bool&times;Bool).
       That is of {@link #Species} and the attributes
       "walks" and "talks" (like a duck).
       It is made from {@link #birdUPM} and {@link #bird_2_walk_N_talk}.
       Note that the attribute-<I>pair</I> is Dependent on Species but
       each attribute is <I>Independent</I> of the other.  */
   public static Dependent bird_N_walk_N_talk =
      new Dependent(pair(birdUPM, bird_2_walk_N_talk));

   /** new {@link NaiveBayes}({@link #bird_N_walk_N_talk}).
       That is, an UnParameterised "NaiveBayes" function-model of
       (Bool&times;Bool)&rarr;{@link #Species}. The story is that
       walk_N_talk_2_bird is made from {@link #bird_N_walk_N_talk}
       which is made from {@link #birdUPM} and {@link #bird_2_walk_N_talk}.
       The latter is made from {@link #walk_N_talk} and walking and
       talking (like a duck) are modelled as <I>Independent</I>
       but nothing actually requires this &mdash; apart from naivety.  */
   public static NaiveBayes walk_N_talk_2_bird =
      new NaiveBayes( bird_N_walk_N_talk );

   /** The particular Naive Bayes function-model used as
       an example in the book.  */
   public static NaiveBayes.M book_eg = (NaiveBayes.M)               // book_eg
      walk_N_talk_2_bird
          .apply( pair( doubles(new double [] {0.5, 0.3, 0.2}),
                        values( new Value [] {
         pair(doubles(new double[]{0.9,0.1}), doubles(new double[]{0.9,0.1})),
         pair(doubles(new double[]{0.1,0.9}), doubles(new double[]{0.1,0.9})),
         pair(doubles(new double[]{0.1,0.9}), doubles(new double[]{0.9,0.1})) }
                )     )       );

   /** A Naive Bayes function-model estimated from data-set {@link #ds}.
       {@link #walk_N_talk_2_bird}.estimator(...).apply({@link #ds})
       is similar to {@link #book_eg} but not quite identical.
       (The estimator's parameters are
       &lang;triv,&lang;triv,triv&rang;&rang;,
       the first triv is for {@link #birdUPM}'s estimator and
       &lang;triv,triv&rang; is for the pair of {@link #boolUPM}
       estimators inside {@link #walk_N_talk}.)  */
   public static NaiveBayes.M estimated = (NaiveBayes.M)           // estimated
      walk_N_talk_2_bird
         .estimator( pair( triv,       // for birdUPM's est...
                           pair(triv, triv) // & walk_N_talk's
                   )     )
         .apply( ds );               // An Estimator maps a data-set to a Model
   // -------------------------------------------------------------------------

   private static void ask(FunctionModel fm, Value obs)
    { System.out.println(
           "If X walks like a duck and talks like a duck, pr(X is a duck) = "
         + x2S( fm.condModel( obs ).pr( duck ) ) + "\n"
         + "using function-model " + fm );
    }//ask(fm)


   /** Walk the walk and talk the talk.  */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run Ducks --");

      Value.Tuple obs = pair(t, t);  // waddles and quacks

      System.out.println( "First, as per the model in the book:\n" );
      ask(book_eg, obs);

      System.out.println( "\nSecond, as estimated from a data-set:\n" );
      ask(estimated, obs);
      System.out.println( "estimated from 'ds',\n"
         + "msg()="
         + x2S(estimated.msg())  + "="
         + x2S(estimated.msg1()) + "+"
         + x2S(estimated.msg2())
         + " nits\n     ="
         + x2S(estimated.msgBits())  + "="
         + x2S(estimated.msg1bits()) + "+"
         + x2S(estimated.msg2bits())
         + " bits"  );

      System.out.println("-- done with Ducks --");
    }//main

 }//Ducks class

// (c) Lloyd Allison, Oct 2017, Sat 9 Dec 2017.  Licence to 'Ducks' is Gnu GPL.
