//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package eg;

import la.la.Type;
import la.la.Value;
import la.maths.Vector;  // A data-set is a Vector of Values
import la.util.*;        // x2String and error
import mml.*;            // For Model, UPModel, Normal, Mixture, Tree, etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.
 *
 * Reference: 
 *   Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.
 *
 * This file is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL) as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this file. If not, see <http://www.gnu.org/licenses/>.
 */
// ----------------------------------------------------------------------------

/** Iris, a simple example of an application program that
    {@link #main runs} some {@link mml.README MML}-analysis
    of "<A HREF="../../FP2013/data/Iris.csv">Iris.csv</A>".
    Includes {@link mml.Mixture Mixture}-Models and
    a (decision-) classification-{@link mml.Tree Tree}.
    Also see
    [<A HREF="https://doi.org/10.1007/978-3-319-76433-7" TARGET="BOOKA">the&nbsp;book</A>]
    and the
    [<A HREF="https://www.cantab.net/users/mmlist/MML/A/" TARGET="mmlist">home&nbsp;page</A>]. */
public class Iris
 {
   /** Return just a few significant digits of&nbsp;x;
       uses {@link la.util.Util}.{@link Util#x2String(double) x2String(x)}. */
   public static String x2S(double x) { return Util.x2String(x, 4); }

   /** The {@link la.la.Type.Enum Enum} Type 'Species' is
       'Iris-setosa | Iris-versicolor | Iris-virginica'. */
   public static Type.Enum Species =
      new Type.Enum( "Species",
         new String [] {"Iris-setosa", "Iris-versicolor", "Iris-virginica"} );

   /** Type 'R4' is 4 {@link la.la.Type#CTS CTS}, i.e., <B>R</B><SUP>4</SUP>,
       that is sepal and petal, length and width. */
   public static Type.Tuple
      R4 = new Type.Tuple.GP("R4",
              new Type [] {Type.CTS, Type.CTS, Type.CTS, Type.CTS});

   /** Type 'R4Species' is a 5-tuple of {@link #R4 dimensions} and
       {@link #Species Species}. */
   public static Type.Tuple
      R4Species = new Type.Tuple.GP("R4Species",
              new Type [] {Type.CTS, Type.CTS, Type.CTS, Type.CTS, Species});

   /** Get the Iris data-set from the named file using {@link Vector#csv}.
       There are five columns (see {@link #R4Species variables}) -- petal and
       sepal, length and width, and the {@link #Species Species}.  */
   public static Vector getData(String filename)
    { java.io.FileInputStream inp = null;
      try{ inp = new java.io.FileInputStream(filename);
         }catch(Exception e){ Util.error("getData("+filename+"): "+e); }
      return Vector.csv(R4Species, inp);
    }//getData(filename)

   /** Using {@link mml.Estimator Estimator} 'e', fit
       a {@link mml.Model Model} to data-set 'ds' and
       show some results. */
   public static Model exercise(String caption, Estimator e, Vector ds)
    { if( caption == null ) caption = "";
      System.out.println( "\n-- " + caption + " --" );
      final int N = ds.nElts();
      System.out.println( " " + N + " data" );
      for( int i = 0; i < N; i ++ )             // show a few data
       { final Value ds_i = ds.elt(i);
         System.out.println( "  " + i + ": " + ds_i );
         if( i == 1 && ds.nElts() >= 6 )
          { System.out.println( " ..." );
            i = N - 3;
       }  }//for i
      Model m = e.apply(ds);                    // e estimates m from ds
      System.out.println( " Estimator: " + e );
      display(m, ds);
      System.out.println( "-- done " + caption + " --" );
      return m;
    }//exercise(,,)

   /** Print out a summary of Model 'm', its 1st and 2nd-part
       message lengths and its performance on data-set 'ds'. */
   public static void display(Model m, Vector ds)
    { System.out.println( " Model    : " + m );
      System.out.println(
         " msg = " + x2S(m.msg())
           + " = " + x2S(m.msg1()) + " + " + x2S(m.msg2()) + " (nits)"
           + " ((sanity check "
               + x2S(m.nlLH(m.stats(ds))) +"~"+ x2S(m.sumNlPr(ds)) + "))\n"
       + "     = " + x2S(m.msgBits())
           + " = " + x2S(m.msg1bits()) + " + " + x2S(m.msg2bits()) + " (bits)\n"
       + " (ds : " + x2S(m.nlLH(m.stats(ds)))  + " (nits) = "
                   + x2S(m.nl2LH(m.stats(ds))) + " (bits))"
       );
    }//display(m)

   /** Compare a no-mixture 'baseEst'-imated Model to a
       'mxEst'-imated Mixture model on data-set 'ds4'
       (the full 'ds5' data-set is <I>only</I> for cross-tabulation). */
   public static void compare(String caption,
                              Estimator baseEst,   // for "1-class" model
                              Mixture.Est mxEst,
                              Vector ds4,          // actual data-set
                              Vector ds5)          // for cross-tabulation
    { System.out.println( "\n-- begin " + caption + " --" );

      Model base_M = exercise( "1. no mixture, baseline", baseEst, ds4);

      Mixture.M   mx4    =  // Parameterised Mixture
         (Mixture.M) exercise("2. Mixture", mxEst, ds4);

      // Cross tabulate inferred classes v. known Species.
      int nSpecies = Species.upb_n()+1, nClasses = mx4.ms().length;
      double [] [] table = new double [nSpecies] [nClasses];
      for( int spcs = 0; spcs < nSpecies; spcs ++ )
         for( int clss = 0; clss < nClasses; clss ++ )
            table [spcs] [clss] = 0;  // initialise

      double [] [] mshps = mxEst.Model2mshp(mx4, ds4);  // NB. by columns!
      for( int i = 0; i < ds5.nElts(); i ++ )           // each datum,
       { int sp_i = ds5.elt(i, 4).n();                  // its species
         for( int clss = 0; clss < nClasses; clss ++ )  // v.
            table [sp_i] [clss]  += mshps [clss] [i];   // classes.
       }//for i

      System.out.println( "\n-- cross tabulate classes and species\n"
                        + "      Classes 0 to " + (nClasses-1) + " ->" );
      for( int spcs = 0; spcs < nSpecies; spcs ++ )
       { System.out.print( "   " + spcs + ": " );
         for( int clss = 0; clss < nClasses; clss ++ )
            System.out.format( "%5.1f,", table[spcs][clss]);
         System.out.println( " -- " + Species.int2value(spcs) );
       }//for spcs

      System.out.println( "\n-- 3. cheat (knowing the 3 species)" );
      double third = 1.0/3.0;
      Model [] ms = new Model [] {          // using
         baseEst.apply(ds4.take(50)),       // info that
         baseEst.apply(ds4.slice(50,100)),  // Mixture
         baseEst.apply(ds4.drop(100)) };    // does not have
      Mixture cheatUPM = new Mixture(baseEst.asUPModel());
      Mixture.M cheat = cheatUPM.sp2Model(
         ms[0].msg1() + ms[1].msg1() + ms[2].msg1(),  // -- msg1, arguably
         0,
         Value.pair(                        // stat params ...
            Vector.doubles( new double[]{third, third, third} ),  // weights
            Vector.values( new Value [] {
               ms[0].statParams(),          // and component ...
               ms[1].statParams(),          // ... sub-models ...
               ms[2].statParams() }         // ...
         ) ) );
      display(cheat, ds4);

      System.out.println( "\n-- done " + caption + " --" );
    }//compare(...)
   // -------------------------------------------------------------------------

   /** Run {@link mml.README MML-analysis} of the Iris data (Iris.csv),
       including a {@link mml.Mixture Mixture}-Model and
       a (decision-) classification-{@link mml.Tree Tree}. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run Iris --");

      Value.Int  four  = Value.four;
      Value.Real zeroR = Value.zeroR, tenR = Value.tenR, pt1 = Value.point1,
               negTenR = Value.negTenR;
      //           i.e.,      0.0    10.0    0.1  -10.0

      Vector ds  = getData("../data/Iris.csv");     // the full data-set, ds
      Vector ds4 = ds.cols(new int[]{0,1,2,3});  // cols 0 to 3 inc., ds4

// Mixture Modelling (clustering) on columns (variables) 0 to 3 inc., ds4, ...
// ... which are assumed to be independent for now ...

      UPModel N4 =                               // i.e., four UnParam. Normals
         new Independent(Value.tuple(            // simplest assumption
            new UPModel[]{MML.Normal, MML.Normal, MML.Normal, MML.Normal} ));
      Value.Tuple Nbounds   =                    // Bounds for prior on...
         Value.quad(zeroR,tenR, pt1,tenR);       // ...mu and sigma,
      Value.Tuple estParams =                    // for the four Normals,
         Value.quad(Nbounds, Nbounds, Nbounds, Nbounds);  // say
      Estimator N4Est = N4.estimator(estParams); // and here's our base Est

      // Now, a Mixture of one or more N4 classes over ds4.
      Mixture mx4UPM = new Mixture(N4);          // UnParameterised Mixture
      Mixture.Est mx4Est =                       // and its Estimator
         (Mixture.Est) mx4UPM.estimator(estParams);

      compare("N4, Normal x 4 over R^4", N4Est, mx4Est, ds4, ds);

// ... and now with some given linear dependencies ...

      int [] parents = new int [] {-1, 0, 0, 2}; // i.e., 0->1, 0->2, 2->3.
      R_D.Forest F = new R_D.Forest( Vector.ints(parents) );
      Value.Tuple FestParams = Value.tuple( new Value [] // mu, sig, b ...
                             {zeroR,tenR, pt1,tenR, negTenR,tenR} );
      R_D.Forest.Est Fest = F.estimator(FestParams);
      Mixture mxFUPM = new Mixture(F);
      Mixture.Est mxFest = (Mixture.Est) mxFUPM.estimator(FestParams);

      compare("R_D.Forest over R^4", Fest, mxFest, ds4, ds);

// ... and now search for the best linear dependencies ...

      R_D.ForestSearch ml1 = new R_D.ForestSearch(four);
      R_D.ForestSearch.Est ml1est = ml1.estimator(FestParams);
      exercise("R_D.ForestSearch, search for dependencies", ml1est, ds4);

// And now for something different  ...
// Classification Tree on cols 0..3 -> col4, that is R4 -> Species ...

      MultiState speciesDistn =             // UnParam. Model over Species
         new MultiState(Species.bounds());

      // First, a Model over Species only for a base-line message length.
      exercise("Species Distribution",
               speciesDistn.estimator(Value.triv), ds.col(4));

      // Now, an UnParameterised classification-Tree;
      // input Type is R4, and Leaf UnParam.-Model is speciesDistn.
      Vector dsTr = ds4.zip(ds.col(4));     // i.e., ((w,x,y,z), s)*, data-set
      Tree treeUP = new Tree( Value.pair(R4, speciesDistn) );  // UnParam.Tree
      Tree.Est treeEst = (Tree.Est) treeUP.estimator(Value.triv);
      exercise("Classification (Decision) Tree, R4->Species", treeEst, dsTr);
      //        caption,                                    estimator, data-set

      System.out.println("-- done Iris --");
    }//main

 }//Iris class

// (c) L.Allison, Jun 2014, 5 July 2017. The licence to 'Iris.java' is Gnu GPL.
