//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package graph;

import la.util.Series;
import la.la.Value;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of {@link Graph Graph} {@link la.la.Type Types}.
    It specifies {@link #isDirected Directed?},
    {@link #selfLoops selfLoops?}, and
    {@link #vType Vertex-} &amp; {@link #eType Edge-} labelling. */
public class Type extends la.la.Type.Structured                // (graph.) Type
 {
   /** Is the Graph Directed, and are self-loops allowed? */
   public final boolean isDirected, selfLoops;

   /** Does the Graph have Vertex- and/or Edge- labels,
       and if so, of what Type(s)? */
   public final la.la.Type vType, eType;

   /** Construct a Graph Type with no Vertex or Edge label Types,
       that is unlabelled. */
   public Type(boolean isDirected)
    { this("unlabelled", isDirected); }//constructor

   /** Assumes no self-loops, and not Vertex- or Edge- labelled. */
   public Type(String name, boolean isDirected)
    { this(name, isDirected, /* self Loops= */ false); }//constructor

   /** Assumes not Vertex- or Edge- labelled. */
   public Type(String name, boolean isDirected, boolean selfLoops)
    { this(name, isDirected, selfLoops, null, null); }//constructor

   /** Construct a Graph Type with specified isDirected(?), selfLoops(?),
       Vertex label Type 'vType', and
       Edge   label Type 'eType. */
   public Type(String name, boolean isDirected, boolean selfLoops,
               la.la.Type vType, la.la.Type eType)
    { super(name);
      this.isDirected = isDirected;
      this.selfLoops  = selfLoops;
      this.vType = vType; this.eType = eType;
    }//constructor

   /** Does {@link Graph Graph} g belong to (satisfy) 'this' Graph Type? */
   public boolean ni(Value g)                                          // ni(g)
    { if( ! (g.type() instanceof Type)
       || ! checkProperties((Graph) g) ) return false;
      Type gt = (Type) g.type();
      return vType == gt.vType  // could be null
          && eType == gt.eType; // ? too strong ?
    }//ni(v)

   /** Return a "cloned" Type but with 'isDirected' set as indicated. */
   public Type clone(boolean isDirected)                            // clone(d)
    { return new Type(name, isDirected, selfLoops, vType, eType); }

   /** What is the maximum number of Edges for a Graph of this Type with
       'vSize' Vertices? (If allowed, a self-loop counts as one.) */
   public int maxEdges(int vSize)
    { return isDirected ? (selfLoops ? vSize*vSize     : (vSize-1)*vSize)
                        : (selfLoops ? (vSize+1)*vSize : (vSize-1)*vSize) >> 1;
    }//maxEdges(vSize)

   /** Check that Graph 'g' does satisfy the requirements of 'this' Type.
       I.e., check for symmetry of Undirected Graphs, and absence of
       &lang;v,v&rang; of &not;selfLoops.  O(|V|<SUP>2</SUP>)-time. */
   public boolean checkProperties(Graph g)
    { if( ! isDirected ) // i.e., isUndirected
         for( int v = 0; v < g.vSize(); v ++ )
          { Series.Discrete ws = g.joinedTo(v); // v--w?
            while( ws.hasSome() )
             { final int w = ws.elt_n();
               if( ! (g.isEdge(v, w) && g.isEdge(w, v)) ) return false;
               ws.advance();
          }  }//for v
      if( ! selfLoops )
         for( int v = 0; v < g.vSize(); v ++ )
            if( g.isEdge(v, v) ) return false;
      return true;  // we did not fail
    }//checkProperties(g)

   public int n() { return -1; } // tag ???

   /** Four, see {@link #elt(int) elt(i)}. */
   public int nElts() { return 4; }

   /** 0:{@link #isDirected isDirected}, 1:{@link #selfLoops selfLoops},
       2:{@link #vType vType},           3:{@link #eType eType}. */
   public Value elt(int i)
    { switch(i)
       { case 0: return isDirected ? ttrue : ffalse;
         case 1: return selfLoops  ? ttrue : ffalse;
         case 2: return vType;
         case 3: return eType;
         default: throw RTE( ".elt(" + i + ")" );
    }  }//elt(i)

 }//(graph.)Type class

// (c) Lloyd Allison July 2014, Fri 31 Oct 2014. Licence to 'Type' is Gnu GPL.
