//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Class Expression defines the abstract-syntax
    (i.e., parse tree) of Expressions in the language;
    also see {@link Value}. */
public abstract class Expression
 {
   /** The use of a StringBuffer, sb, gives linear complexity when
       printing large Values; also see {@link #toString()}. */
   public abstract void  appendSB(StringBuffer sb);             // appendSB(sb)

   /** The use of {@link #appendSB}, and hence a StringBuffer,
       gives linear rather than quadratic complexity. */
   public String toString()                                       // toString()
    { StringBuffer sb = new StringBuffer(); // efficiency!
      this.appendSB(sb);
      return sb.toString();
    }//toString

   /** n(), the numerical code (tag) of a subclass of Expression,
       for example, for <TT>switch</TT>-ing. */
   public abstract int n();                                            // * n()

   /** In general, create a {@link Value.Defer Defer(this,r)},
       a "thunk", to evaluate 'this' Expression later, but also
       see {@link Const#defer} and {@link LambdaExp#defer}. */
   public Value defer(Environment r)                                // defer(r)
    { return new Value.Defer.Exp(this, r); }

   /** A tag; see {@link #n()}. */
   public static final int  // i.e., the Expression "tag" numbers (see n()) ...
         IDENT = 0, CONST = 1, TUPLE = 2, LAMBDAEXP = 3,
         APPLICATION = 4, UNARY = 5, BINARY = 6, IFEXP = 7, BLOCK = 8;

   /** EVALuate 'this' {@link Expression}, using Environment&nbsp;'r',
       to produce a {@link Value}.  Implements the tail-recursion
       optimisation.  Acts directly for {@link Application},
       {@link Block} and {@link IfExp}.  */
   public Value eval(Environment r)                                  // eval(r)
    { Expression e = this;
      while( true )
         switch( e.n() )
          { case APPLICATION:                                    // (\ x.e)p
               Application ap = (Application) e;
               final Value f = ap.fun.eval(r);
               if( f instanceof Function.Native )   // Native
                  return ((Function.Native) f).apply(ap.aparam.defer(r));
            /* else a (lambda) 'Function' ... */
               Value.Lambda func = (Value.Lambda) f;
               if( func.code.fparam == Expression.triv )  // (\().e) ...
                { if( ap.aparam.eval(r) != Value.triv )   // need (\().e)()
                   { error("(lambda ().e) " + ap.aparam + ",  AP is not ()"); }
                  r = func.r;  // else all is good
                }
               else // ((\ x.e1) e2
                { Ident fparam = (Ident) func.code.fparam;
                  Value apv = ap.aparam.defer(r);
                  r = func.r.bind(fparam.id, apv);
                }
               e = func.code.body;  // now eval the func body (i.e. apply)
               // Note that now the Env, r, is, or is derived from, func.r
               continue;  // i.e., loop, not recurse
            case BLOCK:                                   // let decs in exp
               Block b = (Block) e;
               r = r.D(b.d);  e = b.e; // now eval the body of the block
               continue;  // i.e., loop, not recurse
            case IFEXP:                             // if e1 then e2 else e3
               IfExp ife = (IfExp) e;
               e = ((Value.Bool) ife.e1.eval(r)).b ? ife.e2 : ife.e3;
               continue;  // i.e., loop, not recurse
            default:
               // we must not(!) return a Value.Defer(e,r) here...
               return  e.eval(r);             // NB. may be on n-th iteration
          }//switch
    }//eval(r)

   /** An Identifier, 'id', as (it should have been) declared in
       <NOBR><TT> {@link Block let} id = e in e</TT>,</NOBR> or
       <NOBR> {@link LambdaExp &lambda;id.e} .</NOBR>  */
   public static class Ident extends Expression                        // Ident
    { public final String id;
      /** Set by {@link Environment#lookup}. */
      public int levels = -1, offset = -1; // i.e., unknown, for now
      public Ident(String id) { this.id = id; }
      public int n() { return IDENT; }
      /** {@link Environment#lookup lookup} 'this' Ident's Value in 'r'.  */
      public Value eval(Environment r) { return r.lookup(this); }
      public void appendSB(StringBuffer sb) { sb.append(id); }
    }//Ident class

   /** Various Constants (literals) of the language; can be
       <TT>{@link #eval eval}-uated</TT> at "compile" time;
       see {@link #defer defer}.  */
   public static class Const extends Expression                        // Const
    { /** The Value denoted by 'this' Const (literal). */
      public final Value v;
      public Const(Value v) { this.v = v; }
      public int n() { return CONST; }
      /** It is safe to return Value {@link #v}, which is known at
          "compile"-time; also see {@link Expression#defer}. */
      public Value defer(Environment e) { return v; }   // !
      /** Note, a Const was evaluated, to {@link #v}, at "compile"-time. */
      public Value eval(Environment e) { return v; }
      public void appendSB(StringBuffer sb) { sb.append(v); }
    }//Const class


   // NB. It pays to have a single copy of the following...
   /** <TT>{@link Const}ant</TT> true;  see {@link Value#ttrue}. */
   public static final Const ttrue  = new Const(Value.ttrue);          // ttrue
   /** <TT>{@link Const}ant</TT> false; see {@link Value#ffalse}. */
   public static final Const ffalse = new Const(Value.ffalse);        // ffalse
   /** <TT>{@link Const}ant</TT> triv;  see {@link Value#triv}. */
   public static final Const triv  = new Const(Value.triv);             // triv
   /** <TT>{@link Const}ant</TT> nil;   see {@link Value#nilVal}. */
   public static final Const nilCon = new Const(Value.nilVal);        // nilCon


   /** <TT>(e0, e1, ...)</TT>; also see {@link Value.Tuple}. */
   public static class Tuple extends Expression                        // Tuple
    { public final Expression [] subExps;
      public Tuple(Expression [] subExps)
       { this.subExps = subExps; }
      public int n() { return TUPLE; }
      /** A Tuple Expression evaluates to a Tuple Value. */
      public Value eval(Environment r)
       { Value [] subVals = new Value [subExps.length];
         for( int i = 0; i < subExps.length; i ++ )
            subVals[i] = subExps[i].defer(r);  // note lazy
         return Value.tuple(subVals);
       }//eval(r)
      public void appendSB(StringBuffer sb)
       { sb.append(Lexical.Symbol[Lexical.open]);
         for( int i = 0; i < subExps.length; i ++ )
          { if( i > 0 ) sb.append(Lexical.Symbol[Lexical.comma] + " ");
            subExps[i].appendSB(sb);
          }
         sb.append(Lexical.Symbol[Lexical.close]);
       }//appendSB(sb)
    }//Tuple class


   /** <TT> &lambda;&nbsp;x.e </TT> denotes a {@link Function}. */
   public static class LambdaExp extends Expression                // LambdaExp
    { public final Expression fparam, body; // lambda x.e  or  lambda ().e
      public LambdaExp(Expression fparam, Expression body)
       { this.fparam = fparam;  this.body = body; }
      public int n() { return LAMBDAEXP; }
      /** It is in fact safe to return the {@link #eval eval}-uated
          {@link Value.Lambda &lambda;-Function} (which contains the
          given 'r'), after all &lambda;-Function and
          {@link Value.Defer} are very similar.
          (Also see {@link Const#defer}.)  */
      public Value defer(Environment r) { return eval(r); }
      /** Note, evaluate 'this' Lambda Expression to a Function, <I>not</I>
          call a function.  Return the Function Value itself. */
      public Value.Lambda eval(Environment r)
       { return new Value.Lambda(this, r); }
      public void appendSB(StringBuffer sb)
       { sb.append(Lexical.Symbol[Lexical.lambdaSy] + " ");
         fparam.appendSB(sb);  sb.append( ".");
         body.appendSB(sb);
       }
    }//LambdaExp class


   /** A Function application <TT>(e1&nbsp;e2)</TT>, where
       e1 must evaluate to a {@link Function}.  */
   public static class Application extends Expression            // Application
    { /** Represents the application of 'fun' to 'aparam'. */
      public final Expression fun, aparam;
      public Application(Expression fun, Expression aparam)
       { this.fun = fun; this.aparam = aparam; }
      public int n() { return APPLICATION; }
  /// public Value eval(Environment r)   NOT In Use
  ///  { Function func = (Function)(fun.eval(r));
  ///    return func.apply( aparam.defer(r) ); // NB. lazy
  ///  }
      public void appendSB(StringBuffer sb)
       { sb.append("("); fun.appendSB(sb);  sb.append( " " );
         aparam.appendSB(sb); sb.append(")");
       }
    }//Application class, e.g., f(x)


   /** <TT>(op e)</TT>; yes it's just a
       special kind of {@link Application}. */
   public static class Unary extends Expression                        // Unary
    { /** The unary operator, e.g., notSy, etc.. */
      public final int opr;
      /** The subexpression, e.g., 'exp' as in 'not&nbsp;exp'. */
      public final Expression e;
      public Unary(int opr, Expression e) { this.e = e; this.opr = opr; }
      public int n() { return UNARY; }
      /** Apply 'this' unary operator, {@link #opr opr}, to the
          sub-Expression in the given Environment&nbsp;'r'.
          Note, all unary operators are strict. */
      public Value eval(Environment r) { return e.eval(r).uOp(opr); }
      public void appendSB(StringBuffer sb)
       { sb.append( "(" + Lexical.Symbol[opr] + " " );
         e.appendSB(sb);  sb.append( ")" );
       }
    }//Unary class


   /** <TT>(e1 op e2)</TT>; yes it's just a
       special kind of {@link Application}. */
   public static class Binary extends Expression                      // Binary
    { /** The binary operator, e.g., '+', etc.. */
      public final int opr;
      /** lft and rgt, the left and right sub-Expressions. */
      public final Expression lft, rgt;
      public Binary(int opr, Expression lft, Expression rgt)
       { this.opr = opr; this.lft = lft; this.rgt = rgt; }
      public int n() { return BINARY; }
      /** Apply 'this' binary operator, {@link #opr opr}, to
          {@link #lft lft} and {@link #rgt rgt} sub-Expressions
          in the given Environment&nbsp;'r'.
          Note, 'cons' is lazy, 'and' &amp; 'or' are half-lazy,
          '+', '*', etc., are strict.  */
      public Value eval(Environment r)
       { if( opr == Lexical.consSy )    // NB. cons(tructors) lazy
            return new Value.List.Cell(lft.defer(r), rgt.defer(r));
         else
            return lft.eval(r).bOp(opr, rgt.defer(r)); // lft strict
       }//eval
      public void appendSB(StringBuffer sb)
       { sb.append( "(" );
         lft.appendSB(sb);  sb.append( " " + Lexical.Symbol[opr] + " " );
         rgt.appendSB(sb);  sb.append( ")" );
       }
    }//Binary class


   /** <TT>(if e1 then e2 else e3)</TT>; yes it's
       just a special kind of {@link Application} -- of
       <TT>cond&nbsp;=&nbsp;\x1.\x2.\x3.x1&nbsp;x2&nbsp;x3</TT>  */
   public static class IfExp extends Expression                        // IfExp
    { public final Expression e1, e2, e3;
      public IfExp(Expression e1, Expression e2, Expression e3)
       { this.e1 = e1; this.e2 = e2; this.e3 = e3; }
      public int n() { return IFEXP; }
  /// public Value eval(Environment r)   NOT In Use
  ///  { return ((Value.Bool)(e1.eval(r)) == Value.ttrue ? e2 : e3).eval(r); }
      public void appendSB(StringBuffer sb)
       { sb.append( Lexical.Symbol[Lexical.ifSy] + " " );
         e1.appendSB(sb);
         sb.append( " " + Lexical.Symbol[Lexical.thenSy] + " " );
         e2.appendSB(sb);
         sb.append( " " + Lexical.Symbol[Lexical.elseSy] + " " );
         e3.appendSB(sb);
       }
    }//IfExp class


   /** <TT>let &lt;{@link Declaration Decs}&gt; in e</TT>. */
   public static class Block extends Expression                        // Block
    { /** Represents <NOBR><TT> let [rec] d in e </TT></NOBR>  */
      public final Declaration d; public final Expression e;
      public Block(Declaration d, Expression e) { this.d = d; this.e = e; }
      public int n() { return BLOCK; }
  /// public Value eval(Environment r)   NOT In Use
  ///  { return e.eval( r.D(d) ); /* eval e in extended environment r++d */ }
      public void appendSB(StringBuffer sb)
       { sb.append(Lexical.Symbol[Lexical.letSy]+ " ");
         d.appendSB(sb);
         sb.append("\n" + Lexical.Symbol[Lexical.inSy] + " ");
         e.appendSB(sb);
       }
    }//Block class, e.g., let [rec] x1=e1, x2=e2, ... in e2


   /** Throw a RuntimeException.  */
   public void error(String msg)                                  // error(msg)
    { la.util.Util.error("Expression (n=" + n() + "): " + msg); /* throws */ }
    // NB. a RuntimeException need not be mentioned in a method's throws list.

 }//Expression class   http://www.allisons.org/ll/

// (c) Lloyd Allison, Oct 2007, 14 Dec 2016. Licence to 'Expression' is Gnu GPL.
