//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of Functions; note that a Function is a {@link Value}.
    Also see {@link Value.Lambda}, and the standard {@link Library}. */
public abstract class Function extends Value                      // * Function
 {
   /** Return {@link Type#FUNCTION}. */
   public Type type()
    { return Type.FUNCTION; }                                      // type()

   /** Apply 'this' Function to parameter 'p'. */
   public abstract Value apply(Value p);                       // * apply(p)

   /** Return the String "Function". */
   public String toString()
    { return "Function"; }                                     // toString()


   /** A unary operator on 'this' Function f, such that
       (&lt;op&gt;&nbsp;f)(x)&nbsp;=&nbsp;&lt;op&gt;&nbsp;(f(x)).<BR>
       That is,
       <NOBR> f.uOp(op).{@link #apply apply(x)} =
              f.apply(x).{@link Value#uOp uOp(op)}. </NOBR>
       Also see {@link Cts2Cts#uOp} and {@link Value#uOp}.  */
   public Function uOp(final int op)                              // uOp(op)
    { return new Function()
       { public Value apply(Value x)
          { return Function.this.apply(x).uOp(op); }
       };//new Function
    }//uOp(op)

   /** A binary operator on Functions 'this' (f) and 'g', such that
   (f&nbsp;&lt;op&gt;&nbsp;g)(x)&nbsp;=&nbsp;f(x)&nbsp;&lt;op&gt;&nbsp;g(x).
       Also see {@link Value#bOp}.  */
   public Function bOp(final int op, final Function g)          // bOp(op,g)
    { return new Function()
       { public Value apply(Value x)
          { return Function.this.apply(x).bOp(op, g.apply(x)); }
       };//new Function
    }//bOp(op,g)
   // =========================================================================

   /** The class of Cts&nbsp;&rarr;&nbsp;Cts Functions.
       Define {@link #apply_x(double) apply_x(.)} and all else follows.
       A Cts2Cts has a {@link #d_dx() Derivative} and
       an {@link #integral() Integral}. */
   public static abstract class Cts2Cts extends Native             // * Cts2Cts
    { /** apply_x(x) is used by {@link #apply apply(c)}. */
      public abstract double apply_x(double x);              // * apply_x(x)

      /** Calls {@link #apply_x apply_x(c.x())}.  If c.AoM() &gt;zero, it
          also calls the {@link #d_dx() Derivative} to get an
          {@link Value.Cts#AoM() AoM()} for the result.  */
      public Cts apply(Value c)
       { Cts cc = (Cts) c;
         double aom = cc.AoM(), x = cc.x(), fx = apply_x(x);
         return aom == 0 ? real(fx)
                         : cts(fx, aom*d_dx().apply_x(x));
       }//apply(x)

      /** (Static) fromFunction(f) creates a Cts2Cts from a Function,
          'f', where f is not an instance of Cts2Cts but f does have
          the type Cts&nbsp;&rarr;&nbsp;Cts.  */
      public static Cts2Cts fromFunction(final Function f)
       { return new Cts2Cts()
          { public double apply_x(double x)
             { return f.apply(real(x)).x(); }
       }; }//fromFunction(f)

      /** For example, sin.uOp(minus) satisfies
          (sin.uOp(minus))(x)&nbsp;=&nbsp;-sin(x).  */
      public Cts2Cts uOp(final int op)                           // uOp(op)
       { return new Cts2Cts()
          { public double apply_x(double x)
             { return real(Cts2Cts.this.apply_x(x)) .uOp(op) .x(); }
          };//new Cts2Cts
       }//uOp(op)

      /** Find a turning point of 'this' Cts2Cts, 'f', that is find x such
          that {@link Cts2Cts#d_dx() f'(x)}=0, given an initial guess, x0.
          Uses {@link #root root}. */
      public double turning(double x0)                       // turning(x0)
       { return this.d_dx().root(x0); }

      /** Solve f(x)&nbsp;=&nbsp;0, where 'f' is 'this' Cts2Cts, given
          an initial guess, x0.
          Currently it uses {@link #NewtonRaphson NewtonRaphson(x0)}. */
      public double root(double x0)                             // root(x0)
       { return this.NewtonRaphson(x0); }

      /** Use Newton-Raphson to solve f(x)&nbsp;=&nbsp;0, where
          'f' is 'this' Cts2Cts, given an initial guess, x0.
          The central step is <NOBR>x1 = x0 - f(x0) / df_dx_(x0)</NOBR>;
          it helps if f has a closed-form derivative,
          {@link Cts2Cts#d_dx() d_dx}.
          (If Newton-Raphson fails, plan&nbsp;B is binary search.)  */
      public double NewtonRaphson(double x0)
       { final Cts2Cts f = this;       // the function in question
         final int    max   = 30;      // max iterations
         final double delta = 10E-7;   // find x to within one in 1/delta
         final double flat  = delta*delta;  // lower limit on the slope
         final Cts2Cts df_dx = f.d_dx();    // derivative of f
         double x1, y0 = 0, df_dx_x0,
                yp = Double.MAX_VALUE, ypx = 0, ym = -yp, ymx = 0;
         boolean p = false, m = false;
         for( int i = 0; i < max; i ++ )
          { y0 = f.apply_x(x0);
            switch( (int) Math.signum(y0) )
             { case  1: if( y0 < yp ) { p = true; yp = y0; ypx = x0; }; break;
               case -1: if( y0 > ym ) { m = true; ym = y0; ymx = x0; }; break;
               default: return x0;  // only in some weird universe
             }// (Note, ym < 0 < yp, but it is possible that ymx > ypx.)
            df_dx_x0 = df_dx.apply_x(x0);
            if( Math.abs(df_dx_x0) < flat ) // slope too flat
             { if( p && m ) break;
               /*else*/ return x0;
             }
            // the next line is the heart of the Newton-Raphson method...
            x1 = x0 - y0/df_dx_x0;  // i.e., project tangent onto x-axis
            double mag = Math.abs(x1 == 0 ? x0 : x1);
            if( Math.abs(x1 - x0) <= delta*mag ) return x1;          // success
            x0 = x1;
          }//for i

         // If Newton-Raphson did not work, there is a plan B but ...
         double power = 0.1;  // ... but we need both a +ve and -ve f(.)
         if( ! p )  // desperation (maybe f' was too flat?) ...
            for( int i = 0; i < max; i ++ )
             { ypx = x0 + power;      // seek an x s.t. f(x) >= 0
               y0 = f.apply_x(ypx);
               if( y0 >= 0 ) { p = true; yp = y0; break; }
               ypx = x0 - power;      // or this
               y0 = f.apply_x(ypx);
               if( y0 >= 0 ) { p = true; yp = y0; break; }
               power *= 3;
             }
         else /* p */ if( ! m )
            for( int i = 0; i < max; i ++ )
             { ymx = x0 + power;      // seek an x s.t. f(x) <= 0
               y0 = f.apply_x(ymx);
               if( y0 <= 0 ) { m = true; ym = y0; break; }
               ymx = x0 - power;      // or this
               y0 = f.apply_x(ymx);
               if( y0 <= 0 ) { m = true; ym = y0; break; }
               power *= 3;
             } /* else p && m */

         if( p && m )                        // N-R failed, can we ...
          { double mid = ymx, fmid;
            for( int i = 0; i < max; i ++ )  // ... give binary search a try
             { mid = (ymx + ypx) / 2.0;
               fmid = f.apply_x(mid);        // ymx < mid < ypx, or v.v.
               switch( (int) Math.signum(fmid) )
                { case  1: yp = fmid; ypx = mid; break;
                  case -1: ym = fmid; ymx = mid; break;
                  default: return mid;
                }
               if( Math.abs(ymx - ypx) <= delta ) return mid; // bleh
             }//for i
            return mid;  // best I could do ???
            // throw new RuntimeException( "x in [" + ymx + "," + ypx + "], "
            //                         +  "y in [" + ym  + "," + yp  + "]" );
          }// else !p || !m ...
         throw new RuntimeException( "root(" + f + "," + x0
            + "), p=" + p + ", m=" + m
            + ", x0="+x0+", y0="+y0 );  // failure
       }//NewtonRaphson(x0)
      // ----------------------------------------------------------------------

      private Cts2Cts derivative = null; // cache; see below

      /** Return the derivative of 'this' function, d/dx&nbsp;this.
          Caches the result of {@link #make_d_dx()}.  */
      public final Cts2Cts d_dx()                                 // d_dx()
       { if( derivative == null) derivative = make_d_dx();
         return derivative;
       }//d_dx()

      /** The second derivative, d<SUP>2</SUP>/dx<SUP>2</SUP>&nbsp;this,
          that is {@link #d_dx() d_dx()} twice. */
      public final Cts2Cts d2_dx2()
       { return d_dx().d_dx(); }

      /** Slave worker for {@link #d_dx() d_dx()}.  This default
          returns new&nbsp;{@link Derivative Derivative(this)};
          <I>consider overriding it is there is a closed form</I>.  */
      protected Cts2Cts make_d_dx() { return new Derivative(this); }


      private Integral integral = null; // cache; see below

      /** Return the indefinite integral of 'this' function,
          <FONT SIZE="+2">&int;</FONT>&nbsp;this(x)&nbsp;dx.
          Caches the result of {@link #make_integral()}. */
      public final Integral integral()                         // integral()
       { if( integral == null ) integral = make_integral();
         return integral;
       }//integral()

      /** Slave worker for {@link #integral() integral()}. This default
          returns new&nbsp;{@link Integral Integral(this)};
          <I>consider overriding it if there is a closed form</I>,
          as in {@link Derivative#make_integral() Derivative}.  */
      protected Integral make_integral() { return new Integral(this); }

    }//Cts2Cts class
   // -------------------------------------------------------------------------

   /** The default Derivative uses
       finite-{@link #apply_xx(double,double) differences}. The key
       method is {@link #apply_xx(double,double) apply_xx(&delta;,x)}.
       Also see Cts2Cts.{@link Cts2Cts#d_dx() d_dx()} and
       Cts2Cts.{@link Cts2Cts#make_d_dx() make_d_dx()}.  */
   public static class Derivative extends Cts2Cts                 // Derivative
    { /** The Cts2Cts "f" of which "this" is the Derivative,&nbsp;f'. */
      public final Cts2Cts f;

      /** Construct the Derivative of the Cts2Cts, '{@link #f f}'. */
      public Derivative(Function f)
       { this.f = (Cts2Cts) f; }//constructor

      /** Call {@link #apply_xx(double,double) apply_xx(&delta;,x)} for
          some "small" &delta;. If you don't want to take your chances
          with &delta; being chosen for you, set your own with
          {@link #apply_xx(double,double) apply_xx(&delta;,x)}. */
      public double apply_x(double x)                       // apply_x(x)
       { final double delta = 0.001,  // ??? Util.epsilon ???
                      xp    = Math.abs(x);
         return apply_xx(delta * (xp > 1 ? 1 :  xp > delta ? xp :  delta), x);
       }//apply_x(x)

      /** Compute the Derivative of {@link #f f} by finite differences here.
          Called by {@link #apply_x(double) apply_x(x)}.
          Return <NOBR>(f(x+&delta;/2) - f(x-&delta;/2) / &delta;</NOBR>.
          Override it if there is something better (a closed form). */
      public double apply_xx(double delta, double x)
       { final double delta2 = delta * 0.5;
         return (f.apply_x(x+delta2) - f.apply_x(x-delta2)) / delta;
       }//apply_xx(delta,x)

      /** The Derivative of 'this' is the second Derivative of
          {@link #f f}, by finite differences.  */
      protected Derivative make_d_dx()
       { final Cts2Cts f0 = f;  // think about it!
         return new Derivative(this)
          { public double apply_xx(double delta, double x)
             { final double delta2 = delta * 0.5;
               // ((f(x+d2)-f(x))/d2 - (f(x)-f(x-d2))/d2) / d2 = ...
               return (f0.apply_x(x + delta2) + f0.apply_x(x - delta2)
                       - 2*f0.apply_x(x)) / (delta2*delta2);
             }//apply_xx(delta,x)
       }; }//make_d_dx()

      /** The indefinite {@link Integral Integral} of 'this'
          Derivative is {@link #f f}, of course.<BR>  So,
          <SUB>lo</SUB><FONT SIZE="+2">&int;</FONT><SUP>x</SUP> this(x) dx
          =&nbsp;f(x)&nbsp;&minus;&nbsp;f(lo).  */
      protected Integral make_integral()
       { return new Integral(this)
          { /** Override the default (and ignore n). */
            public double apply_x(int n, double lo, double x)
             { return f.apply_x(x) - f.apply_x(lo); }
       }; }//make_integral()

    }//Derivative class
   // -------------------------------------------------------------------------

   /** Integral : int &rarr; real &rarr; real &rarr; real. The key
       method is {@link #apply_nxx(int,double,double) apply_nxx(n,lo,x)}.
       The Integral is done by Simpson's rule.  Note that
       {@link #apply2 (this(n))(lo)} is a {@link Cts2Cts Cts2Cts}.
       Also see {@link Cts2Cts#make_integral Cts2Cts.integral()}
       and {@link Cts2Cts#make_integral Cts2Cts.make_integral()}.  */
   public static class Integral extends Native3                     // Integral
    { /** 'This' Function.Integral is the Integral of&nbsp;'f'. */
      public final Cts2Cts f;

      /** Construct the {@link #apply3 Integral} of Cts2Cts 'f'. */
      public Integral(Function f)
       { this.f = (Cts2Cts) f; }//constructor

      /** Makes use of {@link #apply_nxx apply_nxx(n,lo,x)}.
          Note, ({@link Integral this}(n))(lo) is a Cts2Cts the
          {@link Cts2Cts#d_dx derivative} of which is&nbsp;{@link #f f}.
          E.g.,&nbsp;if f(z)&nbsp;=&nbsp;z say,
          <NOBR> g(z) = &int;<SUB>lo..z</SUB> x dx     </NOBR>
          <NOBR> = {x<SUP>2</SUP>/2}<SUB>lo..z</SUB>   </NOBR>
          <NOBR> = (z<SUP>2</SUP> - lo<SUP>2</SUP>)/2, </NOBR>
          then g'(z)&nbsp;=&nbsp;z&nbsp;= f(z).  */
      public Cts2Cts apply2(final Value n, final Value lo)
       { return new Cts2Cts()
          { public double apply_x(double x)
             { return apply_nxx(n.n(), lo.x(), x); }
            /** The Derivative <I>with respect to x</I> of
                'this' Integral is {@link #f f}, of course. */
            public Cts2Cts make_d_dx() { return f; }
       }; }//apply2(n,x)

      /** {@link #apply_nxx(int,double,double) Integrate} {@link #f f}
          from 'lo' to 'x' in 'n' steps.  Method 'apply3' is what
          makes Integral a {@link Native3}.
          Makes use of {@link #apply_nxx apply_nxx(n,lo,x)}.  */
      public Value apply3(Value n, Value lo, Value x)
       { return real( apply_nxx(n.n(), lo.x(), x.x()) ); }

      /** Integrate {@link #f f} from 'lo' to 'x' in 'n' steps using
          Simpson's rule;  override it if there is a closed form.
          'n'&nbsp;is <I>forced</I> to be &ge;3, and odd. */
      public double apply_nxx(int n, double lo, double x)    // apply_nxx(,,)
       { assert lo < x : (lo + ">=" + x + " in Integral");
         // NB. n should be odd (and >=3), so n-1 must be even (and >=2)!
         final int n_1 = n < 3 ? 2 : (n&1)==0 ? n : n-1;
         double sum = f.apply_x(lo) + f.apply_x(x);  // steps 0 and n-1
         int m = 4; // 4, 2, 4, ..., 2, 4.
         for( int i = 1; i < n_1; i ++ )             // steps 1 ... n-2
          { sum += m * f.apply_x((i*lo + (n_1-i)*x) / n_1); // interpolate
            m = 6 - m;  // 2, 4, 2, ...
          }//for i
 // System.out.print( "{" + "n="+n + ",lo="+lo + ",x="+x + ",sum="+sum + "}" );
         return (x - lo) * sum / (3*n_1);  // i.e. width * average
       }//apply_nxx(lo,x,n)

      /** The Derivative <I>with respect to x</I> of
          'this' Integral is {@link #f f}, of course. */
      // public Cts2Cts d_dx() { return f; }
      // NB. It is not worth caching d_dx() !

    }//Integral class
   // -------------------------------------------------------------------------

   /** Native, <TT>{@link Function}s</TT> specified by arbitrary Java
       code in {@link #apply apply(v)}. Also see {@link Library}.  */
   public static abstract class Native extends Function             // * Native
    {
      /** apply(v) does whatever you specify to implement 'this' Native
          Function <U>BUT</U> it must return a Value in <U>WHNF</U>;
          see {@link #force}. */
      public abstract Value apply(Value v);                      // * apply
    }//Function.Native class
   // -------------------------------------------------------------------------

   /** <I>Curried</I> native <TT>{@link Function}s</TT> specified by
       arbitrary Java code in {@link #apply2 apply2(v0,v1)}.
       Note, Native2 extends (i.e., is&nbsp;a) {@link Native}.  */
   public static abstract class Native2 extends Native             // * Native2
    {
      /** apply2(v0,v1) does whatever you specify, <U>BUT</U> it
          must return a Value in <U>WHNF</U>; see {@link #force}. */
      public abstract Value apply2(Value v0, Value v1);          // * apply2

      /** Take the parameter 'v0' and return a {@link Native} which takes
          the parameter 'v1' to produce the result (this(v0))(v1).
          Hence 'this' is a curried Function.  */
      public Native apply(final Value v0)
       { return new Native() // i.e., a Function
          { public Value apply(Value v1) { return apply2(v0, v1); } };
       }
    }//Function.Native2 class
   // -------------------------------------------------------------------------

   /** <I>3-Curried</I> native <TT>{@link Function}s</TT> specified by
       arbitrary Java code in {@link #apply3 apply3(v0,v1,v2)}.
       Note, Native3 extends (i.e., is&nbsp;a) {@link Native2} and
       hence {@link Native}. */
   public static abstract class Native3 extends Native2            // * Native3
    {
      /** apply3(v0,v1,v2) does whatever you specify, <U>BUT</U> it
          must return a Value in <U>WHNF</U>; see {@link #force}.  */
      public abstract Value apply3(Value v0, Value v1, Value v2);// * apply3

      /** Take the given parameters 'v0' and 'v1' and return a
          {@link Native} which takes 'v2' and produces a result.  */
      public Native apply2(final Value v0, final Value v1)
       { return new Native() // i.e., is a Function
          { public Value apply(Value v2) { return apply3(v0, v1, v2); } };
       }

      /** Take the given parameter 'v0' and return a {@link Native2} which
          takes parameters 'v1' and 'v2' and produces a result.  */
      public Native2 apply(final Value v0)
       { return new Native2() // NB. a Native2 is a Native is a Function
          { public Value apply2(final Value v1, final Value v2)
             { return apply3(v0, v1, v2); }
       }; }
    }//Function.Native3 class
   // -------------------------------------------------------------------------

   /** Return a Function, of one parameter, based on the unary operator
       'op' (not, -, hd, etc.). For example, (fromUop(-))x&nbsp;=&nbsp;-x.
       (Note that fromUop is a <TT>static</TT> method.)
       Also see {@link Function#uOp}, {@link Value#uOp},
       and {@link #fromBop}. */
   public static Function.Native fromUop(final int op)           // fromUop(op)
    { return new Function.Native()
       { public Value apply(Value v) { return v.uOp(op); }
    }; }//fromUop(op)

   /** Return a {@link Native2 Curried} Function based on the binary
       operator, 'op' (+, *, etc.).  For example,
       ((fromBop(+))x)y&nbsp;=&nbsp;x+y.
       (Note that fromBop is a <TT>static</TT> method.)
       Also see {@link Function#bOp}, {@link Value#bOp},
       and {@link #fromUop}. */
   public static Function.Native2 fromBop(final int op)          // fromBop(op)
    { return new Function.Native2()
       { public Value apply2(Value v1, Value v2) { return v1.bOp(op, v2); }
    }; }//fromBop(op)
   // -------------------------------------------------------------------------

   /** main() allows class Function to be <I>very slightly</I> tested,
       in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Function --");

      Value a=new Int(100), b=new Int(20), c=new Int(3);

      System.out.println( "  fromUop(-) 100 = "
         + fromUop(Lexical.minus).apply(a)
         + ", fromBop(+) " + a + " " + b + " = "
              + fromBop(Lexical.plus).apply(a).apply(b)
         );

      Native3 s3 = new Native3()  // s3 :: Int -> Int -> Int
       { public Int apply3(Value a, Value b, Value c)
          { return new Int(a.n() + b.n() + c.n()); }
       };
      System.out.println( "  on Native3: "
         + s3.apply3(a,b,c) + " = "         // all ...
         + s3.apply(a).apply2(b,c) + " = "  // partitions ...
         + s3.apply2(a,b).apply(c) + " = "  // of a,b,c ...
         + s3.apply(a).apply(b).apply(c)
         + "  -- s3(a,b,c), ..., s3 a b c"
         );

      final Cts2Cts
         sin = new Cts2Cts()
          { public double apply_x(double x) { return Math.sin(x); } };
      Cts2Cts
         cos = new Cts2Cts()
          { public double apply_x(double x) { return Math.cos(x); }
            // If the following closed-form is out,
            // it defaults to finite differences...
            // public Cts2Cts make_d_dx() { return sin.uOp(Lexical.minus); }
          };
      Cts2Cts d_cos  = cos.d_dx(),
              d2_cos = d_cos.d_dx();
      Integral I_cos = cos.integral();

      System.out.println( "  on cos (with its closed-form d/dx off):"
         +  "\n    cos 0 = " + cos.apply(zeroR)
         + ",\n    (d/dx cos) 0 = " + d_cos.apply(zeroR)
         + ",\n    (d2/dx2 cos) 0 = " + d2_cos.apply(zeroR)
         + ",\n    I cos steps=10 0..pi/2 = "
              + I_cos.apply3(ten, zeroR, PIby2)
         +  ",\n    d/dx (I cos steps=10 0..) (pi/2) = "
                    +  I_cos.apply2(ten, zeroR).d_dx().apply(PIby2)
         + ",\n    cos.root(guess=1) = " + cos.root(1)
         + ",\n    cos.turning(guess=1) = " + cos.turning(1)
         );

      System.out.println("-- done Function --");
    }//main

 }//Function class

// (c) Lloyd Allison 2012, Fri 25 Aug 2017. Licence to 'Function' is Gnu GPL.
