//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;

import java.io.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A stateful Lexical analyser, also see {@link Syntax}. */
public class Lexical
 { protected final InputStream inp;

   /** Variable components of the "state" of the Lexical analyser. */
   private byte[]  buffer = new byte[1];
   private int    lineNum = 0, charNum = 0; // i.e., 1st line is #0
   private int         sy = -1;       // see word, ... below
   private boolean lines1stSy = true; // is sy first on line?
   private char     rawCh = ' ',
                       ch = ' ';      // where {nl, tab} -> ' '
   private boolean    eof = false;
   private String theWord = "<?>";
   private int     theInt = 666;
   private double  theDbl = 6.9, theAcc = 0.1;  // i.e., 6.9+/-0.05

   /** Integer codes for the various lexical symbols;
       also see {@link #Symbol}. */
   public static final int  // ...
   word          =  0,
   numeral       =  1,
   numeralR      =  2,
   charLiteral   =  3 /* 'z'   */,
   stringLiteral =  4 /* "abc" */,
   triv          =  5 /* ()    */,
   nilSy         =  6,
   trueSy        =  7,
   falseSy       =  8,
   open          =  9,
   close         = 10,
   sqopen        = 11,
   sqclose       = 12,
   curlopen      = 13,
   curlclose     = 14,
   letSy         = 15,
   recSy         = 16,
   inSy          = 17,
   comma         = 18,
   colon         = 19,
   ifSy          = 20,
   thenSy        = 21,
   elseSy        = 22,
   lambdaSy      = 23,
   dot           = 24,
   consSy        = 25,
   orSy          = 26,
   andSy         = 27,
   eq            = 28,
   ne            = 29,
   lt            = 30,
   le            = 31,
   gt            = 32,
   ge            = 33,
   plus          = 34,
   minus         = 35,
   times         = 36,
   over          = 37,
   nullSy        = 38,
   hdSy          = 39,
   tlSy          = 40,
   notSy         = 41,
   eofSy         = 42;

   /** String representations of the various lexical symbols. */
   public static final String[] Symbol = new String[]
    { "<word>", "<numeral>", "<numeralR>", "<char>", "<string>",
      "()", "nil", "true", "false",
      "(", ")", "[", "]", "{", "}",
      "let", "rec", "in", ",", ":",
      "if", "then", "else", "lambda", ".",
      "::", "or", "and",
      "=", "<>", "<", "<=", ">", ">=",
      "+", "-", "*", "/",
      "null", "hd", "tl", "not",
      "<eof>"
    };//Symbol

   /** Return the n-th Lexical Symbol as a String, if possible. */
   public static String int2sy(int n)
    { return n >= 0 && n <= eofSy ? Symbol[n] : "???"; }
   // -------------------------------------------------------------------------

   /** Construct a Lexical analyser of a source InputStream, 'inp'.
       Also see {@link #forString} and {@link #forFile(File)}.  */
   public Lexical(InputStream inp)
    { this.inp = inp; insymbol(); }//constructor


   /** Return a Lexical analyser of a source String, 'str'. */
   public static Lexical forString(String str)
    { return new Lexical(new ByteArrayInputStream(str.getBytes())); }

   /** Return a Lexical analyser of a source File, 'file'.
       Also see {@link #forFile(String)}. */
   public static Lexical forFile(File file)
    { InputStream fip = null;
      try{ fip = new FileInputStream(file); }
      catch(Exception e)
       { throw new RuntimeException( "cannot find file \'" + file + "\'" ); }
      return new Lexical(fip);
    }//forFile(f)

   /** Return a Lexical analyser of a source File named 'fn'.
       Also see {@link #forFile(File)}. */
   public static Lexical forFile(String fn)
    { return forFile(new File(fn)); }


   /** Return the kind of the current input symbol. */
   public int sy() { return sy; }

   /** Return a String representation of the current {@link #sy() symbol},
       e.g., for debugging or tracing purposes. */
   public String syInfo()
    { if( sy == word || sy == charLiteral || sy == stringLiteral)
         return theWord;
      if( sy == numeral  ) return "" + theInt;
      if( sy == numeralR ) return "" + theDbl;
      /* else */ return sy + "(" + int2sy(sy) + ")";
    }//syInfo()

   /** Is the current input symbol, the first one on the line? */
   public boolean lines1stSy() { return lines1stSy; }

   /** Is this Lexical at its 'end of input', or not? */
   public boolean eoi() { return sy == eofSy; }

   /** Return the current word (identifier or keyword), if any. */
   public String theWord() { return theWord; }

   /** Return the current integer numeral, if any. */
   public int theInt() { return theInt;  }

   /** Return the current real-valued numeral, if any. */
   public double theDbl() { return theDbl;  }

   /** Return the supposed precision of the real-valued numeral, if any. */
   public double theAcc() { return theAcc;  }


   /** Get the next {@link #sy() symbol}, from the InputStream 'inp'. */
   public void insymbol()                                           // insymbol
    { if( sy == eofSy ) return; // else...
      lines1stSy = false;  // ?1st sym on line? see getch()
      do
       { while( ch == ' ' ) getch(); // skip white space
         if( ch == '{' )             // skip comment
          { do{ getch(); }while( ch != '}' && ! eof );
            getch();
          }
       }while( ch == ' ' || ch == '{' );

      if( eof ) sy = eofSy;

      else if( Character.isLetter(ch) )               // words
       { StringBuffer w = new StringBuffer();
         while( Character.isLetterOrDigit(ch) )
          { w.append(ch); getch(); }
         theWord = w.toString();  sy = word;
         for(int i = 0; i < eofSy; i++) // maybe speed up one day?
          { if(Symbol[i].compareTo(theWord) == 0) { sy = i; break; } }
       }

      else if( Character.isDigit(ch) )                // numbers
       { theInt = 0; theDbl = 0.0; theAcc = 0.0;
         StringBuffer n = new StringBuffer();
         while( Character.isDigit(ch) )
          { n.append(ch); getch(); }      // 123
         theInt = Integer.parseInt(n.toString());
         sy = numeral;
         if( ch != '.' ) return;          // i.e., int
         n.append(ch); getch();           // '.'
         StringBuffer acc = new StringBuffer(); // accuracy
         while( Character.isDigit(ch) )  // .456
          { n.append(ch);
            acc.append( acc.length()==0 ? "0." : "0" );
            getch();
          }
         acc.append( acc.length() == 0 ? "1.0" : "1" );
         int ePos = n.length();
         if( ch == 'e' || ch == 'E' )     // E
          { n.append(ch); getch();
            if( ch == '+' || ch == '-' )  // +|-
             { n.append(ch); getch(); }
            while( Character.isDigit(ch) )
             { n.append(ch); getch(); }   // 45
          }
         sy = numeralR;                   // i.e., real
         try{ theDbl = Double.parseDouble(n.toString());
              theAcc = Double.parseDouble(
                          acc.toString() + n.toString().substring(ePos));
              return ;
            }
         catch(Exception e) { error("bad real number"); }
       }

      else if( ch == '\'' )                           // character literals
       { getch();
         theWord = new String(new char[]{ch});  getch();
         if( ch == '\'' ) { sy = charLiteral; getch(); }
         else error("bad char literal");
       }

      else if( ch == '\"' )                           // string literals
       { getch(); // skip opening '\"'
         StringBuffer s = new StringBuffer();
         while( ch != '\"' )
          { if( rawCh == '\n' /* eoln */ ) error("no closing \" to string");
            s.append(ch); getch();
          }
         getch(); // skip closing '\"'
         theWord = s.toString(); sy = stringLiteral;
       }

      else switch( ch )                               // special symbols
       { case '<': getch();
                   if( ch == '=' ) { getch(); sy = le; break; }
                   if( ch == '>' ) { getch(); sy = ne; break; }
                   sy = lt; break;
         case '>': getch();
                   if( ch == '=' ) { getch(); sy = ge; break; }
                   sy = gt; break;
         case '(': getch();
                   if( ch == ')' ) { getch(); sy = triv; break; }
                   sy = open; break;
         case ':': getch();
                   if( ch == ':' ) { getch(); sy = consSy; break; }
                   sy = colon; break;

         case '+': case '-': case '*': case '/': case '=':
         case ')': case '.': case ',':
         case '[': case ']': case '{': case '}': case '\"':
            int ch2 = ch; getch();
            switch( ch2 )
             { case '+': sy = plus;    break;  case '-': sy = minus;    break;
               case '*': sy = times;   break;  case '/': sy = over;     break;
               case '=': sy = eq;      break;  case ')': sy = close;    break;
               case '.': sy = dot;     break;  case ',': sy = comma;    break;
               case '[': sy = sqopen;  break;  case ']': sy = sqclose;  break;
               case '{': sy =curlopen; break;  case '}': sy =curlclose; break;
               default: error("should not be here!"); break;
             }
            break;

         default: error("bad symbol");
       }
    }//insymbol()

   /** Note, getch() changes the state variables,
       including 'ch', as a side-effect. */
   protected void getch()                                            // getch()
    { if( /* last */ rawCh == '\n' )
       { lineNum ++ ; charNum = -1;
         lines1stSy = true; // also see insymbol()
       }
      rawCh = '?'; ch = rawCh;
      if( eof ) return;
      try { int n = 0;
            if( inp.available() > 0 ) n = inp.read(buffer);
            if( n <= 0 ) eof = true;
            else { ch = (char)buffer[0]; rawCh = ch; charNum ++ ; }
          }
      catch(Exception e){ }
      if( ch == '\n' || ch == '\t' ) ch = ' '; // NB. but not rawCh
    }//getch()


   /** {@link #skipRest(int,int) Skip}, and return, up to
       a "few" remaining input lines and characters. */
   public String skipRest()                                       // skipRest()
    { return skipRest(5 /* lines */, 400 /* chars */); }

   /** Skip, and return, the rest of the input up to the limits 'maxLines'
       lines and 'maxChars' characters. Note, returns <I>all</I>
       characters if maxLines and maxChars are both negative. */
   public String skipRest(int maxLines, int maxChars)            // skipRest(,)
    { if( inp == null ) return " No input"; // (e.g., file error)
      int line1 = lineNum;
      StringBuffer sb = new StringBuffer();
      int n = 0;
      while( ! eof )
       { if( n == maxChars || lineNum-line1 == maxLines )
          { sb.append(" ...etc.\n"); break; }
         sb.append(rawCh);  n++;  getch();
       }
      return sb.toString();
   }//skipRest(nl,nc)

   /** Print an error message, msg, and stop. */
   public void error(String msg)                                       // error
    { System.out.println( "Error: " + msg
                   + ", sy=" + sy + " (" + int2sy(sy) + "), ch=" + ch
                   + ", theWord=" + theWord + ", theInt=" + theInt
                   + ", theDbl=" + theDbl + " to " + theAcc
                   + "\n(line " + lineNum + "," + charNum + ")skipped:"
                   + skipRest()
                   + "-- Stop --" );
      System.exit(1);  //nonzero indicates abnormal termination
    }//error(msg)
   // -------------------------------------------------------------------------

   /** main() allows Lexical to be tested, a little, in isolation,
       for example <NOBR> java ...Lexical &lt; inputfile </NOBR> */
   public static void  main(String[] argv)
    { System.out.println("-- Lexical --");
      for(int i=0; i < argv.length; i++) // command line params if any
         System.out.print("argv[" + i + "]=" + argv[i] + "\n");

      Lexical lex = argv.length <= 0 ? new Lexical(System.in) // std input
                  : Lexical.forFile(new File(argv[0]));       // named file
      while( ! lex.eoi() )
       { int sy = lex.sy(); String sy_str = "      " + sy;
         System.out.print( sy_str.substring(sy_str.length()-4)
                         + " " + int2sy(sy) + " " );
         switch(sy)
          { case Lexical.word: case charLiteral:
               System.out.print( "<" + lex.theWord() + ">" ); break;
            case Lexical.numeral:
               System.out.print( "<" + lex.theInt() + ">" ); break;
            case Lexical.numeralR:
               System.out.print("<"+lex.theDbl()+" to "+lex.theAcc+">"); break;
          }
         System.out.println();
         lex.insymbol();
       }

      System.out.println("-- end Lexical --");
    }//main

 }//Lexical class

// Lambda, (c) Lloyd Allison, October 2007, Dec 2013. Licence is GNU GPL.
