//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The standard library of predefined {@link Function Functions} and
    other {@link Value Values}; also see {@link Environment}. */
public class Library
 {
   /** The identity Function, <NOBR><TT> id x = x </TT>.</NOBR> */
   public static final Function id =                                    // id x
      new Function.Native()
       { /** Note, &quot;{@link Value#force forces}&quot; 'x' to WHNF;
             see {@link Function.Native#apply}.  */
         public Value apply(Value x)
          { return x.force(); }//apply(x)
       };//id x

   // /** A different way to define the identity function,
   //     from (a String for) the Lambda Expression.  */ 
   // public static final Function id =                             // other id
   //    new Value.Lambda( (Expression.LambdaExp)
   //           new Syntax(Lexical.forString( "lambda p.p" )).exp(),
   //           Environment.empty );


   /** Return the first element of a {@link Value.Tuple}. */
   public static final Function fst =                                    // fst
      new Function.Native()           // fst
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .fst(); } 
       };//fst

   /** Return the second element of a {@link Value.Tuple}. */
   public static final Function snd =                                    // snd
      new Function.Native()
       { public Value apply(Value p) 
          { return ((Value.Tuple) p.force()) .snd(); }
       };//snd

   /** Return the third element of a {@link Value.Tuple}. */
   public static final Function thrd =                                  // thrd
      new Function.Native()
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .thrd(); }
       };//thrd

   /** Return the fourth element of a {@link Value.Tuple}. */
   public static final Function frth =                                  // frth
      new Function.Native()
       { public Value apply(Value p)
          { return ((Value.Tuple) p.force()) .frth(); }
       };//frth


   /** Return a Function based on the unary operator,
       'op' (not, -, hd, etc.). Also see
       {@link Value#uOp} and {@link #bOp}. */
   public static Function.Native uOp(final int op)                   // uOp(op)
    { return new Function.Native()
       { public Value apply(Value v) { return v.uOp(op); }
    }; }//uOp(op)

   /** Return a curried Function based on the binary operator, 'op'
       (+, *, etc).  Also see {@link Value#bOp} and {@link #uOp}. */
   public static Function.Native2 bOp(final int op)                  // bOp(op)
    { return new Function.Native2()
       { public Value apply2(Value v1, Value v2) { return v1.bOp(op, v2); }
    }; }//bOp(op)


   /** curry: <TT>((t, u) &rarr; v) &rarr; (t &rarr; u &rarr; v)</TT>.
       <NOBR><TT>curry fu x y = fu(x, y)</TT>.</NOBR>  */
   public static final Function.Native3 curry =                        // curry
      new Function.Native3()
       { public Value apply3(Value f, Value p0, Value p1)
          { return ((Function) f.force()) .apply(Value.pair(p0, p1));
          }//apply3(f,p0,p1)
       };//curry

   /** uncurry: <TT>(t &rarr; u &rarr; v) &rarr; ((t, u) &rarr; v)</TT>.
       <NOBR><TT>uncurry fc (x, y) = fc x y</TT>.</NOBR> */
   public static final Function.Native2 uncurry =                  // uncurry f
      new Function.Native2()
       { public Value apply2(Value f, Value p)
          { if( ! Type.PAIR.ni(p) ) error( "uncurry(f,p), p:" + p.type() );
            final Value.Tuple p0p1 = (Value.Tuple) p.force();  // WHNF
            final Value p0 = p0p1.elt(0), p1 = p0p1.elt(1);
            return ((Function) // NB. all apply(.)s return Values in WHNF
                      ((Function) f.force()) .apply(p0)) .apply(p1);
          }//apply2(f,p)
       };//uncurry


   /** <TT>whnf p0 p1</TT> can be used to make a parameter (p0) strict, say:
       {@link Value#force force} p0 to weak head normal form (WHNF), but
       return p1 (also in WHNF). */
   public static final Function whnf =                            // whnf p0 p1
      new Function.Native2()
       { public Value apply2(Value p0, Value p1)
          { p0.force(); return p1.force(); }
       };//whnf p0 p1
   // -------------------------------------------------------------------------

   private final static String [] ids  = new String[]                    // ids
      { "id", "fst", "snd", "thrd", "frth", "whnf", "curry", "uncurry" };

   private final static Value  [] vals = new Value []                   // vals
      {  id,   fst,   snd,   thrd,   frth,   whnf,   curry,   uncurry  };

   /** An Environment binding names (Strings) to standard Functions,
       for example, "id"&rarr;{@link #id id},
       "fst"&rarr;{@link #fst fst}, etc.  */
   public final static Environment env =                                 // env
      Environment.empty.bind(ids, vals);
   // -------------------------------------------------------------------------

   /** Run a few simple tests on Library. */
   public static void main(String[] argv)
    { System.out.println("-- test Library --");

      Value a=new Value.Int(10), b=new Value.Int(2);
      Value.Tuple ab = Value.pair(a, b);

      Function.Native u2 = new Function.Native() // u2 :: (Int, Int) -> Int,
       { public Value.Int apply(Value ab)        // i.e., uncurried
          { Value.Tuple abT = (Value.Tuple) ab.force();
            return new Value.Int(abT.fst().n() + abT.snd().n());
       }  };

      Function.Native2 c2 = curry.apply(u2);     // curried

      System.out.println( "  currying: "
                        + u2.apply(ab)  + " = "
                        + c2.apply(a).apply(b) + " = "
                        + c2.apply2(a, b) + " = "
                        + uncurry.apply(c2).apply(ab)
                        );

      System.out.println( "  uOp, bOp: "
                        + "(-) " + a + " = " + uOp(Lexical.minus).apply(a)
                        + ", (+) " + a + " " + b + " = "
                           + bOp(Lexical.plus).apply(a).apply(b)
                        );
 
      System.out.println("-- done Library --");
    }//main

 }//Library class

// (c) Lloyd Allison 2012, 23 Nov 2016. Licence to 'Library' is Gnu GPL.
