//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The Types of {@link Value Values};
    it is for a very simple dynamic type system;
    {@link Type.Enum} and {@link Type.Option} are interesting. */
public abstract class Type extends Value.Option                       // * Type
 {
   /** Integer codes for various "types" of Type. */
   public static final int TRIV_N   = 0,
                           BOOL_N   = 1,
                           CHAR_N   = 2,
                           INT_N    = 3,
                           CTS_N    = 4,
                           ENUM_N   = 5,
                           OPTION_N = 6,
                           VECTOR_N = 7,
                           TUPLE_N  = 8,
                           FUNCTION_N = 9,
                           MODEL_N = 10,
                           TYPE_N  = 11;

   /** The (optional) name of 'this' Type.  */
   protected final String name;

   /** Special case when ids={} and arities={}. */
   public Type(String name)
    { this.name = name; }//constructor

   /** Return {@link #TYPE TYPE}. */
   public Type type() { return TYPE; }

   /** This default returns 0; override it if necessary. */
   public int nElts() { return 0; }

   /** This default throws an error, inapplicable. */
   public Value elt(int i)
    { throw RTE( ".elt(" + i + ") is inapplicable" ); }

   /** Return a String representation of 'this' Type. */
   public String toString()                                       // toString()
    { final boolean some = nElts() > 0;
      StringBuilder sb = new StringBuilder( (some ? "(" : "")
            + (name == null || name.isEmpty()
               ? getClass().getName() : name) );
      for( int i = 0; i < nElts(); i ++ )
       { sb.append( " " );  Value ei;
         try{ ei = elt(i); }                // may be ...
         catch( Throwable t ) {ei = null; } // ... unknown
         sb.append( ei == null ? "<?>" : ei.toString() );
       }
      if( some ) sb.append( ")" );
      return sb.toString();
    }//toString()

   /** Also see {@link Expression#appendSB}. */
   public void  appendSB(StringBuffer sb)                       // appendSB(sb)
    { sb.append(toString()); }

   /** &iquest;Is Value v's <B>in</B> this Type; does Value 'v' have
       exactly 'this' Type? (--???or_Inclusion???) */
   public boolean ni(Value v)                                          // ni(v)
    { return v.type() == this; }

   /** Does 'this' Type contain (or equal) Type 't'?  */
   public boolean contains(Type t)
    { return t == this; }
   // -------------------------------------------------------------------------

   /** The Type of a Type is TYPE. TYPE also has Type TYPE.
       See class {@link Type.TYPE}.  */
   public final static Type TYPE = new TYPE();                          // TYPE

   /** DNA, an example {@link Type.Enum} Type. */
   public final static Type.Enum DNA  =                                  // DNA
      new Enum("DNA", new String[]{"A", "C", "G", "T"});

   /** Also see {@link Value.Triv}.  */
   public final static Triv TRIV = new Triv("TRIV");                    // TRIV

   /** Also see {@link Value.Bool}.  */
   public final static Enum BOOL =                                      // BOOL
      new Enum("BOOL", new String[]{ Lexical.int2sy(Lexical.falseSy),
                                     Lexical.int2sy(Lexical.trueSy ) })
       { public int n() { return BOOL_N; }
         public Value.Bool int2value(int n)
          { switch(n)
             { case 0:  return Value.ffalse;
               case 1:  return Value.ttrue;
               default: throw RTE(" n=" + n);
       }  }  };

   /** Also see {@link Value.Int}.  */
   public final static Int INT  = new Int("INT");                        // INT

   /** Also see {@link Value.Char}.  */
   public final static Char CHAR = new Char("CHAR");                    // CHAR

   /** Also see {@link Value.Chars}.  */
   public final static Type CHARS = new Vector("STRING", CHAR);        // CHARS

   /** Also see {@link Value.Cts}.  */
   public final static Type CTS  = new Cts("CTS");                       // CTS

   public final static Type VECTOR = new Vector();                    // VECTOR

   public final static Type.Vector VECTOR_CTS =                   // VECTOR_CTS
      new Vector(CTS);
   public final static Type MATRIX_CTS = new Vector(VECTOR_CTS);  // MATRIX_CTS

   public final static Type.Vector VECTOR_INT =                   // VECTOR_INT
      new Vector(INT);
   public final static Type MATRIX_INT = new Vector(VECTOR_INT);  // MATRIX_INT

   public final static Type.Vector VECTOR_CHARS =               // VECTOR_CHARS
      new Vector(CHARS);

   /** Also see {@link Value.Tuple}. */
   public final static Type.Tuple PAIR   = new Tuple("PAIR",   2),      // PAIR
                                  TRIPLE = new Tuple("TRIPLE", 3);    // TRIPLE
   private static Tuple [] ts = null;
   /** [null, null, PAIR, TRIPLE, 4_Tuple, ...]. */
   public static Tuple tuple(int k)                                 // tuple(k)
    { if( ts == null ) // cache Tuple 0, ..., 21.
       { ts = new Tuple[21];
         ts[0] = null; ts[1] = null; ts[2] = PAIR; ts[3] = TRIPLE;
         for(int i = 4; i < ts.length; i ++)
            ts[i] = new Tuple(i+"_Tuple", i);
       }
      return ts[k];
    }//tuple(k)

   /** Also see {@link Value.List}.  */
   public final static Type.Option LIST =                               // LIST
      new Option("LIST",
          new String []{ Lexical.Symbol[Lexical.nilSy],   // ids[]
                         Lexical.Symbol[Lexical.consSy] },
          new int   [] { 0, 2 }  // arities
      );//new Option(...)

   /** {@link Value.Maybe Maybe} T, for optional Values, missing data, etc. */
   public final static Type.Option MAYBE =                             // MAYBE
      new Option("MAYBE",
          new String []{ "None", "Just" }, // ids[]
          new int []   {  0,      1 }      //arities
      );//new Option(...)

   /** Inclusive Or, as in {@link Value.Inc_Or Inc_Or} Int Char, say. */
   public final static Type.Option INC_OR =                           // INC_OR
      new Option("INC_OR",
          new String [] { "Left", "Right", "Both" },  // ids[]
          new int []    {  1,      1,       2 }       // arities
      );//new Option(...)

   /** Also see {@link Function}.  */
   public final static Type FUNCTION = new Function();              // FUNCTION

   /** For statistical {@link mml.Model Models} and {@link mml.README mml}. */
   public final static Type MODEL= new Model();                        // MODEL
   // -------------------------------------------------------------------------

   /** The class of the Type of a Type; see {@link #TYPE TYPE}. */
   public static class TYPE extends Type
    { public TYPE() { super("TYPE"); }//constructor
      public int n() { return TYPE_N; }
    }//TYPE class

   /** The superclass of Atomic Types, notably of {@link Discrete Discrete}
       and {@link #CTS CTS}. By default, such a Type is
       {@link #ordered() ordered}. Also see {@link Value.Atomic}.  */
   public abstract static class Atomic extends Type                   // Atomic
    { public Atomic(String name) { super(name); }

      /** This default return true; i.e., the Type is assumed to be ordered. */
      public boolean ordered() { return true; }
    }//Atomic class
   // -------------------------------------------------------------------------

   /** The class of the {@link #CTS CTS} Type.  A subclass is a possibility
       for a bounded subset of Cts. Also see {@link Value.Cts}. */
   public static class Cts extends Atomic                                // Cts
    { public Cts(String name) { super(name); }//constructor
      public int n() { return CTS_N; }
    }//Cts class


   /** The superclass of Discrete Types. */
   public static abstract class Discrete extends Atomic           // * Discrete
    { private final boolean hasLwb, hasUpb;
      private final int     lwb,    upb;
      private final boolean ordered;

      /** Note, this is assumed to be un-{@link #ordered() ordered}. */
      public Discrete(String name, boolean hasLwb, int lwb,
                                   boolean hasUpb, int upb)
       { this(name, hasLwb, lwb, hasUpb, upb, /*ord*/ false); }//constructor

      /** Construct a Discrete Type, which may or may not
          {@link #hasLwb have}   a  {@link #lwb lower} bound,
          {@link #hasUpb and/or} an {@link #upb upper} bound, and be
          {@link #ordered() ordered} or not. */
      public Discrete(String name, boolean hasLwb, int lwb,
                                   boolean hasUpb, int upb,
                                   boolean ordered)
       { super(name);
         this.hasLwb = hasLwb; this.lwb = lwb;
         this.hasUpb = hasUpb; this.upb = upb;
         this.ordered = ordered;
       }//constructor

      /** Convert an int, n, into a Value of 'this' Discrete Type.
          Also see {@link #string2n string2n} and
          {@link #string2value string2value}.  */
      public abstract Value.Discrete int2value(int n);        // * int2value(n)

      /** Return the int "code" for a constant denoted by String 's'.
          Also see {@link #string2value string2value} and
          {@link #int2value int2value}.  */
      public abstract int string2n(String s);                  // * string2n(s)

      /** Return
          {@link #int2value int2value}({@link #string2n string2n}(s)).  */
      public Value.Discrete string2value(String s)          //  string2value(s)
       { return int2value(string2n(s)); }

      /** Does 'this' Discrete have both {@link #hasLwb lower}
          and {@link #hasUpb upper} bounds? */
      public boolean bounded() { return hasLwb() && hasUpb(); }

      /** Does 'this' Discrete have a {@link #lwb}? */
      public boolean hasLwb() { return hasLwb; }

      /** Does 'this' Discrete have a {@link #upb}? */
      public boolean hasUpb() { return hasUpb; }

      /** 'this' Discrete's &lang;{@link #lwb() lwb}, {@link #upb() upb}&rang;
          if appropriate, else exception. */
      public Value.Tuple bounds() { return Value.pair(lwb(), upb()); }

      /** 'this' Discrete's lower bound if any otherwise an Exception;
           also see {@link #lwb_n}. */
      public Value lwb() { return int2value(lwb_n()); }

      /** 'this' Discrete's upper bound if any otherwise an Exception;
           also see {@link #upb_n}. */
      public Value upb() { return int2value(upb_n()); }

      /** 'this' Discrete's <B>int</B> lower bound - if any,
          else exception, alse see {@link #lwb}. */
      public int lwb_n()
       { if( hasLwb ) return lwb;
         /* else */ throw RTE(".lwb_n()");
       }//lwb_n()

      /** 'this' Discrete's <B>int</B> upper bound - if any,
           else exception, alse see {@link #upb}. */
      public int upb_n()
       { if( hasUpb ) return upb;
         /* else */ throw RTE(".upb_n()");
       }//upb_n()

      /** Is 'this' Type ordered, or not? */
      public boolean ordered() { return ordered; }

      /** Return <TT>not&nbsp;{@link #ordered() ordered()}</TT>. */
      public boolean unOrdered() { return ! ordered(); }

    }//Discrete class


   /** The class of the {@link #TRIV TRIV} Type.
       Also see {@link Value#triv}. */
   public static class Triv extends Enum                                // Triv
    { public Triv(String name)
       { super(name, new String[]{Lexical.int2sy(Lexical.triv)}); }
      public int n() { return TRIV_N; }
      public Value.Triv int2value(int n)
       { switch(n)
          { case 0:  return Value.triv;
            default: throw RTE( ".int2value(" + n + ")" );
       }  }//int2value
    }//Triv class


   /** The class of the {@link #INT INT} Type. Also see {@link Value.Int}. */
   public static class Int extends Discrete                              // Int
    { /** Construct "the" Int Type. */
      public Int(String name)
       { this(name, /*lwb*/ false, -1, /*upb*/ false, -1); }//constructor

      /** Construct an Int Type possibly bounded above, or below, or both. */
      public Int(String name, boolean hasLwb, int lwb,
                              boolean hasUpb, int upb)
       { super(name, hasLwb, lwb, hasUpb, upb, /*ord*/ true); }//constructor

      /** Return {@link #INT_N INT_N}. */
      public int n() { return INT_N; }

      /** Return (a new) {@link Value.Int Value.Int}(n). */
      public Value.Int int2value(int n)
       { if( (hasLwb() && n < lwb_n())
          || (hasUpb() && n > upb_n()) ) error( ".int2value(" + n + ")" );
       /*else*/ return new Value.Int(n);  // ??? Type
       }//int2value(n)

      /** Return an int from a String such as "123", "-456" or even
          "&nbsp;+&nbsp;789".  Note, the use of {@link Lexical} to
          parse&nbsp;s permits FP comments in it. */
      public int string2n(String s)
       { final Lexical lex = Lexical.forString(s);
         final int sy = lex.sy();
         final int sign = sy == Lexical.minus ? -1 : 1;
         if( sign < 0 || sy == Lexical.plus ) lex.insymbol();  // [+|-]
         if( lex.sy() == Lexical.numeral )
            return sign * lex.theInt();
         else throw RTE( ".string2n(\"" + s +  "\") bad int value" );
       }//string2n(s)

    }//Int Class


   /** The class of the {@link #CHAR CHAR} Type.
       Also see {@link Value.Char}. */
   public static class Char extends Discrete                            // Char
    { public Char(String name)
       { super("CHAR", /*lwb*/ true, Character.MIN_VALUE,
                       /*upb*/ true, Character.MAX_VALUE,
                       /*ordered*/ true);
       }//constructor
      public int n() { return CHAR_N; }
      public Value.Char int2value(int n)
       { return new Value.Char((char) n); }
      /** String s must be a single character; return it's int code. */
      public int string2n(String s)
       { if( s.length() != 1 ) error( ".string2n(\"" + s + "\")" );
         return (int) s.charAt(0);
       }//string2n(s)
    }//Char class


   /** Enum Types, e.g., DNA={A,C,G,T}, types of {@link Value.Enum}.
       An Enum Value, d, is specified by an int code, d.n(), and
       its Enum Type, d.type(), tells how to interpret d.n(), etc..
       Also see {@link Option}.  */
   public static class Enum extends Discrete                            // Enum
    { /** String <TT> ids[i] </TT> denotes Value
          <NOBR><TT> vals[i], i=0..&nbsp; </TT>.</NOBR> */
      public final String     [] ids;

      /** String <TT> ids[i] </TT> denotes Value
          <NOBR><TT> vals[i], i=0..&nbsp; </TT>.</NOBR> */
      private final Value.Enum [] vals;  // NB. BOOL bypasses vals[]

      /** An Enum Type created thus is assumed to be
          <NOBR><B>un</B>-{@link #ordered() ordered}</NOBR>. */
      public Enum(String name, String[] ids)
       { this(name, ids, false); }//constructor

      /** E.g., <NOBR>DNA = A | C | G | T</NOBR>.  DNA is unordered,
          and <NOBR>quality = poor | avg | good</NOBR>, ordered. */
      public Enum(String name, String[] ids, boolean ordered)
       { super(name, true, 0, true, ids.length-1, ordered);
         this.ids = ids; vals = new Value.Enum[ids.length];
         for( int i = 0; i < ids.length; i ++ )
            vals[i] = new Value.Enum.GP(this, i);
       }//constructor

      public int n() { return ENUM_N; }

      /** Return the n-th Value, vals[n], of 'this' Enum. */
      public Value.Enum int2value(int n) { return vals[n]; }
      // Note that BOOL overrides this.

      // Just to get the return type Value.Enum, not Value.Discrete ...
      public Value.Enum string2value(String s)
       { return int2value(string2n(s)); }

      /** Return the int code corresponding to 'str'.
          Throw an {@link #RTE RTE} in case of error. */
      public int string2n(String str)
       { for( int i = 0; i < ids.length; i ++ )
            if( str.compareTo(ids[i]) == 0 ) return i;
         throw RTE( ".string2n(" + str + ") does not match" );
       }//string2n(str)

    }//Enum class
  // -------------------------------------------------------------------------

   /** Superclass of {@link Type.Vector} etc; also see {@link Atomic}.  */
   public static abstract class Structured extends Type           // Structured
    { public Structured(String name) { super(name); }
    }

   public static class Vector extends Structured                      // Vector
    { /** The element Type of a Vector (if known). */
      public final Type eltType;

      public Vector() { this(null); }

      public Vector(Type eltType)
       { this("VECTOR", eltType); }

      public Vector(String name, Type eltType)
       { super(name); this.eltType = eltType; }//constructor

      /** Is Value v's type in 'this' Vector Type?  */
      public boolean ni(Value v)
       { final Type vt = v.type();
         if( ! (vt instanceof Vector) ) return false;
         if( eltType == null ) return true; // I don't care
         return eltType == ((la.maths.Vector) v).eltType();
       }//ni(v)

      /** Does 'this' Vector Type contain Type&nbsp;'t'?  */
      public boolean contains(Type t)
       { if( t == this ) return true;
         if( ! (t instanceof Vector) ) return false;
         if( eltType == null ) return true;  // don't care about t's elements
         return( eltType.contains( ((Vector) t).eltType) );
       }//contains(t)

      public int n() { return VECTOR_N; }

      public int nElts() { return 1; }

      public Type elt(int i)
       { if( i < 0 || i >= nElts() )
            throw RTE( ".elt(" + i + ") inapplicable" ); // else ...
         return eltType;
       }
    }//Vector class

   /** The Types of k-Tuples; there is a different type for each value of k.
       Also see {@link #tuple}, {@link Value.Tuple}.  */
   public static class Tuple extends Structured                    // (k-)Tuple
    { /** The number of componets (fields) of 'this' Typle Type. */
      public final int k;

      public Tuple(String name, int k)
       { super(name); this.k = k; }//constructor

      public int  n()        { return TUPLE_N; }

      public int  nElts()    { return k; }
      /* elt(i) is nevertheless inapplicable, although see GP.elt(i). */

      /** Is Value v's Type in 'this' Tuple Type?  */
      public boolean ni(Value v)
       { final Type t = v.type();
         if( t == this ) return true;
         if( (! (t instanceof Tuple) ) || ((Tuple) t) .k != k ) return false;
         return true;
       }//ni(v)

      /** Does 'this' Tuple Type contain Type 't'?  */
      public boolean contains(Type t)
       { if( this == t ) return true;
         if( (! (t instanceof Tuple)) || (k != ((Tuple) t).k)  ) return false;
         return true;
       }

      /** A Tuple Type with specific field Types, {@link #elts}.
          (The lack of multiple inheritance means that
          this Type.Tuple cannot also be a Value.Tuple.)  */
      public static class GP extends Tuple                          // Tuple.GP
       { /** The component Types (fields) of 'this' Tuple. */
         public final Type [] elts;

         public GP(String name, Type [] elts)
          { super(name, elts.length); this.elts = elts; }//constructor

         /** Return the i-th {@link #elts component Type}. */
         public Type elt(int i) { return elts[i]; }

         /** Does Value v structurally match 'this' Type? */
         public boolean ni(Value v)
          { if( ! super.ni(v) ) return false;
            final Value.Tuple vt = (Value.Tuple) v;
            for( int i = 0; i < k; i ++ ) // each field...
               if( ! elt(i).ni( vt.elt(i) ) ) return false;
            return true;
          }//ni(v)

         /** Does 'this' Tuple(t1,t2,...) Type contain Type 't'? */
         public boolean contains(Type t)
          { if( t == this ) return true;
            if( ! (t instanceof Tuple.GP) ) return false;
            final Tuple.GP tT = (Tuple.GP) t;
            if( k != tT.k ) return false;
            for( int i = 0; i < k; i ++ ) // each field...
               if( ! elt(i).contains( tT.elt(i) ) ) return false;
            return true;
          }
       }//Tuple.GP class

    }//Tuple class

   /** Option Types, e.g.,
       <NOBR>{@link Value.List} : {@link #LIST},</NOBR> or <TT>Tree</TT> where
       <NOBR><TT> Tree e = emptyT | fork (Tree e) e (Tree e)</TT>.</NOBR>
       Also see {@link Type.Enum}.  */
   public static class Option extends Structured                      // Option
    { /** For example, {"emptyT", "fork"}.  */
      public final String [] ids;

      /** Special case when ids={} and arities={}. */
      public final int    [] arities;

      /** Special case when ids={} and arities={}. */
      public Option(String name)
       { this(name, new String[]{}, new int[]{}); }

      /** Construct a Type.Option with name 'name' etc. */
      public Option(String name, String[] ids, int[] arities)
       { super(name);
         this.ids = ids; this.arities = arities;
         if( ids.length != arities.length )
            error( ".constructor, lengths "+ids.length+"!="+arities.length );
       }//constructor

      public int n() { return OPTION_N; }
    }//Option class
   // -------------------------------------------------------------------------

   /** Function Types. */
   public static class Function extends Type                        // Function
    { /** Input- and output- Types. */
      private final Type ipType, opType;
      public Function() { this("FUNCTION"); }//constructor
      public Function(String name)
       { this(name, null, null); }//constructor
      public Function(String name, Type ipType, Type opType)
       { super(name); this.ipType = ipType; this.opType = opType;
       }//constructor
      public int n() { return FUNCTION_N; }
      /** Two elements (if known), the input and output Types. */
      public int nElts() { return 2; }
      /** elt(0) returns the input Type, elt(1) the output, if known! */
      public Value elt(int i)
       { switch(i)
          { case 0:  return ipType == null ? super.elt(i) : ipType;
            case 1:  return opType == null ? super.elt(i) : opType;
            default: return super.elt(i);
       }  }
    }//Function class

   /** Model Types. */
   public static class Model extends Type                              // Model
    { private final Type dataspace;
      public Model() { this("MODEL"); }//constructor
      public Model(String name) { this(name, null); }//constructor
      public Model(String name, Type dataspace)
       { super(name); this.dataspace = dataspace; }//constructor
      public int n() { return MODEL_N; }
      /** One element (if known), the dataspace (Type). */
      public int nElts() { return 1; }
      /** elt(0) returns the dataspace (Type), if known! */
      public Value elt(int i)
       { switch(i)
          { case 0:  return dataspace == null ? super.elt(i) : dataspace;
            default: return super.elt(i);
       }  }
    }//Model class
   // -------------------------------------------------------------------------

                                                  // a few very(!) simple tests
   /** Show a Value 'v' -- which must not be infinite, or even "big". */
   public static String show(Value v)
    { return v + ":" + v.type(); }

   /** {@link #show(Value)} several Values. */
   public static String show(Value[] vs)
    { String ans = "";
      for(int i = 0; i < vs.length; i ++ ) ans += show(vs[i]) + ", ";
      return ans;
    }

   /** 'main' allows Type to be tested (v.slightly) on its own
       e.g., <NOBR><TT>java Type</TT>.</NOBR>  */
   public static void main(String[] argv)                               // main
    { System.out.println("-- test Type --");

      System.out.println(show(new Value[]
         {Value.triv, DNA.vals[0], Value.ttrue,
          new Value.Int(8), Value.nilVal}));

      System.out.println("-- done Type --");
    }//main

 }//Type class

// (c) Lloyd Allison, 2012, Wed 23 Nov 2016. Licence to 'Type' is Gnu GPL.
