//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.la;

import java.io.*;  // for class PrintStream; see print(ps)

import la.util.*;   // for Series etc.
import la.maths.*;  // for Vector etc.
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** <CODE>Value = int + bool + char + triv + ...
              + (Value&nbsp;&times;&nbsp;Value)
              + (Value&nbsp;&rarr;&nbsp;Value)</CODE>,
    that is, first class Values.  Also see {@link Type}.  */
public abstract class Value implements Comparable <Value>
 {
   /** A class that implements Scannable can produce (by toSeries()) a
       {@link Series Series} of Values. A Scannable must be able to
       do this repeatedly, although there might be a time penalty.
       For example, see
       {@link Vector Vector}.{@link Vector#toSeries() toSeries()}.  */
   public static interface Scannable                             // * Scannable
    { /** Return a {@link Series Series} of Values. */
      public Series toSeries();
    }//Scannable interface

                                                     // parameterless constants
   public static final Triv triv   = new Triv();                     // triv
   public static final List nilVal = List.NIL;                       // nil
   public static final Bool ffalse = new Bool(0);                    // ffalse
   public static final Bool ttrue  = new Bool(1);                    // ttrue

   public static final Int  negOne = new Int( -1),
                            zero   = new Int(  0),
                            one    = new Int(  1),
                            two    = new Int(  2),
                            three  = new Int(  3),
                            four   = new Int(  4),
                            five   = new Int(  5),
                            six    = new Int(  6),
                            seven  = new Int(  7),
                            eight  = new Int(  8),
                            nine   = new Int(  9),
                            ten    = new Int( 10),
                            C      = new Int(100);

   public static final Real negTenR= real( -10.0),   //  -10.0
                            negOneR= real(  -1.0),
                            zeroR  = real(   0.0),   //    0.0
                            point1 = real(   0.1),
                            point5 = real(   0.5),
                           point01 = real(   0.01),
                           quarter = real(   0.25),
                            half   = real(   0.5),
                            point9 = real(   0.9),
                            oneR   = real(   1.0),   //    1.0
                            twoR   = real(   2.0),
                          threeR   = real(   3.0),
                           fourR   = real(   4.0),
                            fiveR  = real(   5.0),
                            tenR   = real(  10.0),   //   10.0
                            CR     = real( 100.0),   //  100.0
                            negCR  = real(-100.0),   // -100.0
                            E      = real(Math.E),
                            PI     = real(Math.PI),
                            PIby2  = real(Math.PI / 2.0);
   // .........................................................................

   /** Return the {@link Type} of 'this' Value.  */
   public abstract Type type();                                     // * type()

   /** Note, not all Values are actually Comparable.
       If&nbsp;ok it returns '-1', '0', '+1' for '&lt;', '=', '&gt;'.  */
   public int compareTo(Value v2)                              // compareTo(v2)
    { return bOp(Lexical.lt, v2) == ttrue ? -1
           : bOp(Lexical.gt, v2) == ttrue ?  1 : 0;
    }//compareTo(v2)

   /** <CODE>comparator.compare(v1,&nbsp;v2)</CODE> returns
       <CODE>v1.{@link #compareTo compareTo(v2)}</CODE>.
       Note, not all Values are actually Comparable.  */
   public static java.util.Comparator <Value> comparator          // comparator
    = new java.util.Comparator <Value>()
       { public int compare(Value v1, Value v2)
          { return v1.compareTo(v2); }
       };//new ...

   /** Apply unary Operator 'op' to 'this' Value, if implemented (this
       default throws an Exception). The int codes for <CODE>op</CODE>
       come from {@link Lexical}. Also see {@link #bOp(int,Value)}.  */
   public Value uOp(int op)                                          // uOp(op)
    { throw UOE( ".uOp(" + Lexical.int2sy(op) + ") is not valid" ); }

   /** Apply binary Operator 'op' to Values 'this' and 'rgt', if implemented
       (this default throws an Exception). The int codes for <CODE>op</CODE>
       come from {@link Lexical}. Also see {@link #uOp(int)}. */
   public Value bOp(int op, Value rgt)                           // bOp(op,rgt)
    { throw UOE(".bOp(" + Lexical.int2sy(op) + "," + rgt + ") is not valid"); }

   /** This default implementation returns 'this' Value, but
       the real interest is in {@link Defer#force()}. */
   public Value force() { return this; };                            // force()

   /** Note, evaluation is print driven; this default print(ps) calls
       <CODE>ps.print({@link #toString()})</CODE> and is only used by small,
       atomic values; other sub-classes <I>must override it</I>. */
   public void print(PrintStream ps)                               // print(ps)
    { ps.print( toString() ); ps.flush(); }

   /** Used, for example, in {@link #print printing} of
       {@link Expression.Const constants}, and in calls to {@link #error}.
       The result does not necessarily fully specify a large, or
       infinite(!), Value.  (This default returns the Class Name of 'this'.)
       Also see {@link #print(PrintStream) print(ps)}.  */
   public String toString()                                       // toString()
    { return "<" + this.getClass().getName() + ">"; }
   // .........................................................................

   // The following are only valid for certain subclasses. Their inclusion
   // here (i)&nbsp;reduces the need for casts, and (ii)&nbsp;makes them
   // methods of Defer, which <I>may</I> return a Value for which they
   // are valid.

   /** n() corresponds to an {@link Int#n() Int.n()} or an
       {@link Option#n() Option.n()} etc.; this default throws a
       RuntimeException but see {@link Bool#n() Bool.n()}, say.
       The inclusion of n(), and others, for all Values reduces the need
       for casts and also works nicely with {@link Defer}.  */
   public int n() { throw RTE(".n() is invalid"); }

   /** x() returns a {@link Cts#x() Cts.x()}, and also see
       {@link Int#x() Int.x()}, but this default throws a
       RuntimeException (see {@link #n()}). */
   public double       x() { throw RTE(".x() is invalid"); }
   /** See {@link Cts#AoM()}.
       (This default throws a RuntimeException.) */
   public double     AoM() { throw RTE(".AoM() is invalid"); }
   /** See {@link Cts#nlAoM()}.
       (This default throws a RuntimeException.) */
   public double   nlAoM() { throw RTE(".nlAoM() is invalid"); }

   /** This default throws an exception but see {@link Structured#nElts()}. */
   public int      nElts() { throw RTE(".nElts() is invalid"); }
   /** See {@link Structured#elt(int)}. */
   public Value elt(int i) { throw RTE(".elt(" + i + ") is invalid"); }

   /** This default throws an exception but see {@link Function#apply}. */
   public Value apply(Value v) { throw RTE(".apply(v) is invalid"); }
   // -------------------------------------------------------------------------

   /** The abstract class of atomic, non-{@link Structured structured} Values,
       notably {@link Discrete Discrete} and {@link Cts Cts}. */
   public static abstract class Atomic extends Value                // * Atomic
    { }//Atomic class
   // -------------------------------------------------------------------------

   /** Discrete Values, such as {@link Bool}, are subclasses of Discrete. */
   public static abstract class Discrete extends Atomic           // * Discrete
    {
      /** The <CODE>int</CODE> value or tag corresponding to 'this'
          Discrete Value.  */
      public abstract int n();

      // NB. There are no Unary operators (uOp(op)) as such on (all) Discretes.

      /** Apply binary operator 'op' to Discrete 'this', and to 'rgt'; this
          particular 'bOp(,)' takes care of the comparison operators.  */
      public Value bOp(int op, Value rgt)
       { if( ((1L << op) & Syntax.comparisonOprs) != 0 ) // =, <, etc.
          { if( ! type().ni(rgt) ) // makes rgt WHNF
               error( ".bOp(" + Lexical.int2sy(op) + "," + rgt + ")" );
            final int n = n(), rn = rgt.n();
            boolean ans;
            switch(op)
             { case Lexical.eq: ans = n == rn; break; // comparisons...
               case Lexical.ne: ans = n != rn; break;
               case Lexical.gt: ans = n >  rn; break;
               case Lexical.ge: ans = n >= rn; break;
               case Lexical.lt: ans = n <  rn; break;
               case Lexical.le: ans = n <= rn; break;
               default: throw RTE( ".bOp("+Lexical.int2sy(op)+","+rgt+")" );
             }
            return ans ? ttrue : ffalse;
          }//else...
         return super.bOp(op, rgt);
       }//bOp(op,rgt)
    }//Discrete class

   /** <CODE>() : {@link Type#TRIV TRIV} </CODE>. */
   public static class Triv extends Enum                                // Triv
    { public Triv() { super(0); }
      public Type.Triv type() { return Type.TRIV; }
      /** Returns zero, 0, that is the only case of Triv.  */
      public int n() { return 0; }
      /** Return the String representation of Triv.  */
      public String toString() { return Lexical.Symbol[Lexical.triv]; }
    }//Triv class

   /** Bool = false |&nbsp;true; also see {@link Type#BOOL}. */
   public static class Bool extends Enum                                // Bool
    { /** The <CODE>boolean</CODE> of 'this' Bool Value.  */
      public  final boolean b;
      /** n=0 for false, n=1 for true.
          Also see {@link #ffalse ffalse} and {@link #ttrue ttrue}.  */
      public Bool(int n)
       { super(n); this.b = n==1 ? true : false; }//constructor
      // NB. Must use class Enum (not Enum.GP) to avoid circularity of
      //     Type.BOOL : ttrue and ffalse
      /** The Type is {@link Type#BOOL}. */
      public Type.Enum type() { return Type.BOOL; }
      /** Apply unary operator, op, that is 'not', on 'this' Bool. */
      public Value uOp(int op)                                    // uOp(op)
       { switch(op)
          { case Lexical.notSy: return b ? ffalse : ttrue;
            default: return super.uOp(op);
       }  }//uOp(op)
      /** Apply a binary operator, 'op', that is 'and', 'or' or
          the comparisons, to Bools 'this' and 'rgt'. */
      public Value bOp(int op, Value rgt)                     // bOp(op,rgt)
       { switch(op) // try to be lazy on rgt...
          { case Lexical.andSy: return b ? (Bool)rgt.force() : ffalse;
            case Lexical.orSy:  return b ? ttrue : (Bool)rgt.force();
            default: return super.bOp(op, rgt);  // i.e., as for Discrete
       }  }//bOp(op,rgt)
    }//Bool class

   /** Also see {@link Type#INT}. */
   public static class Int  extends Discrete                             // Int
    { /** The <CODE>int</CODE> of 'this' Int Value.  */
      public final int n;
      public Int(int n) { this.n = n; }
      /** The Type is {@link Type#INT}. */
      public Type type() { return Type.INT; }
      /** The Java <CODE>int</CODE>, 'n', itself.  */
      public int    n() { return n; }
      /** Apply unary operator op, that is '+' or '-', on 'this' Int.
          Also see {@link Cts#uOp(int)}. */
      public Value uOp(int op)
       { switch(op)
          { case Lexical.plus:  return this;  // +x = x
            case Lexical.minus: return new Int(-n());
            default: return super.uOp(op);
          }
       }//uOp(op)
      /** Apply binary operator 'op', that is '+', '-', '*', '/' or
          the comparisons, to 'this' Int, and to 'rgt'. */
      public Value bOp(int op, Value rgt)
       { if( ((1L << op) & Syntax.arithOprs) != 0 )
          { if( Type.INT.ni(rgt) ) // makes rgt WHNF
             { final int rn = rgt.n();  int ans;
               switch(op)
                { case Lexical.plus:  ans = n + rn; break;
                  case Lexical.minus: ans = n - rn; break;
                  case Lexical.times: ans = n * rn; break;
                  case Lexical.over:  ans = n / rn; break;
                  default: throw RTE( ".bOp("+Lexical.int2sy(op)+","+rgt+")" );
                }//switch
               return new Int(ans);
             }
            else if( Type.CTS.ni(rgt) )     // e.g., 2 + 3.4 ...
               return real(n).bOp(op, rgt); // ... 2.0 + 3.4
          }
         return super.bOp(op, rgt); // eg. =, !=, etc.
       }//bOp(op,rgt)
      /** Float the {@link #n()} to a <CODE>double</CODE>. */
      public double x() { return (double) n(); }
      /** Return zero, i.e., Ints are "exact". */
      public double AoM() { return 0.0; }
      /** Return the String (numeral) of 'this' Int.  */
      public String toString() { return Integer.toString(n); }
    }//Int class


   /** A Char Value; also see {@link Type#CHAR}. */
   public static class Char extends Discrete                            // Char
    { /** The <CODE>char</CODE> of 'this' Char Value.  */
      public final char ch;
      public Char(char ch) { this.ch = ch; }
      /** The Type is {@link Type#CHAR}. */
      public Type type() { return Type.CHAR; }
      /** Return the Char's <CODE>int</CODE> code. */
      public int n() { return (int) ch; }
      /** Return the <CODE>'char'</CODE> of this Char.  */
      public String toString() { return "\'" + ch + "\'"; }
    }//Char class

   /** A Chars (String) Value; also see {@link Type#CHARS}. */
   public static class Chars extends Vector                            // Chars
    { /** The <CODE>String</CODE> of 'this' Chars.  */
      public final String s;
      public  Chars(String s) { this.s = s; }
      /** The Type is {@link Type#CHARS}. */
      public Type      type() { return Type.CHARS; }
      /** The element Type is {@link Type#CHAR}. */
      public Type   eltType() { return Type.CHAR;  }
      public int      nElts() { return s.length(); }
      public Char  elt(int i) { return new Char(s.charAt(i)); }
      /** Return the <CODE>"string"</CODE> of this Chars.  */
      public String toString() { return "\"" + s + "\""; }
    }//Chars class


   /** Enum Values, for example, A:DNA; also see {@link Type.Enum}. */
   public abstract static class Enum extends Discrete                 // * Enum
    { /** The option number of 'this' Enum Value.  */
      private final int  n;
      public Enum(int n)
       { this.n = n; }//constructor
      public abstract Type.Enum type();
      /** Return the Enum Value's int "code". */
      public int n() { return n; }
      /** Return the String representation of 'this' Enum Value.
          See {@link Type.Enum#ids}.  */
      public String toString() { return type().ids[n]; }

      /** A basic "general purpose" (GP) implementation of an Enum Value
          having a given {@link Type.Enum}&nbsp;t.  */
      public static class GP extends Enum                            // Enum.GP
       { private final Type.Enum t;
         /** 't' is the Enum Value's {@link Type.Enum type}, and
             'n' is the Enum Value's {@link Type.Enum#n() int&nbsp;code}.  */
         public GP(Type t, int n)
          { super(n); this.t = (Type.Enum) t; }
         public Type.Enum type() { return t; }
       }//Enum.GP class

    }//Enum class

   /** Continuous Values,
   <SPAN STYLE="white-space:nowrap">x()+/-({@link #AoM() AoM()}/2);</SPAN>
       see {@link Type#CTS} and {@link Real}.
       Also see {@link #cts(double,double)}.  */
   public static abstract class Cts extends Atomic                     // * Cts
    { public Type type() { return Type.CTS; }

      /** Return the <CODE>double</CODE> corresponding to 'this' Cts. */
      public abstract double x();

      /** The Accuracy of Measurement (AoM), <SPAN STYLE="white-space:nowrap">
          <CODE> {@link #x() x()}&plusmn;(AoM()/2), </CODE></SPAN>
          as in 3.14 to the nearest 0.01, say.  Incidentally, the
          "expected |error|", the distance from x() to the "true" Value,
          is AoM/4. */
      public abstract double AoM();

      /** The negative log {@link #AoM() AoM()} of 'this' Cts. */
      public double nlAoM()
       { final double aom = AoM();
         if( aom <= 0 ) error(".AoM = " + aom);
         return - Math.log(aom);
       }//nlAoM()

      /** Apply the unary operator, 'op', i.e., '+' or '-', to 'this' Cts.
          Also see {@link Int#uOp(int)} and {@link Lexical}. */
      public Value uOp(int op)
       { switch(op)
          { case Lexical.plus:  return this;
            case Lexical.minus: return cts(-x(), AoM());
            // NB if there were an op, inv(x), would have AoM()/x^2,
            //    e.g., 10.0 to 0.1 -> 0.1 to 0.001.
            default: return super.uOp(op);
       }  }//uOp(op)

      /** Apply the binary operator 'op' to 'this', and 'rgt', Cts.
          The matter of setting an effective AoM for the result of
          <SPAN STYLE="white-space:nowrap">+, -, *, /,</SPAN>
          is difficult so we set AoM to zero (for now???).
          Note, for (x+/-a) the expected |error| is a/2, and
          for (x+/-a)+(b+/-a) it is is 2a/3. And <I>arguably</I>
          the comparisons should be probabilistic for AoM&gt;0! */
      public Value bOp(int op, Value rgt)
       { if( ((1L << op) & (Syntax.arithOprs | Syntax.comparisonOprs)) != 0 )
          { if( Type.CTS.ni(rgt) || Type.INT.ni(rgt) )  // Cts 'op' (Cts|Int)
             { double ans;
               final double x1 = x(), x2 = rgt.x();
               switch(op)
                { case Lexical.plus:  ans = x1 + x2; break;  // +, -, * /
                  case Lexical.minus: ans = x1 - x2; break;
                  case Lexical.times: ans = x1 * x2; break;
                  case Lexical.over:  ans = x1 / x2; break;
                  case Lexical.eq: return x1 == x2 ? ttrue : ffalse;
                  case Lexical.ne: return x1 != x2 ? ttrue : ffalse;
                  case Lexical.gt: return x1 >  x2 ? ttrue : ffalse;
                  case Lexical.ge: return x1 >= x2 ? ttrue : ffalse;
                  case Lexical.lt: return x1 <  x2 ? ttrue : ffalse;
                  case Lexical.le: return x1 <= x2 ? ttrue : ffalse;
                  default: throw RTE( ".bOp("+Lexical.int2sy(op)+","+rgt+")" );
                }//switch
               return real(ans);
          }  }
         return super.bOp(op, rgt);
       }//bOp(op,rgt)

      /** Return a String representation of 'this' Cts Value.  */
      public String toString()
       { final double aom = AoM();
         return x() + (aom > 0 ? "+/-" + Util.x2String(aom/2,2) : "");
       }//toString()

    }//Cts class

   /** Return a {@link Cts Cts} Value representing x&plusmn;AoM/2.
       Also see {@link #real(double) real(x)}.  */
   public static Cts cts(final double x, final double AoM)
    { return AoM == 0 ? real(x) :
         new Cts()
          { public double   x() { return x;   }
            public double AoM() { return AoM; }
    };    }//cts(x,AoM)


   /** A class of <I>exact</I> {@link Cts} Values -- those having
       <CODE> {@link #AoM}&nbsp;=&nbsp;0 </CODE>.
       Continuous data (experimental or observational) cannot be
       exact Real, of course.  Also see {@link #real(double)}.  */
   public static abstract class Real extends Cts                      // * Real
    { // public double x() remains abstract.
      /** Returns 0.0 because a Real Value is precise. */
      public double AoM() { return 0.0; }
      /** Throws an exception because a Real Value is precise. */
      public double nlAoM()
       { throw RTE(" is exact, AoM=0, nlAoM=-infinity"); }
    }//Real class

   /** Return an exact {@link Real Real} Value corresponding to the
       double x. Continuous data (experimental or observational) cannot
       be exact Real, of course.
       Also see {@link #cts(double,double) cts(x,aom)}.  */
   public static Real real(final double x)
    { return new Real() { public double x() { return x; } };
    }//real(x)
   // -------------------------------------------------------------------------

   /** The super-class of Values, such as {@link Tuple Tuples} and
       {@link Vector Vectors}, having zero or more elements
       (components, fields), <CODE>{@link #elt(int) elt(i)}</CODE>.
       Also see {@link Atomic Atomic} Values.  */
   public static abstract class Structured extends Value        // * Structured
    {
      /** The number of elements (components, fields) in
          'this' Structured Value. */
      public abstract int nElts();                              // * nElts()

      /** The i-th element (component, field) of 'this'
          Structured Value, counting from zero. */
      public abstract Value elt(int i);                          // * elt(i)

      /** fst(),  short for {@link #elt elt(0)}. */
      public Value  fst() { return elt(0); }
      /** snd(),  short for {@link #elt elt(1)}. */
      public Value  snd() { return elt(1); }
      /** thrd(), short for {@link #elt elt(2)}. */
      public Value thrd() { return elt(2); }
      /** frth(), short for {@link #elt elt(3)}. */
      public Value frth() { return elt(3); }

      /** All the elements (components, fields) as an array; this default
          implementation may well be bettered in a sub-class. */
      public Value [] elts()                                       // elts()
       { final int nElts = nElts();
         if( nElts == 0 ) return null; // else ...
         final Value [] ans = new Value[nElts];
         for( int i = 0; i < nElts; i ++ ) ans[i] = elt(i);
         return ans;
       }//elts()

      /** The {@link Value#AoM() AoM} of 'this'
          <I>complete</I> Structured Value.
          Returns the exponential of -&nbsp;{@link #nlAoM()}.  */
      public double AoM() { return Math.exp( - nlAoM() ); }         // AoM()

      /** The <I>total</I> {@link Value#nlAoM() nlAoM} of 'this'
          <I>complete</I> Structured Value.  This default sums the
          nlAoM of all elements.
          There may be better ways in some sub-classes.  */
      public double nlAoM()                                       // nlAoM()
       { double ans = 0;
         for( int i = 0; i < nElts(); i ++ )
          { final Value ei = elt(i);
            if( ei.AoM() <= 0 )
               error(".nlAoM(), .elt("+i+").AoM()=" + ei.AoM() + "<=0");
            ans += ei.nlAoM();
          }//for
         return ans;
       }//nlAoM

      /** This default print(ps) outputs all elements,
          (elt(0),&nbsp;elt(1),&nbsp;...), to 'ps'.
          Also see {@link #opens()}, {@link #separator()},
          {@link #closes()},  and {@link #toString()}.  */
      public void print(PrintStream ps)                         // print(ps)
       { ps.print(opens());
         for( int i = 0; i < nElts(); i ++ )
          { ps.print(i > 0 ? separator() : "");  elt(i).print(ps); }
         ps.print(closes());
       }//print(ps)

      /** toString() <I>may</I> be acceptable for small Structured Values
          or you may want {@link #print print(ps)} instead?
          Also see {@link #opens()}, {@link #separator()},
          {@link #closes()}, and {@link Value#toString()}.  */
      public String toString()                                 // toString()
       { StringBuilder sb = new StringBuilder(opens());
         for( int i = 0; i < nElts(); i ++ )
            sb.append((i > 0 ? separator() : "") + elt(i));
         sb.append(closes());
         return sb.toString();
       }//toString()

      /** Is "(", but can override to change {@link #print formatting}. */
      protected String     opens() { return "("; }

      /** Is ",&nbsp;" but can override to change {@link #print formatting}. */
      protected String separator() { return ", "; }

      /** Is ")", but can override to change {@link #print formatting}. */
      protected String    closes() { return ")"; }

    }//Structured class
   // .........................................................................

   /** The class of <I>heterogeneous</I> k-Tuples, that is pairs, triples,
       and so on; also see {@link Type.Tuple} and {@link Vector}.  */
   public abstract static class Tuple extends Structured             // * Tuple
    {
      public abstract Type.Tuple type();

      /** For a Tuple consisting entirely of <I>bounded Discrete</I> Values,
          return the obvious integer encoding, hoping it doesn't overflow. */
      public int n()                                                     // n()
       { long n = 0;
         for( int i = 0; i < nElts(); i ++ )
          { Value ei = elt(i); Type.Discrete ti = (Type.Discrete) ei.type();
            if( ti.hasLwb() && ti.hasUpb() )
               n = n * (ti.upb_n() - ti.lwb_n() + 1) + ei.n();
            else RTE( ".n(), component " + i + " is not bounded" );
          }
         return (int) n;
       }//n()

      public abstract int nElts();

      /** Return a sub-Tuple of 'this', made of just the
          elements at the selected positions, 'ps'. */
      public Tuple elts(final int [] ps)
       { final Tuple t = this;  final int nE = nElts();
         for( int i = 0; i < ps.length; i ++ )
          { final int ps_i = ps[i];
            if( ps_i < 0 || ps_i >= nE )
               error( ".elts(ps), ps[" + i + "]=" + ps[i] );
          }//AOK
         return new Tuple()
          { public Type.Tuple type() { return Type.tuple(ps.length); }
            public int       nElts() { return ps.length; }
            public Value  elt(int i) { return t.elt(ps[i]); }
       }; }//elts(ps)

      /** Returns the i-th elt ({@link #force forced}). */
      public abstract Value elt(int i);

      /** Apply unary operator, 'op', element-wise to 'this' Tuple.
          Also see {@link Lexical} for op codes.  */
      public Tuple uOp(int op)
       { Value[] es = new Value[nElts()];
         for( int i = 0; i < es.length; i ++ )
            es[i] = elt(i).uOp(op);
         return tuple(es);
       }//uOp(op)

      /** Apply binary operator, 'op', element-wise to Tuples 'this'
          and 'rgt'. Also see {@link Lexical} for op codes. */
      public Tuple bOp(int op, Value /*Tuple*/ rgt)
       { if( (! type().ni(rgt)) ) // includes "arity"
            error( ".bOp(" + op + "," + rgt + ") type error" );
         final Tuple rt = (Tuple) rgt.force();
         final Value[] es = new Value[nElts()];
         for( int i = 0; i < nElts(); i ++ )
            es[i] = elt(i).bOp(op, rt.elt(i));
         return tuple(es);
       }//bOp(op,rgt)

      /** Is 'this' Tuple &lt; (-1), &gt; (+1), or equal (0) to 'rgt'?
          Note, 'this' and rgt must have the same shape.  */
      public int compareTo(Value /*Tuple*/ rgt)             // ZZZ ???
       { if( (! type().ni(rgt)) ) // includes "arity"
            error( ".compareTo(" + rgt + ") type error" );
         final Tuple rt = (Tuple) rgt.force();
         int c = 0;
         for( int i = 0; (c == 0) && (i < nElts()); i ++ )
            c = elt(i).compareTo(rt.elt(i));
         return c;
       }//compareTo(rgt)

      /** From 'this' Tuple of Vectors (which is checked), return a
          {@link Vector#zip(Vector[]) zipped} Vector of Tuples. */
      public Vector zip()
       { final Vector[] cols = new Vector[ nElts() ];
         for( int c = 0; c < cols.length; c ++ )
          { final Value ec = elt(c);
            if( ! (ec instanceof Vector) )
               error( ".zip(), column " + c + " is not a Vector" );
            cols[c] = (Vector) ec;
          }
         return Vector.zip(cols);
       }//zip()

      /** A simple general purpose (GP) implementation of a Tuple Value. */
      public static class GP extends Tuple                          // Tuple.GP
       { /** The elements of 'this' Tuple.GP. */
         public final Value [] elts;
         /** Construct a Tuple.GP from several elements 'elts'. */
         public GP(Value [] elts) { this.elts = elts; }//constructor
         public Type.Tuple type() { return Type.tuple(elts.length); }
         public int       nElts() { return elts.length; }
         public Value  elt(int i) { return elts[i]; }
       }//Tuple.GP class

    }//Tuple class


   /** Check that 'this' Value really is a k-Tuple and, if it is,
       return it as a Tuple. E.g., useful in a Function (method) that
       has a Pair as a parameter, e.g.,
       <CODE>Tuple&nbsp;tv=v.isTuple(2)</CODE>.  */
   public Tuple isTuple(int k)
    { if( ! Type.tuple(k).ni(this) )  error( " is not a " + Type.tuple(k) );
      return (Tuple) this;
    }//isTuple(k)

   /** tuple, a convenience function to make
       a <CODE>k-{@link Tuple Tuple}</CODE>.  */
   public static Tuple.GP tuple(final Value [] vs)
    { return new Tuple.GP(vs); }

   // single, a convenience function to make
   //  a <CODE>1-{@link Tuple Tuple}</CODE>.  */
   // p ublic static Tuple single(Value e0)  not a good idea?
   //  { return tuple(new Value[]{e0}); }

   /** pair, a convenience function to make
       a <CODE>2-{@link Tuple Tuple}</CODE>. */
   public static Tuple.GP pair(Value e0, Value e1)
    { return tuple(new Value[]{e0, e1}); }

   /** triple, a convenience function to make
       a <CODE>3-{@link Tuple Tuple}s</CODE>. */
   public static Tuple.GP triple(Value e0, Value e1, Value e2)
    { return tuple(new Value[]{e0, e1, e2}); }

   /** quad, a convenience function to make
       a <CODE>4-{@link Tuple Tuple}</CODE>. */
   public static Tuple.GP quad(Value e0, Value e1, Value e2, Value e3)
    { return tuple(new Value[]{e0, e1, e2, e3}); }
   // -------------------------------------------------------------------------

   /** An Option Value has a "tag", {@link #n()}, and zero or more
       sub-Values, {@link #elt(int)} depending on the tag.
       For example, see linked {@link List Lists}.
       Also see {@link Type.Option}. */
   public static abstract class Option extends Structured           // * Option
    {
      /** Return this Option Value's number within its
          Option&nbsp;{@link Type.Option Type}. */
      public abstract int n();                                      // * n()

      /** Return this Option's number of elements (fields). */
      public int nElts()
       { return ((Type.Option) type()) .arities[n()]; }

      /** Is 'this' Option &lt;&nbsp;(-1), &gt;&nbsp;(+1), or =&nbsp;(0) rgt?
          Note, 'this' and 'rgt' must be of the same Option Type.
          The result is firstly based on {@link #n() n()} then,
          if n()s are equal, on the elements in order.   */
      public int compareTo(Value rgt)
       { if( ! type().ni(rgt) )
            error( ".compareTo(" + rgt + ") type error" );
         final Option rgtO = (Option) rgt;
         final int n0 = n(), n1 = rgtO.n();
         if( n0 < n1 ) return -1; // else...
         if( n0 > n1 ) return  1; // else...
         int c = 0;
         for( int i = 0; (c == 0) && (i < nElts()); i ++ )
            c = elt(i).compareTo(rgtO.elt(i));
         return c;
       }//compareTo(rgt)


      // See class Structured for Option's print(ps).

      /** Is ["("]+id where id is the {@link Type.Option#ids name}
          of 'this' Option Value; see {@link #print}.
          For example, EmptyTree &nbsp;|&nbsp; (Fork&nbsp;...).  */
      protected String opens()
       { final boolean some = nElts() > 0;
         return (some?"(":"") + ((Type.Option)type()) .ids[n()]
              + (some?separator():"");
       }//opens()

      /** Is space, "&nbsp;"; see {@link #print}. */
      protected String separator() { return " "; }

      /** Is ")" if nElts()&gt;0; see {@link #print}. */
      protected String    closes() { return nElts() > 0 ? ")" : ""; }


      /** A general purpose (GP) implementation of an {@link Option} Value;
          also see {@link Type.Option}, and {@link List}.  */
      public static class GP extends Option                        // Option.GP
       { /** The Type of 'this' Option. */
         public final Type.Option t;
         /** The particular Option number, 'n', within Type 't'. */
         public final int n;
         /** The elements (components, fields) 'elts', if any. */
         public final Value [] elts;

         /** Construct an Option Value of Type t, number n, zero elements. */
         public GP(Type t, int n)
          { this.t = (Type.Option) t; this.n = n; this.elts = null;
            check();
          }//constructor

         /** Construct an Option Value of Type t, number n, one element v. */
         public GP(Type t, int n, Value v)
          { this(t, n, new Value [] {v}); }//constructor

         /** Construct an Option Value of Type t, number n, elements elts. */
         public GP(Type t, int n, Value [] elts)
          { this.t = (Type.Option) t; this.n = n; this.elts = elts;
            check();
          }//constructor

         private void check()
          { final int ne = elts == null ? 0 : elts.length;
            if( this.t.arities[n] != ne )
               error( ".constructor, #"+n+" has "+this.t.arities[n]+" elts" );
          }//check()

         public Type.Option type() { return t; }
            /** Return the Option number, {@link #n}. */
         public int         n() { return n; }
         public Value [] elts() { return elts; }
            /** Return 'this' Option's i-th element,
                {@link #elt elt(i)}, in {@link #force WHNF}. */
         public Value elt(int i)
          { final Value ei = elts[i].force(); elts[i] = ei; return ei; }
       }//Option.GP class

    }//Option class


   /** Linked Lists, a special implementation of the abstract {@link Option}.
       Note that the tail (tl) of a non-empty List must itself be a List. */
   public static abstract class List extends Option                   // * List
    {
      private static final String sep =
         " " + Lexical.Symbol[Lexical.consSy] + " ";
         // sep = ??? " " + Lexical.Symbol[Lexical.consSy] + "\n";

      /** The {@link Type#LIST LIST} Type, of course. */
      public Type.Option type() { return Type.LIST; }
      /** Generally use {@link #hd()} or {@link #tl()} instead.  */
      public Value elt(int i)                                      // elt(i)
       { switch( i ) { case 0: return hd();  case 1: return tl(); }
         throw RTE(".elt(" + i + ")");
       }//elt(i)
      /** Is 'this' List 'nil' (empty), or not?  */
      public abstract Value.Bool isNil();                       // * isNil()
      /** The head (1st element) of the List. */
      public abstract Value hd();                                  // * hd()
      /** The tail (all but the {@link #hd()}) of the List. */
      public abstract Value tl();                                  // * tl()
      /** Note, the tail of a List is printed <I>iteratively</I>. */
      public void print(PrintStream ps)                         // print(ps)
       { if( isNil().b ) ps.print(this);
         else
          { ps.print("("); hd().print(ps);
            Value l = tl();
            while( l instanceof Cell )
             { final Cell lc = (Cell) l;
               ps.print(sep); lc.hd().print(ps);
               l = lc.tl();
             }
            if( ! (l instanceof List) ) l.print(ps);  // should not happen!
            ps.print(")");
          }//if
       }//print(ps)

      /** <I>The</I> empty list, NIL; also see {@link Cell}. */
      public static final List NIL = new List()                     // List.NIL
       { /** NIL is Option number&nbsp;0 of List. */
                      public int            n() { return 0; }
         /** zero */  public int        nElts() { return 0; }
         /** ttrue */ public Value.Bool isNil() { return ttrue; }
         /** error */ public Value hd() { throw RTE(".hd()"); }
         /** error */ public Value tl() { throw RTE(".tl()"); }
         // NB. No binary operators as such on List, cons handled otherwise
         /** Unary operator, op, i.e., null, hd, or tl, on NIL.
             (Or operators on the super class?)
             Also see {@link Lexical} for op codes.  */
         public Value uOp(int op)
          { switch(op)
             { case Lexical.hdSy:   return hd();
               case Lexical.tlSy:   return tl();
               case Lexical.nullSy: return ttrue;
               default: return super.uOp(op);
          }  }//uOp(op)
       };//new List()

      /** A List Cell; also see {@link #NIL}. */
      public static class Cell extends List                        // List.Cell
       { /** The head, hd, and tail, tl, of 'this' List Cell. */
         protected Value hd, tl; // NB. tl is quite likely Defer-ed.
         public Cell(Value hd, Value tl)
          { this.hd = hd; this.tl = tl; }//constructor
         /** A Cell is Option number&nbsp;1 of List (NIL is number&nbsp;0). */
         public int            n() { return 1; }
         /** 2 elements */
         public int        nElts() { return 2; }
         /** ffalse */
         public Value.Bool isNil() { return ffalse; }
         /** The head of the List. */
         public Value         hd() { hd = hd.force(); return hd; }
         /** The tail (rest) of the List. */
         public Value         tl() { tl = tl.force(); return tl; }
         // NB. No binary operators as such on List, cons handled otherwise
         /** Unary operator, op, i.e., null, hd, or tl, on Cells.
             (Or operators on the super class?)  */
         public Value uOp(int op)
          { switch(op)
             { case Lexical.hdSy:   return hd();
               case Lexical.tlSy:   return tl();
               case Lexical.nullSy: return ffalse;
               default: return super.uOp(op);
          }  }//uOp(op)
       }//List.Cell class
    }//List class

   /** Convenience function returning the List h::t.
       Note, t must be a List. */
   public static List.Cell cons(Value h, Value t)
    { return new List.Cell(h, t); }
   // -------------------------------------------------------------------------

   /** Maybe t = None |&nbsp;Just&nbsp;t, for where a Value may be missing
       ({@link #None None}), or present ({@link Just Just&nbsp;v}).
       Also see {@link Inc_Or Inc_Or&nbsp;t0&nbsp;t1}.  */
   public static abstract class Maybe extends Option                   // Maybe
    { /** NONE = 0, JUST = 1. */
      public static final int NONE = 0, JUST = 1;

      /** Its Type is {@link Type#MAYBE}. */
      public Type.Option type() { return Type.MAYBE; }

      /** None, where the optional Value is missing, absent, late,... . */
      public static final Maybe None =                            // Maybe.None
      new Maybe()
       { public int     n() { return NONE; }
         public int nElts() { return 0; }
         public Value elt(int n) { throw RTE( ".elt(" + n + ")" ); }
       };//new Maybe()

      /** The Value is present, Just&nbsp;v. */
      public static class Just extends Maybe                      // Maybe.Just
       { /** The Value, v, in Just&nbsp;v is present. */
         public final Value v;
         public Just(Value v) { this.v = v; }//constructor
         public int     n() { return JUST; }
         public int nElts() { return 1; }
         public Value elt(int n)
          { if( n != 0 ) error( ".elt(" + n + ")" );
            /* else */ return v;
          }//elt(n)
       }//Maybe.Just class
    }//Maybe class

   /** Equal to Maybe.None */
   public static final Maybe None = Maybe.None;                         // None

   /** Convenience function, returns new Maybe.Just(v).  */
   public static Maybe.Just just(Value v)                            // just(v)
    { return new Maybe.Just(v); }
   // -------------------------------------------------------------------------

   /** Inc_Or&nbsp;t0&nbsp;t1 = Left&nbsp;t0 | Right&nbsp;t1
                              | Both&nbsp;t0&nbsp;t1,
       for where one, or both, of v0:t0 and v1:t1 can be present.
       Also see {@link Maybe Maybe&nbsp;t}.  */
   public static abstract class Inc_Or extends Option                 // Inc_Or
    { /** LEFT = 0, RIGHT = 1, BOTH = 2. */
      public static final int LEFT = 0, RIGHT = 1, BOTH = 2;

      /** Its Type is {@link Type#INC_OR}. */
      public Type.Option type() { return Type.INC_OR; }

      /** Left v0, the first (Left) Option alone. */
      public static class Left extends Inc_Or                    // Inc_Or.Left
       { /** The Value, v0, in Left&nbsp;v0. */
         public final Value v0;
         public Left(Value v0) { this.v0 = v0; }//constructor
         public int     n() { return LEFT; }
         public int nElts() { return 1; }
         public Value elt(int n)
          { if( n != 0 ) error( ".elt(" + n + ")" );
            return v0;
          }//elt(n)
       }//Inc_Or.Left class

      /** Right v1, the second (Right) Option alone. */
      public static class Right extends Inc_Or                  // Inc_Or.Right
       { /** The Value, v1, in Right&nbsp;v1. */
         public final Value v1;
         public Right(Value v1) { this.v1 = v1; }//constructor
         public int     n() { return RIGHT; }
         public int nElts() { return 1; }
         public Value elt(int n)
          { if( n != 0 ) error( ".elt(" + n + ")" );
            return v1;
          }//elt(n)
       }//Inc_Or.Right class

      /** Both v0 v1. */
      public static class Both extends Inc_Or                    // Inc_Or.Both
       { /** The Values, v0, v1, in Both&nbsp;v0&nbsp;v1. */
         public final Value v0, v1;
         public Both(Value v0, Value v1)
          { this.v0 = v0; this.v1 = v1; }//constructor
         public int     n() { return BOTH; }
         public int nElts() { return 2; }
         public Value elt(int n)
          { if( n == 0 ) return v0;
            if( n == 1 ) return v1;
            throw RTE( ".elt(" + n + ")" );
          }//elt(n)
       }//Inc_Or.Both class
    }//Inc_Or class
   // -------------------------------------------------------------------------

   /** Function produced from a {@link Expression.LambdaExp} and an
       {@link Environment}.  Also see {@link Function.Native}.  */
   public static class Lambda extends Function                        // Lambda
    { public final Expression.LambdaExp code;  public final Environment r;

      public Lambda(Expression.LambdaExp code, Environment r)
       { this.code = code;  this.r = r; }// constructor, i.e., a closure

      /** Calls <CODE>code.body.{@link Expression#eval eval(r')}</CODE>, where
          r' binds the formal parameter to the actual parameter.<BR>
          Also see {@link Expression#eval} on tail recursion.  */
      public Value apply(Value aparam)
       { if( code.fparam == Expression.triv ) // ((\ ().e) exp, type check!
            if( aparam.force() == Value.triv ) return code.body.eval(r);
            else throw RTE(" (L ().e)exp, exp not ()");
         else                                 // ((\ x.e) exp
          { Expression.Ident fparam = (Expression.Ident) code.fparam;
            return code.body.eval( r.bind(fparam.id, aparam) );
          } // result is in WHNF
       }//apply
    }//Lambda class
   // -------------------------------------------------------------------------

   /** The class of lazy, that is, not yet computed, <I>Deferred</I> Values,
       {@link Defer.Exp} and {@link Defer.App}.  The action happens in
       {@link #force()}.  Also see {@link Expression#defer}. */
   public static abstract class Defer extends Value                  // * Defer
    { /** v holds the once Deferred Value when (if) it is eventually
          computed, by {@link #force}, to at least WHNF. */
      protected Value v = null;

      /** Cause 'this' lazy, Deferred Value to be computed to at last WHNF
          and cached in '{@link #v v}'. */
      public abstract Value force();                               // * force()

      /** Note, <CODE>exp.defer(env).print(ps)</CODE> is what causes
          <CODE>{@link Expression#eval <U>eval</U>}uation</CODE> to happen by
          <CODE>{@link #force() force()}.print()</CODE>-ing the program.  */
      public void print(PrintStream ps) { force().print(ps); }
      // .....................................................

      /** All unary operators are strict, so force() 'this' and
          return v.uOp(op). */
      public Value uOp(int op)
       { force(); return v.uOp(op); }

      /** All binary operators (other than 'cons') are strict on 'this', so
          force() and return v.bOp(op,rgt) for binary operator 'op'. */
      public Value bOp(int op, Value rgt)
       { force(); return v.bOp(op, rgt); }  // strict on this

      /** force(), and return the v.type() of this Deferred Value. */
      public Type      type() { force(); return v.type(); }

      /** force(), and return the v.n() of this Deferred Value. */
      public int          n() { force(); return v.n(); }

      /** force(), and return the v.x() of this Deferred Value. */
      public double       x() { force(); return v.x(); }

      /** force(), and return the v.AoM of this Deferred Value. */
      public double     AoM() { force(); return v.AoM(); }

      /** force(), and return the v.nlAoM of this Deferred Value. */
      public double   nlAoM() { force(); return v.nlAoM(); }

      /** force(), and return the v.nElts() of this Deferred Value. */
      public int      nElts() { force(); return v.nElts(); }

      /** force(), and return the v.elt(i) of this Deferred Value. */
      public Value elt(int i) { force(); return v.elt(i); }

      /** force() the Function and return v.apply(p). */
      public Value apply(Value p) { force(); return v.apply(p); }
      // .........................................

       /** A lazy unevaluated, Deferred Expression is a closure, (e,&nbsp;r),
           a <NOBR>({@link Expression}, {@link Environment}).</NOBR> */
      public static class Exp extends Defer                        // Defer.Exp
       { /** The Expression to be
             <CODE>{@link Expression#eval eval}-uated</CODE>,
             using Environment {@link #r r}, at some later date, maybe. */
         protected Expression  e;
         /** The Environment to be used with Expression {@link #e e},
             at some later date, maybe. */
         protected Environment r;
         /** Construct an un-eval-uated
             ({@link Expression},&nbsp;{@link Environment}) pair. */
         public Exp(Expression e, Environment r)
          { this.e = e; this.r = r; }//constructor

         /** Cause 'this' Deferred Expression {@link #e} to be
             <CODE>{@link Expression#eval eval}-uated</CODE> to WHNF
             and cached in '{@link #v v}'. */
         public Value force()
          { if( v == null )
             { v = e.eval(r);      // overwrite. Note, v now in WHNF
               e = null; r = null; // free up space
             }
            return v;
          }//force()

         /** Note, may be used in error messages, and for debugging,
             <B>only</B>; you probably want {@link #print print(ps)}
             instead!  */
         public String toString()
          { return "Defer(" + (e != null ? e : "null") + ",Env)"; }
       }//Defer.Exp class
      // .........................................

      /** A lazy, un-apply-ed, Deferred Function-application, a
      <SPAN STYLE="white-space:nowrap">(Function, actual-parameter)</SPAN>
          pair. */
      public static class App extends Defer                        // Defer.App
       { /** The Function, 'f', and actual parameter, 'ap', to be
             applied at a later date, maybe. */
         protected Value f, ap;
         /** Construct an unapplied
             <NOBR>(Function, actual-parameter)</NOBR> pair. */
         public App(Value f, Value ap)
          { this.f = f; this.ap = ap; }//constructor

         /** Cause Function '{@link #f}' to be
             <CODE>{@link Function#apply apply}-ed</CODE> to actual parameter
             '{@link #ap}' and the result cached in '{@link #v v}'. */
         public Value force()
          { if( v == null )
             { v = ((Function) f.force()) .apply(ap); // overwrite
               f = null; ap = null;                   // free up space
             }
            return v;
          }//force()

         /** Note, may be used in error messages, and for debugging,
             <B>only</B>; you probably want {@link #print} instead! */
         public String toString()
          { return "Defer(" + (f != null ?  f : "null") + " "
                           + (ap != null ? ap : "null") + ")";
          }//toString()
       }//Defer.App class

    }//Defer class
   // -------------------------------------------------------------------------

   /** Used to prevent recursion within {@link #errMsg(String)}!  */
   private int errMsgCount = 0;

   /** It has all gone pear-shaped; identify 'this' and 'msg'.  See, e.g.,
       {@link #error error(str)},
       {@link #RTE   RTE(str)}, and
       {@link #UOE   UOE(str)}.  */
   protected String errMsg(String msg)                           // errMsg(msg)
    { String thisS = "this";
      if( errMsgCount <= 0 )           // Beware...
         try{ errMsgCount ++ ;         // ...this.toString()...
              thisS = this.toString(); // ...might cause another errMsg ?
              errMsgCount -- ;         // ...which would be inf.-recursion.
            }catch(Exception e){ thisS = getClass().getName(); }
      return "error, " + thisS + msg;
    }//errMsg(msg)

   /** Short for RuntimeException({@link #errMsg errMsg(msg)})  */
   protected RuntimeException RTE(String msg)                       // RTE(msg)
    { return new RuntimeException(errMsg(msg)); }
      // NB. a RuntimeException need not be listed in a method's throws list.

   /** Short for UnsupportedOperationException({@link #errMsg errMsg(msg)}) */
   protected UnsupportedOperationException UOE(String msg)          // UOE(msg)
    { return new UnsupportedOperationException(errMsg(msg)); }

   /** Throws {@link #RTE RTE(msg)}. */
   protected void error(String msg)                               // error(msg)
    { throw RTE(msg); }
   // -------------------------------------------------------------------------

   private static String show(Value v) { return v + ":" + v.type(); }

   /** main() allows Value to be (very slightly) tested in isolation. */
   public static void  main(String[] argv)
    { System.out.println("-- test Value --");

      Char z     = new Char('z');
      Real six   = real(6);
      Cts pi_ish = cts(3.14, 0.01);
      Tuple  tpl = tuple( new Value[]{PI, ten} );
      Vector vec = Vector.values( new Value[]{ten,ten,ten} );
      Vector vec2= Vector.doubles(2.0, new double[]{1, 2, 3});
      System.out.println( "   triv=" + show(triv) + ", ttrue=" + show(ttrue)
                      + ", z=" + show(z)
                      + ",\n   ten=" + show(ten)
                      + ", pi=" + show(PI) + ", pi_ish=" + show(pi_ish)
                      + ",\n   vec=" + show(vec)
                      + ", vec2=" + show(vec2)
                      + ",\n   tpl=" + show(tpl)
                      + ",\n   nilVal=" + show(nilVal)
                          + ", 6::6::nil=" + show(cons(six,cons(six,nilVal)))
                      + ",\n   Just pi=" + show(new Maybe.Just(PI))
                           + ", None=" + show(Maybe.None)
                      + ",\n   Both six z=" + show(new Inc_Or.Both(six, z))
                            + ", Left six=" + new Inc_Or.Left(six)
                            + ", Right z=" + new Inc_Or.Right(z)
                        );

      System.out.println("-- done Value --");
    }//main

 }//Value class, also see  http://www.allisons.org/ll/Computing/

// (c) Lloyd Allison, Oct 2007, Sat 18 Nov 2017. Licence to 'Value' is Gnu GPL.
