//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.maths;

import java.util.Comparator;

import la.util.Util;
import static la.util.Util.error;
import la.la.Value;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Useful mathematical constants and functions.
    Also see {@link README README}.  */
public class Maths
 {
   /** A tiny amount, epsilon, to allow for rounding error where necessary. */
   public static final double epsilon = 0.000001;

   /** twoPI = 2&pi;. */
   public static final double twoPI = 2.0 * Math.PI;

   /** Is |x-y| &le; a standard '{@link #epsilon epsilon}'-fraction
       of <NOBR>max(|x|, |y|) ?</NOBR> */
   public static boolean nearlyEqual(double x, double y)
    { return Math.abs(x - y) <= Math.max(Math.abs(x), Math.abs(y)) * epsilon; }

   /** Is |x-y| &le; an epsilon-fraction of <NOBR>max(|x|, |y|) ?</NOBR> */
   public static boolean nearlyEqual(double x, double y, double epsilon)
    { return Math.abs(x - y) <= Math.max(Math.abs(x), Math.abs(y)) * epsilon; }
   // -------------------------------------------------------------------------

   /** log<SUB>e</SUB>(<SUP>n</SUP>C<SUB>k</SUB>).
       <NOBR>n choose k, <SUP>n</SUP>C<SUB>k</SUB> = n! / (k! (n-k)!).</NOBR>
       Calls {@link #logFactorial logFactorial}.  */
   public static double log_nCk(int n, int k)
    { assert k >= 0 && n >= k : "log_nCk(n=" + n + ", k=" + k + ")";
      if( n-k < k ) k = n-k;  // symmetry k : (n-k)
      if( k > 6 )
         return logFactorial(n) - logFactorial(k) - logFactorial(n-k);
      double sum = 0; // = log(1), of course.
      for( int i = n; i > n-k; i -- ) // n (n-1) ... (n-k+1) ...
         sum += Math.log(i);          // ... i.e., n! / (n-k)!
      for( int i = 1; i <= k; i ++ )
         sum -= Math.log(i);          // i.e., n! / ( (n-k)! k! )
      return sum;
      // nC0 = 1, ..., (2m)Cm = (2m)!/(m!^2), ..., nCn = 1
    }//log_nCk(n,k)

   /** 'logSum', aka 'logPlus',<BR>
       <NOBR> logSum<SUB>i</SUB>( - log(p[i]))
        = - log( &sum;<SUB>i</SUB> exp( log(p[i]) )
        = - log( &sum;<SUB>i</SUB> p[i] )</NOBR>.
       (But it is not implemented that way.)
       For&nbsp;example,
       <NOBR>logSum({10, 10}) = 10 - log(2) nits.</NOBR>  */
   public static double logSum(double [] nlXs)
    { final double limit = 30; // i.e., limit nits ~ prob of zero.
      int min = 0;  double nlXs_min = nlXs[0];
      for( int i = 1; i < nlXs.length; i ++ )
         if( nlXs[i] < nlXs_min )
          { min = i;  nlXs_min = nlXs[i]; }
      double sum = 0;
      for( int i = 0; i < nlXs.length; i ++ )
       { final double nlXs_i = nlXs[i] - nlXs_min;  // # of nits > the min
         if( nlXs_i <= limit ) sum += Math.exp( - nlXs_i );
       }
      return nlXs_min - Math.log(sum);
    }//logSum(nlXs)

   /** Equivalent to
       <NOBR><TT>{@link #logSum(double[]) logSum}({nlX, nlY})</TT>,</NOBR>
       but a little quicker. */
   public static double logSum(double nlX, double nlY)
    { double a, b;
      if( nlX >= nlY ) { a = nlX; b = nlY; } // a >= b
      else             { a = nlY; b = nlX; } // a >= b
      final double ab = a-b;   // >= 0
      if( ab > 30 ) return b;  // a >= b, so exp(-a) <= exp(-b)
      return b - Math.log(1.0 + Math.exp( - ab ));
    }//logSum(nlX,nlY)

   /** Convert an array, nlPr, of message lengths (negative log probabilities)
       into a normalised array of probabilities.
       Also see {@link #nlPrs2odds(double[],double[])}.  */
   public static double [] nlPrs2odds(double [] nlPr)
    { double [] pr = new double [nlPr.length];
      nlPrs2odds(nlPr, pr);
      return pr;
    }//nlPrs2prs(nlPr)

   /** Convert an array, nlPr, of negative log probabilities
       into a normalised array of probabilities, pr.  Afterwards,
       &sum;<SUB>i</SUB>&nbsp;pr[i]&nbsp;=&nbsp;1.  */
   public static void nlPrs2odds(double [] nlPr, double [] pr)
    { final int N = nlPr.length;
      assert pr.length == N : "nlPr[" + N + "] : pr[" + pr.length + "]";
      double min = nlPr[0];
      for( int i = 1; i < N ; i ++ ) if( nlPr[i] < min ) min = nlPr[i];
      double sum = 0;
      for( int i = 0; i < N; i ++ )
       { pr[i] = Math.exp( min - nlPr[i] ); // i.e., -(nlPr[i]-min)
         assert pr[i] >= 0 && pr[i] <= 1 : "pr[" + i + "]=" + pr[i];
         sum += pr[i];
       }
      for( int i = 0; i < N; i ++ ) pr[i] /= sum; // normalise
    }//nlPrs2prs(nlPr, pr)
   // -------------------------------------------------------------------------

   /** Return the N<SUP>th</SUP> Catalan number, N&nbsp;&ge;&nbsp;0.
       <NOBR>This is <SUP>2N</SUP>C<SUB>N</SUB> / (N+1).</NOBR> The series
       starts 1,&nbsp;1,&nbsp;2, 5, 14, 42, 132, ...  Takes O(N)-time.
       The N-th Catalan number is the number of full binary trees with
       N internal nodes, equivalently with N+1 leaves.
       Also see {@link #cummulativeCatalans cummulativeCatalans(N)}. */
   public static long catalan(int N)
    { if( N < 0 ) error( "catalan(" + N + ")" ); // else...
      if( N == 0 ) return 1;                     // else{N > 0} ...
      long last = 1;
      for( int n = 1; n <= N; n ++ )
       { final long two_n = n+n, nSq = n*n;
         final long next = last * 2 * (two_n - 1) / n;
         last = next;
       }//n==N when done
      return last / (N+1);
    }//catalan(N)

   /** Return &sum;<SUB>i=0..N</SUB> {@link #catalan catalan(i)}, N&ge;0.
       Takes O(N)-time. */
   public static long cummulativeCatalans(int N)
    { if( N < 0 ) error( "cummulativeCatalans(" + N + ")" ); // else...
      if( N == 0 ) return 1;                     // else{N > 0} ...
      long sum = 1, last = 1;
      for( int n = 1; n <= N; n ++ )
       { final int two_n = n+n, n1 = n+1, nSq = n*n;
         final long next = last * 2 * (two_n - 1) / n;
         sum += next / n1;
         last = next;
       }//n==N when done
      return sum;
    }//cummulativeCatalans(N)
   // -------------------------------------------------------------------------

   /** Math.exp({@link #logGamma logGamma(x)}), for real&nbsp;x. Note,
       <NOBR> &Gamma; n = (n-1)!, </NOBR> for int&nbsp;n&ge;1, and
       <NOBR> &Gamma; (x+1) = x &Gamma; x. </NOBR>
       Also defined for non-integer x&lt;0 but there will be
       numerical problems for large negative values.
       (Recall that for int n&lt;0,&nbsp; n! is infinite.)   */
   public static double Gamma(double x)
    { return x >= 0 ? Math.exp(logGamma(x))
                    : Math.PI / (Math.sin(Math.PI*x) * Gamma(1-x)); // Euler
                    // NB. undefined (infinite) on integers n <= 0.
    }//Gamma(x)

   /** Implemented as {@link #logGamma(double) logGamma(n+1)}.  */
   public static double logFactorial( int n )
    { return logGamma( n+1 );
      // double sum = 0;
      // for( int i = 1; i <= n; i++ ) sum += Math.log( i );
      // return sum;
    }//logFactorial(n)

   /** Implemented as {@link #logGamma(double) logGamma(x+1)}.  */
   public static double logFactorial( double x )
    { return logGamma( x+1 ); }

   /** log(&Gamma;(x)), real x&gt;0; &nbsp;
       <NOBR> note, &Gamma; n = (n-1)!, </NOBR> for int&nbsp;n&ge;1.
       &nbsp; Also, <NOBR> n! = &Gamma;(n+1), </NOBR> and
       &nbsp; <NOBR> &Gamma; (x+1) = x &Gamma; x. </NOBR>
       &nbsp; <NOBR>See
       [<A HREF="http://www.allisons.org/ll/AlgDS/Numerical/Stirling/"
       TARGET="LL">www.allisons.org/ll/AlgDS/Numerical/Stirling/</A>]</NOBR>
       and {@link #logFactorial logFactorial}.  */
   public static double logGamma(double x)                     // requires x>=0
    { if( x < 0.5 ) // Euler's "reflection": G(1-z) G(z) = pi/sin(pi z) ...
       { if( x <= 0 ) error( "logGamma(x="+x+"), x<=0" ); // else...
         return Math.log(Math.PI) -Math.log(Math.sin(Math.PI*x)) -logGamma(1-x);
       }
    //else, x >= 0.5
      final double g = 7;
      // Coefficients used by the GNU Scientific Library
      final double [] p = new double []
       {    0.99999999999980993, 676.5203681218851,
        -1259.1392167224028,     771.32342877765313, -176.61502916214059,
           12.507343278686905,    -0.13857109526572012, 9.9843695780195716e-6,
            1.5056327351493116e-7 };
      x -= 1;
      double sum = p[0];
      for( int i = 1; i < g+2; i++ )
         sum += p[i] / (x+i);
      double t = x + g + 0.5;
      return 0.5*Math.log(2*Math.PI) +(x+0.5)*Math.log(t) -t +Math.log(sum);
    }//logGamma(x)
   // -------------------------------------------------------------------------

   /** The log Beta (&beta;) function, log(&Beta;(x,y)), returns
   <NOBR>{@link #logGamma log&Gamma;}(x)+log&Gamma;(y)-log&Gamma;(x+y)</NOBR>.
       For <NOBR>x, y &gt; 0,</NOBR>
       <NOBR> B(x,y) = <SUB>0</SUB>&int;<SUP>1</SUP>
       t<SUP>x-1</SUP> (1-t)<SUP>y-1</SUP> dt.</NOBR> */
   public static double logBeta(double x, double y)
    { assert x > 0 && y > 0 : "x=" + x + ", y=" + y;
      return logGamma(x) + logGamma(y) - logGamma(x+y);
    }//logBeta(x,y)
   // -------------------------------------------------------------------------

   /** I<SUB>&alpha;</SUB>(x), the modified Bessel function of the 1st kind;
       <NOBR>x &ge; 0</NOBR>, and the order <NOBR>&alpha; &ge; 0</NOBR>,
       are reals here.
       Also see [<A HREF="http://en.wikipedia.org/wiki/Bessel_function"
                    TARGET="EX">wikip</A>] ...
       Note, the derivative
       <NOBR> I'<SUB>&nu;</SUB>(z) =
              (I<SUB>&nu;-1</SUB>(z) + I<SUB>&nu;+1</SUB>(z)) / 2; </NOBR>
       <NOBR> I'<SUB>0</SUB> = I<SUB>1</SUB>(z). </NOBR>
       Also note,
       <NOBR> I(a,x) / I(a-1,x) = 1 / (a/x + 1/((a+2)/x + ...)), </NOBR>
       a continued fraction (Watson1995).  */
   public static double Bessel_I(double alpha, double x)
    { if( alpha < 0 || x < 0 ) error( "Bessel I_[" + alpha + "](" + x + ")" );
      return Bessel_I_series(alpha, x);
    }//Bessel_I(a,x)

   /** Uses {@link #Bessel_I_series Bessel_I_series(,)}
       for "small" x, {@link #Bessel_I_asymptotic Bessel_I_asymptotic(,)}
       for "large" x, and both at the cross-over (continuity). */
   public static double Bessel_I2(double alpha, double x)
    { if( alpha < 0 || x < 0 ) error( "Bessel I_[" + alpha + "](" + x + ")" );
      final double lo = Math.sqrt(14*(14+alpha)),  // >= 14
                   hi = lo+1;
      if( x <= lo ) return Bessel_I_series(alpha, x);      // else...
      if( x >= hi ) return Bessel_I_asymptotic(alpha, x);  // else...
      return ( (hi-x)*Bessel_I_series(alpha, x)            // x in (lo, hi)
             + (x-lo)*Bessel_I_asymptotic(alpha, x) ) / (hi - lo);
      // the interpolation above is to keep the function continuous
    }//Bessel_I2(a,x)

   /** For "small" x, uses the series expansion.
       Is accurate, but slow for large x. */
   public static double Bessel_I_series(double alpha, double x)
    { final double                // note, alpha >= 0 and x >= 0 ...
       t = alpha == 0 ? 1.0                    // else alpha > 0 ...
         :     x == 0 ? 0.0                    // else     x > 0 ...
         : Math.exp(alpha * Math.log(x/2.0));  // i.e., (x/2)^alpha
      final double x2_4 = x*x*0.25;
      double R = 1.0,             // 0-th term
             I = 1.0,             // m=0 sum
             m = 1.0;             // next, m=1
      // m! = m*(m-1)!, and
      // Gamma(m+alpha+1) = (m+alpha)*Gamma(m+alpha), truly
      // because Gamma(x+1) = x*Gamma(x), for all x > 0.
      do{ R *= x2_4 / (m*(alpha+m));  // the m-th term
          if( R < java.lang.Math.ulp(I) ) break; // ulp(I) = dist to next dbl
          I += R;                     // total
          m += 1.0;
        } while( true );
      return I * t / Gamma(alpha+1);
    }//Bessel_I_series(a,x)

   /** For "large" x, uses Hankel's asymptotic expansion.  See
       [<A HREF="http://people.math.sfu.ca/~cbm/aands/page_377.htm">www</A>]
       and [<A HREF="http://dlmf.nist.gov/10.40">www</A>]. */
   public static double Bessel_I_asymptotic(double alpha, double x)
    { if( x < alpha ) error( "alpha=" + alpha + ", x=" +x );
      final double mu = alpha * alpha * 4,  inv_8x = 1.0 / (x*8);
      double a_k = 1, sum = 1;
      for( int k = 1; k < 50; k++ )
       { final int two_k_1 = 2*k-1;  // i.e., 1, 3, 5, 7, ...
         a_k *= - (mu - two_k_1*two_k_1) * inv_8x / k;
         if( Math.abs(a_k) < java.lang.Math.ulp(sum) ) break;
         sum += a_k;
       }
      return (Math.exp(x) / Math.sqrt(twoPI * x)) * sum;
    }//Bessel_I_asymptotic(a,x)
   // -------------------------------------------------------------------------

   public static void main(String[] args)
    { System.out.println("-- test Maths --");

      System.out.println("-- see la.maths.Test --");

      System.out.println("-- done Maths --");
    }//main

 }//Maths class   http://www.allisons.org/ll/

// (c) Lloyd Allison, Oct 2011, Wed 1 Feb 2017. Licence to 'Maths' is Gnu GPL.
