//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package la.util;

import java.util.Comparator;

import la.la.Value;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** A few general purpose constants and utility functions/methods.
    Also see {@link README README}.  */
public class Util
 {
   /** Return the String "error, " + msg; see {@link #error error(msg)}. */
   private static String errMsg(String msg)                      // errMsg(msg)
    { return "error, " + msg; }

   /** Convenience function for new RuntimeException(msg). */
   public static RuntimeException RTE(String msg)                   // RTE(msg)
    { return new RuntimeException(errMsg(msg)); }

   /** Throw a RuntimeException, {@link #RTE RTE(msg)}. */
   public static void error(String msg)                           // error(msg)
    { throw RTE(msg); }
   /* ---------------------------------------------------------------------- */

   /** Render a double, x, compactly and to at least three
       significant digits; also see {@link #x2String(double,int)}. */
   public static String x2String(double x)
    { return x2String(x, 3); }

   /** Render a double, x, as compactly as possibly while
       guaranteeing 'sigDigits' significant digits. */
   public static String x2String(double x, int sigDigits)
    { sigDigits = Math.max(sigDigits, 1);
      String F = ".0", G = "0.", H = "0.";
      for( int i = 1 ; i < sigDigits; i ++ ) // F        G        H
       { F += "#"; G += "#"; H = "#"+H; }    // .0###    0.0##    ###0.
      F += "E0"; G += "E0";                  // .0###E0  0.0##E0
      double xp = x >= 0 ? x : -x;
      String xs = new java.text.DecimalFormat( F ).format(xp);
      int ep  = xs.indexOf("E");
      if( ep < 0 || ep >= xs.length() ) // just paranoid
         return new java.text.DecimalFormat(G).format(x);
      int exp = new Integer(xs.substring(ep+1)).intValue();  // exponent
// System.out.println("{x="+x + ", xs="+xs + ", ep="+ep + ", exp="+exp + "}");
      if( exp > sigDigits + 2 )
         return new java.text.DecimalFormat(G).format(x);  // D.dddED
      else if( exp > sigDigits )
         return new java.text.DecimalFormat(H).format(x);  // DDDD..
      if( exp >= 0 &&  exp <= sigDigits )
       { String I = "";
         for( int i = 0; i < exp; i ++ ) I = "#" + I;      // ##.
         I += ".";
         for( int i = exp; i < sigDigits; i ++ ) I += "#"; // ##.##
         return new java.text.DecimalFormat(I).format(x);  // dd.dd
       }
      else if( exp >= -3 )
       { String I = ".";
         for( int i = 0; i < -exp; i ++ ) I += "0";
         I += F.substring(1, sigDigits+1);
         return new java.text.DecimalFormat(I).format(x);
       }
      else return new java.text.DecimalFormat(G).format(x);
    }//x2String(x,sf)

   /** Binary search 'ns[]' for 'n'; return n's location or else -1.
       Also see {@link #search search(ns,n)}.   */
   public static int location(int [] ns, int n)
    { int lo = 0, hi = ns.length;
      if( hi < 1 ) return -1; // else...
      while( lo < hi - 1 )
       { final int mid = (lo + hi) / 2;  // lo < mid < hi
         if( n >= ns[mid] ) lo = mid;
         else/* < */        hi = mid;
       }//lo == hi-1
      return ns[lo] == n ? lo : -1;
    }//search(ns,n)

   /** {@link #location Binary search} 'ns[]' for 'n'.
       Is its {@link #location location} within the array?  */
   public static boolean search(int [] ns, int n)
    { return location(ns, n) >= 0; }
   // -------------------------------------------------------------------------

   /** Table for use by interleave16(m,n) below. */
   private final static short [] MortonTable256 = new short []
    { 0x0000, 0x0001, 0x0004, 0x0005, 0x0010, 0x0011, 0x0014, 0x0015,
      0x0040, 0x0041, 0x0044, 0x0045, 0x0050, 0x0051, 0x0054, 0x0055,
      0x0100, 0x0101, 0x0104, 0x0105, 0x0110, 0x0111, 0x0114, 0x0115,
      0x0140, 0x0141, 0x0144, 0x0145, 0x0150, 0x0151, 0x0154, 0x0155,
      0x0400, 0x0401, 0x0404, 0x0405, 0x0410, 0x0411, 0x0414, 0x0415,
      0x0440, 0x0441, 0x0444, 0x0445, 0x0450, 0x0451, 0x0454, 0x0455,
      0x0500, 0x0501, 0x0504, 0x0505, 0x0510, 0x0511, 0x0514, 0x0515,
      0x0540, 0x0541, 0x0544, 0x0545, 0x0550, 0x0551, 0x0554, 0x0555,
      0x1000, 0x1001, 0x1004, 0x1005, 0x1010, 0x1011, 0x1014, 0x1015,
      0x1040, 0x1041, 0x1044, 0x1045, 0x1050, 0x1051, 0x1054, 0x1055,
      0x1100, 0x1101, 0x1104, 0x1105, 0x1110, 0x1111, 0x1114, 0x1115,
      0x1140, 0x1141, 0x1144, 0x1145, 0x1150, 0x1151, 0x1154, 0x1155,
      0x1400, 0x1401, 0x1404, 0x1405, 0x1410, 0x1411, 0x1414, 0x1415,
      0x1440, 0x1441, 0x1444, 0x1445, 0x1450, 0x1451, 0x1454, 0x1455,
      0x1500, 0x1501, 0x1504, 0x1505, 0x1510, 0x1511, 0x1514, 0x1515,
      0x1540, 0x1541, 0x1544, 0x1545, 0x1550, 0x1551, 0x1554, 0x1555,
      0x4000, 0x4001, 0x4004, 0x4005, 0x4010, 0x4011, 0x4014, 0x4015,
      0x4040, 0x4041, 0x4044, 0x4045, 0x4050, 0x4051, 0x4054, 0x4055,
      0x4100, 0x4101, 0x4104, 0x4105, 0x4110, 0x4111, 0x4114, 0x4115,
      0x4140, 0x4141, 0x4144, 0x4145, 0x4150, 0x4151, 0x4154, 0x4155,
      0x4400, 0x4401, 0x4404, 0x4405, 0x4410, 0x4411, 0x4414, 0x4415,
      0x4440, 0x4441, 0x4444, 0x4445, 0x4450, 0x4451, 0x4454, 0x4455,
      0x4500, 0x4501, 0x4504, 0x4505, 0x4510, 0x4511, 0x4514, 0x4515,
      0x4540, 0x4541, 0x4544, 0x4545, 0x4550, 0x4551, 0x4554, 0x4555,
      0x5000, 0x5001, 0x5004, 0x5005, 0x5010, 0x5011, 0x5014, 0x5015,
      0x5040, 0x5041, 0x5044, 0x5045, 0x5050, 0x5051, 0x5054, 0x5055,
      0x5100, 0x5101, 0x5104, 0x5105, 0x5110, 0x5111, 0x5114, 0x5115,
      0x5140, 0x5141, 0x5144, 0x5145, 0x5150, 0x5151, 0x5154, 0x5155,
      0x5400, 0x5401, 0x5404, 0x5405, 0x5410, 0x5411, 0x5414, 0x5415,
      0x5440, 0x5441, 0x5444, 0x5445, 0x5450, 0x5451, 0x5454, 0x5455,
      0x5500, 0x5501, 0x5504, 0x5505, 0x5510, 0x5511, 0x5514, 0x5515,
      0x5540, 0x5541, 0x5544, 0x5545, 0x5550, 0x5551, 0x5554, 0x5555
    };  // e.g., 1111,1111_2 -> 0101,0101,0101,0101_2

   /** Return the lower 16 bits of 'm' and of 'n' interleaved.
       Any higher bits of m and n are ignored.  (See
       [<A HREF="http://graphics.stanford.edu/~seander/bithacks.html"
           TARGET="EX">www</A>]['15].)   */
   public static int interleave16(int m, int n)            // interleave16(m,n)
    { return MortonTable256[(m & 0xFF00) >> 8] << 17 |
             MortonTable256[(n & 0xFF00) >> 8] << 16 |
             MortonTable256[ m &   0xFF      ] <<  1 |
             MortonTable256[ n &   0xFF      ];
    }//interleave16(m,n)
   // -------------------------------------------------------------------------

   public static void main(String[] args)
    { System.out.println("-- test Util --");

      System.out.println("-- see la.util.Test --");

      System.out.println("-- done Util --");
    }//main

 }//Util class   http://www.allisons.org/ll/

// (c) Lloyd Allison, Oct 2011, Tue 13 Dec 2016. Licence to 'Util' is Gnu GPL.
