//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised <I>BestOf</I> Model &ndash; choose the best one
    out of a given Tuple of alternative UnParameterised Models.
    The alternatives are put in {@link #upms upms[]};
    all are considered to be equally likely apriori.
    All members of upms[] must of course be Models of the same kind of data,
    but their sufficient statistics and the parameters of their estimators
    and their parameterised Models can differ, one from another.
    The fully parameterised BestOf-Model is {@link M}. */
public class BestOf extends UPModel                                   // BestOf
 {
   /** The alternative UnParameterised Models from which 'this' BestOf
       is to choose just one in estimating {@link M}. */
   public final UPModel [] upms;

   /** Nupms = upms.length. It is used a lot.  */
   private final int Nupms;

   /** The problem definition parameter 'upms' is a Tuple of
       UnParameterised Models; saved as {@link #upms upms[]}.  */
   public BestOf(Value upms)
    { super(upms);
      Tuple upmsT = (Tuple) upms;
      this.Nupms  = upmsT.nElts();
      this.upms   = new UPModel [Nupms];
      for( int i = 0; i < this.Nupms; i ++ )
         this.upms[i] = (UPModel) upmsT.elt(i);
    }//constructor

   /** Given first- and second-part message lengths, msg1 and msg2, and
       statistical parameter(s), sp, return a fully parameterised
       {@link M} Model. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); } // for the sake of the type

   /** Calculate the  sufficient statistics of ds[lo,hi) for
       <I>all</I> possible alternatives {@link #upms upms[.]}.
       Return a Tuple of these sufficient statistics.
       Note that {@link M#stats M.stats(...)} is simpler because
       one of the alternatives has been chosen in&nbsp;M.  */
   public Value.Tuple stats(Vector ds, int lo, int hi)
    { Value [] ss = new Value [Nupms];
      for( int i = 0; i < Nupms; i ++ )
         ss[i] = upms[i].stats(ds, lo, hi);
      return Value.tuple(ss);
    }//stats(ds,lo,hi)

   /** Combine statistics ss0 and ss1, either ss0 "+" ss1,
       or ss0 "-" ss1, depending on 'add' being true or false.
       Return a Tuple of these sufficient statistics.  */
   public Value.Tuple stats(boolean add, Value ss0, Value ss1)
    { Value [] ss = new Value [Nupms];
      Tuple ss0T = ss0.isTuple(Nupms), ss1T = ss1.isTuple(Nupms);
      for( int i = 0; i < Nupms; i ++ )  // for each upms[i] ...
         ss[i] = upms[i].stats(add, ss0T.elt(i), ss1T.elt(i));
      return Value.tuple(ss);
    }//stats(add,ss0,ss1)

   /** Return an Estimator of a fully parameterised {@link M}
       Model (essentially pick the best of the {@link #upms upms[]}).
       Estimator's parameter 'ps' must be a Tuple having one element
       (possibly trivial) per {@link #upms upms[.]}.estimator(). */
   public Est estimator(Value ps)
    { return new Est(ps)
       { Tuple psT = ps.isTuple(Nupms);
         /** Given sufficient statistics, ss, of a data-set,
             return a fully parameterised {@link BestM} Model. */
         public M ss2Model(Value ss)
          { Tuple ssT = ss.isTuple(Nupms);
            Model  best     = null;
            int    best_i   = -1;
            double best_msg = Double.MAX_VALUE;
            for( int i = 0; i < Nupms; i ++ )
             { Model m = upms[i].estimator(psT.elt(i)).ss2Model(ssT.elt(i));
               double msg = m.msg();
               if( m.msg() < best_msg )
                { best = m;  best_i = i;  best_msg = msg; }
             }//for i
            return new M( Math.log(Nupms) + best.msg1(),
                          best.msg2(),
                          pair(new Int(best_i), best.statParams()) );
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(ps)

   // .........................................................................

   /** The fully parameterised Model;
       {@link BestOf} is the UnParameterised Model.  */
   public class M extends UPModel.M                                 // BestOf.M
    {
      /** The index of the best of the {@link #upms upms[]} (parameterised). */
      public int   choice;

      /** The fully parameterised upms[{@link #choice choice}]-Model,
          the best choice out of the {@link #upms upms[]}.  */
      public Model mdl;

      /** Two part message lengths, msg1 and msg2,
          and statistical parameter(s) sp=&lang;i,sp<SUB>i</SUB>&rang;.
          'i'&nbsp;is the index of the best Model out of {@link #upms umps[]}.
          'sp<SUB>i</SUB>'&nbsp;is the statistical parameter(s)
          for that Model.   */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple spT = sp.isTuple(2);
         choice = spT.fst().n();
         mdl    = upms[choice].apply(spT.snd());
       }//constructor

      /** Return {@link #mdl mdl}'s negative log probability of data 'd'.  */
      public double nlPr(Value d)  { return mdl.nlPr(d);  }

      /** Return {@link #mdl mdl}'s stats for ds[lo,hi).
          Also see {@link BestOf#stats BestOf.stats(...))}
          which is more extensive.  */
      public Value stats(Vector ds, int lo, int hi)
       { return mdl.stats(ds, lo, hi); }

      /** Return {@link #mdl mdl}'s stats ss0&plusmn;ss1,
          "+"&nbsp;if add is true otherwise&nbsp;"-".
          Also see {@link BestOf#stats BestOf.stats(...))}.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return mdl.stats(add, ss0, ss1); }

      /** Given sufficient statistics, ss=stats(ds), of a data-set ds,
          return {@link #mdl mdl}'s negative log LikeliHood of ds. */
      public double nlLH(Value ss) { return mdl.nlLH(ss); }

      /** Return a Value from {@link #mdl mdl}. */
      public Value random()        { return mdl.random(); }

    }//BestOf.M class, the fully parameterised Model

 }//BestOf class, the UPModel

// (c) Lloyd Allison Sun 6 Aug 2017. Licence to 'BestOf' is Gnu GPL.
