//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** CPT -- Conditional Probability Table -- an (UnParameterised)
    FunctionModel, <I>from</I> an input datum with a working
    {@link #n() n()}, such as a {@link la.la.Value.Discrete} or a
    {@link la.la.Value.Tuple Tuple} of Discretes, <I>to</I> a Model of the
    output datum. The fully parameterised Model is {@link CPT.M}.
    In the simplest situation the output datum is modelled by
    {@link MultiState} Models but there is greater generality. */
public class CPT extends UPFunctionModel                                 // CPT
 {
   /** lwb, upb, the bounds on the input datum. Note, the
       bounds and the input datum have the same Type which
       may even be Tuple of Discrete Values.  */
   public final Value lwb, upb;

   /** lwb_n, upb_n, the bounds on the input datum, as <I>ints</I>. */
   public final int lwb_n, upb_n;

   /** The UnParameterised Model to be {@link M#condMdls parameterised}
       for each case (Value) of the input datum.  */
   public final UPModel upm;

   /** Given problem-definition parameters,
       dp&nbsp;=&nbsp;(lwb,upb,upm), construct an (UnParameterised) CPT
       with (upb_n&nbsp;-&nbsp;lwb_n&nbsp;+&nbsp;1) entries, each entry
       being (eventually) an <I>fully parameterised</I> upm-Model. */
   public CPT(Value dp)
    { super(dp);
      Value.Tuple l_u_upm = dp.isTuple(3);  // (lwb, upb, upm)
      lwb = l_u_upm.fst();  upb = l_u_upm.snd();
      upm = (UPModel) l_u_upm.thrd();
      if( ! lwb.type().ni(upb) ) error( " bounds " + lwb + "!~" + upb );
      lwb_n = lwb.n();  upb_n = upb.n();
    }//constructor

   /** Given a data-set, ds, calculate sufficient statistics
       (of the output datum) for each possible case (Value) of
       the input datum.  Uses {@link #upm}.stats(.,.,.).
       (TODO: Implementation is multi-pass ELSE need INCREMENTAL
       suff stats. Which IS Possible now!)
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { Value [] sss = new Value [upb_n - lwb_n + 1];  // statisticses
      Vector ids = ds.col(0), ods = ds.col(1);  // ?check cols=2?
      double [] wt = new double [ ds.nElts() ]; // OK, for now???
      for( int id_n = lwb_n; id_n <= upb_n; id_n ++ ) // each case
       { for( int j = lo; j < hi; j ++ )
            wt[j] = ids.elt(j).n() == id_n ? 1 : 0;  // in | out
         sss[id_n - lwb_n] = upm.stats( ods.weight(wt), lo, hi );
       }//for id_n
      return Vector.values(sss);
    }//stats(ds,lo,hi)

   /** Calls upon {@link #upm upm}.stats(add,.,.).  */
   public Vector stats(boolean add, Value ss0, Value ss1)
    { Vector ss0v = (Vector) ss0, ss1v = (Vector) ss1;
      Value [] sss = new Value [upb_n - lwb_n + 1];  // statisticses
      for( int i = 0; i < sss.length; i ++ ) // each case
         sss[i] = upm.stats(add, ss0v.elt(i), ss1v.elt(i));
      return Vector.values(sss);
    }//stats(add,ss0,ss1)

   /** Given sps, a Vector of statistical parameters, one for each
       input Value (that is, one for each entry of the CPT), return
       a fully parameterised {@link M CPT Model}.  */
   public M sp2Model(double msg1, double msg2, Value sps)
    { return new M( msg1, msg2, (Vector)sps ); }

   /** Return an Estimator of fully parameterised CPTs -- of
       {@link M}. The parameter, 'ps' (e.g., on a prior),
       is passed to the per-case-Estimator,
       {@link #upm upm}.{@link UPModel#estimator estimator(ps)}. */
   public Est estimator(final Value ps)                        // estimator(ps)
    { return new Est(ps)
       { Estimator est = CPT.this.upm.estimator(ps);
         /** Given sufficient statistics, {@link #stats ss}, return a
             {@link M}. Note ss is a Vector of stats, one per case. */
         public M ss2Model(Value ss)
          { Vector sss = (Vector) ss;  // stats x cases
            Value [] sps = new Value [upb_n - lwb_n + 1];
            double m1 = 0, m2 = 0;
            for( int i = 0; i < sps.length; i ++ )  // for each case...
             { Model mi = est.ss2Model(sss.elt(i)); // use 'est'
               m1 += mi.msg1(); m2 += mi.msg2();
               sps[i] = mi.statParams();
             }
            return new M(m1, m2, Vector.values(sps));
          }//ss2Model(ss)
       };//new Estimator
    }//estimator(ps)
   // .........................................................................

   /** CPT.M, a <I>fully parameterised</I> Function Model, being a
       conditional probability table. The UnParameterised
       Function Model is {@link CPT}.  */
   public class M extends UPFunctionModel.M                            // CPT.M
    {
      /** An array of now fully parameterised (conditional)
          {@link #upm upm}-Models, one Model for each possible
          case (Value) of the input datum. */
      public final Model [] condMdls;

      /** Given (m1,m2,sps), where sps is a Vector of
          statistical parameters, one for each entry of
          the CPT, construct the fully parameterised CPT.
          Also see {@link #condMdls}.  */
      public M(double msg1, double msg2, Vector sps)
       { super(msg1, msg2, sps);
         if( sps.nElts() != upb_n-lwb_n+1 )
            error( " the # of stat params, " + sps.nElts() );
         condMdls = new Model[upb_n - lwb_n + 1];
         for( int i = lwb_n; i <= upb_n; i ++ )
            condMdls[i] = upm.sp2Model(0 /* ! */, 0 /* ! */, sps.elt(i));
       }//constructor

      /** Return the {@link #condMdls Model} of the ouput datum conditional
          upon the Value of the input datum (variable), 'id'.
          The kind of Model that can returned is arbitrary, as specified
          by {@link CPT}; it is not restricted to MultiState Models.  */
      public Model condModel(Value id)
       { if( ! lwb.type().ni(id) )
            error( ".condModel(" + id + "), types" );
         return condMdls[ id.n() - lwb_n ];
       }//condModel(id)

      /** Given a data-set, ds, calculate statistics
          (of the output datum) for each case (Value) of the
          input datum.  Uses {@link #condMdls}[.]stats(.,.,.,.,.).
          More on stats {@link UPModel#stats here}.  */
      public Vector stats(Vector ds, int lo, int hi)
       { if( ds.nElts() > 0 &&  ds.nCols() != 2 )        // hmmm?
            error( "ds has " + ds.nCols() + " columns" );
         final Vector ids = ds.col(0), ods = ds.col(1);  // ?check cols=2?
         final Value [] sss = new Value [upb_n - lwb_n + 1];  // statisticses
         for( int j = 0; j <= upb_n - lwb_n; j ++ )
            sss[j] = condMdls[j].stats(ods, 0, 0);  // !
         for( int i = lo; i < hi; i ++ )
          { final int theCase = ids.elt(i).n() - lwb_n; // >= 0
            sss[theCase] =
               condMdls[theCase].stats(true, sss[theCase], ods, i, i+1);
          }//for i
         return Vector.values(sss);
       }//stats(ds,lo,hi)

      /** Calls upon {@link #condMdls condMdls}[.].stats(add,.,.).  */
      public Vector stats(boolean add, Value ss0, Value ss1)
       { Vector ss0v = (Vector) ss0, ss1v = (Vector) ss1;
         Value [] sss = new Value [upb_n - lwb_n + 1];  // statisticses
         for( int i = 0; i < sss.length; i ++ ) // each case
            sss[i] = condMdls[i].stats(add, ss0v.elt(i), ss1v.elt(i));
         return Vector.values(sss);
       }//stats(add,ss0,ss1)

      /** Given sufficient statistics,
          ss&nbsp;=&nbsp;{@link CPT#stats stats(ds)}, of a
          data-set, ds, return the negative log likelihood of ds. */
      public double nlLH(Value ss)
        { double sum = 0;
          for( int i = lwb_n; i <= upb_n; i ++ )
             sum += condMdls[i].nlLH( ((Vector) ss) .elt(i) );
          return sum;
        }//nlLH(ss)

    }//CPT.M, the (CPT-) FunctionModel class

 }//CPT class, the UPFunctionModel

// (c) Lloyd Allison, 2012, Fri 8 Dec 2017. Licence to 'CPT' is Gnu GPL.
