//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** Continuous may be used to define (UnParameterised) Models over
    {@link la.la.Value.Cts Cts} Values by specifying a (-log) probability
    density function, {@link Continuous.M#nlPdf_x}, on a Value's 'x()'. */
public abstract class Continuous extends ByPdf                  // * Continuous
 { public Continuous(Value dp)
     { super(dp); }//constructor

   public abstract M sp2Model(double msg1, double msg2, Value sp);         // *
   // ............................................................

   /** The (abstract) class of fully parameterised Continuous Models. */
   public abstract class M extends ByPdf.M                    // * Continuous.M
    { public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** The negative log probability density of x;
          also see {@link #nlPdf}. */
      public abstract double nlPdf_x(double x);                 // * nlPdf_x(x)

      /** Calls <TT>{@link #nlPdf_x nlPdf_x}(d.x())</TT>. */
      public double nlPdf(Value d)
       { return nlPdf_x( ((Value.Cts) d) .x() ); }

      /** Call {@link #random_x()}, if implemented, to return
          an <I>exact</I> random {@link la.la.Value.Real}; it might be
          necessary to do something different if, for example,
          you want accuracy (AoM) to depend on the result's x().
          But if you want random Values with zero or other
          constant AoM, the present strategy is fine.  */
      public Value.Real random()
       { return Value.real( random_x() ); }

      /** Return a random
          {#link la.la.Value#cts Cts}({@link #random_x() random_x()},AoM). */
      public Value.Cts random(double AoM)
       { return Value.cts(random_x(), AoM); }

      /** Return a double[n] of {@link #random_x() random_x()}. */
      public double[] random_x(int n)
       { double[] ans = new double[n];
         for( int i = 0; i < n; i ++ ) ans[i] = random_x();
         return ans;
       }//random_x(n)

      /** Return n random Cts each with the specified AoM. */
      public Vector random(int n, double AoM)
       { return Vector.doubles(random_x(n), AoM); }

      /** Called by {@link #random()}; this default is
          not implemented (throws an Exception) but see
          {@link Continuous.Uniform.M#random_x()} and
          {@link NormalUPM.M#random_x()}, say.  Note, there is an issue
          about generating random Cts Values with non-zero AoM if AoM
          depends on&nbsp;x.  */
      public double random_x()                                  // ? random_x()
       { throw UOE( ".random_x() is not yet(?) supported, sorry" ); }

    }//Continuous.M (abstract) class, the Model
   // .........................................

   /** UnParameterised Bounded Continuous Models, over
       a range, <NOBR>[lwb, upb],</NOBR> of Cts Values. */
   public static abstract class Bounded extends Continuous      // * ...Bounded
    { public Bounded(Value dp)
       { super(dp); }//constructor
             /** Return (lwb, upb) of the data-space. */
      public abstract Tuple bounds();                             // * bounds()
             /** The lower bound, {@link #bounds() bounds()}.fst(). */
      public Cts    lwb()   { return (Cts) bounds().fst(); }
             /** The upper bound, {@link #bounds() bounds()}.snd(). */
      public Cts    upb()   { return (Cts) bounds().snd(); }
             /** The lower bound of the data-space as a double. */
      public double lwb_x() { return lwb().x(); }
             /** The upper bound of the data-space as a double. */
      public double upb_x() { return upb().x(); }
      // ............................

      /** Fully parameterised Bounded Continuous Models. */
      public abstract class M extends Continuous.M            // * ...Bounded.M
       { public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); }//constructor
             /** Bounded.this.{@link Bounded#bounds() bounds()}. */
         public Tuple bounds() { return Bounded.this.bounds(); }
             /** Bounded.this.{@link Bounded#lwb() lwb()}. */
         public Cts      lwb() { return Bounded.this.lwb(); }
             /** Bounded.this.{@link Bounded#upb() upb()}. */
         public Cts      upb() { return Bounded.this.upb(); }
             /** Bounded.this.{@link #lwb_x() lwb_x()}. */
         public double lwb_x() { return Bounded.this.lwb_x(); }
             /** Bounded.this.{@link #lwb_x() upb_x()}. */
         public double upb_x() { return Bounded.this.upb_x(); }
       }//Bounded.M (abstract) class, the Model

    }//Bounded (abstract) class, the UPModel
   // ......................................

   /** The "UnParameterised" Uniform Continuous Model on the range
       <NOBR>[lwb, upb]</NOBR>. Also see the "fully parameterised"
       {@link Uniform.M}, and the {@link Discretes.Uniform}.  */
   public static class Uniform extends Bounded            // Continuous.Uniform
    { /** The lower (upper) bound on the .x() of a datum. */
      private final double lwb_x, upb_x;
      /** The negative log probability density of a (every) data Value. */
      private final double nlPdf;

      /** Construct an "UnParameterised" Uniform Model (distribution) over
          the <NOBR> bounds = [lwb, upb]</NOBR>.  */
      public Uniform(Value bounds)
       { super(bounds.isTuple(2));
         this.lwb_x = lwb().x(); this.upb_x = upb().x();
         final double range = upb_x - lwb_x;
         this.nlPdf = Math.log(range);
       }//constructor

      public Tuple bounds() { return (Tuple) defnParams(); }
      public double lwb_x() { return lwb_x;  }
      public double upb_x() { return upb_x;  }

      /** Return a "fully parameterised" Uniform Continuous {@link M Model};
          note that the statistical parameter, t, is trivial. */
      public M sp2Model(double m1, double m2, Value t /* = triv */)
       { return new M(m1, m2, t); }

      /** Given a data-set, ds, return sufficient statistics, ss, that
          is the (weighted) number of elements and their nlAoM.
          More on stats {@link UPModel#stats here}.  */
      public Value.Tuple stats(Vector ds, int lo, int hi)
       { return Value.pair( Value.real(ds    .wts(lo,hi)),
                            Value.real(ds.wtNlAoM(lo,hi)) );
       }//stats(ds,lo,hi)

      public Value.Tuple stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
         return add
            ? Value.pair( Value.real(ss0t.fst().x() + ss1t.fst().x()),
                          Value.real(ss0t.snd().x() + ss1t.snd().x()) )
            : Value.pair( Value.real(ss0t.fst().x() - ss1t.fst().x()),
                          Value.real(ss0t.snd().x() - ss1t.snd().x()) );
       }

      /** Given sufficient statistics,
         <NOBR><TT>ss = {@link Poisson0UPM#stats stats(ds)}</TT>,</NOBR>
         of a data-set, ds, return the negative log likelihood of ds.
         The statistical parameter of {@link Uniform.M} is trivial,
         so nlLH does not depend on it and nlLH can be here in
         the UnParameterised Model. */
      public double nlLH(Value ss)
       { final Value.Tuple ssT = ss.isTuple(2);
         final double N = ssT.fst().x(), nlAoM = ssT.snd().x();
         return nlPdf * N + nlAoM;
       }//nlLH(ss)

      /** The trivial Estimator that is a Uniform's. */
      public Estimator estimator(final Value t /* = triv */)
       { return Mdl.asEstimator(t); }
      // ............................

      /** The fully parameterised Uniform continuous Model
          on <NOBR>[lwb, upb]</NOBR>.  */
      public final M Mdl = new M(0, 0, triv);

      /** {@link #Mdl} should be sufficient for many purposes.
          M, the class of fully- (trivially) parameterised Uniform
          continuous Model(s) on the range <NOBR>[lwb, upb]</NOBR>.  */
      public class M extends Bounded.M                  // Continuous.Uniform.M
       { /** requires msg1=0, sp=triv. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

         public double nlPdf_x(double d) { return nlPdf; }

         /** Return {@link Uniform}.{@link Uniform#nlLH nlLH(ss)}. */
         public double nlLH(Value ss)
          { return Uniform.this.nlLH(ss); }

         /** Called by {@link #random()}; note, the generated
             {@link la.la.Value.Cts} is exact, its AoM=0. */
         public double random_x()
          { return lwb_x + RNG.nextDouble() * (upb_x - lwb_x); }

       }//Continuous.Uniform.M, the Model

    }//Continuous.Uniform, the UPModel

 }//Continuous (abstract) class, Model

// (c) Lloyd Allison, 2012, Fri 31 January 2014.  Licence is Gnu GPL.
