//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** UnParameterised Models of Directions, that is of
    {@link R_D <B>R</B><SUP>D</SUP>}-Vectors, of <I>known</I>
    norm (length, size, radius).  Such a Vector has
    <NOBR>{@link #K() K} = {@link #D() D()} - 1</NOBR>
    degrees of freedom. Fully parameterised Direction Models
    start with {@link Direction.M} */
public abstract class Direction extends R_D                      // * Direction
 { public Direction(Value dp) { super(dp); }//constructor

   /** Degrees of freedom (surface dimension),
       <NOBR>K = {@link #D() D} - 1</NOBR>, on the
       unit-radius K-Sphere in <B>R</B><SUP>D</SUP>. */
   public int K() { return D() - 1; }                                    // K()

   // public abstract int D(),  is inherited from R_D                  // * D()
   // public Value stats(Vector ds), from R_D, but note nlAoM ~ normalising!

   /** Return the log of the surface-area of the unit-radius K-Sphere, the
       surface of the unit-radius D-Ball, where <NOBR>K = D - 1</NOBR>.
       For example, <NOBR>D=2, K=1, "1_area" = 2&pi;;</NOBR>
       and <NOBR>D=3, K=2, 2_area = 4&pi;.</NOBR> */
   public double logArea()                                         // logArea()
    { double halfK1 = D() / 2.0;  // K=D-1
      return log2 + halfK1*logPI - logGamma(halfK1);
      // NB. '+', return '+log Area'
      // e.g., D=2, K=1, halfK1=1,   area = 2 pi
      // e.g., D=3, K=2, halfK1=1.5, area = 4 pi
    }//logArea()

   /** (Abstract) a fully parameterised Model of Direction
       is made by defining {@link #nlPdf nlPdf(v)}.
       The UnParameterised Model is {@link Direction}.  */
   public abstract class M extends R_D.M                       // * Direction.M
    { public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** The negative log probability density of datum Vector&nbsp;v must
          be defined on the unit-radius K-sphere, the surface of a D-ball
          in <B>R</B><SUP>D</SUP>, {@link #K() D=K+1}. <B>Note</B>, it is
          most general if nlPdf(v) does not require v to be normalised.
          The given {@link #nlPr nlPr(v)} adjusts for v.norm().
          (If you do require v to be normalised, some other code will be
          doing that <I>and</I> you'll still have to check it anyway!) */
      public abstract double nlPdf(Value v); // anyway from R_D.M

      /** The negative log probability of datum Vector v; note that
          v need not be normalised but it is taken to be a Direction
          with v.norm() being "common knowledge".  That leaves
          {@link #K() K()} degrees of freedom.  nlPr returns
          <NOBR>{@link #nlPdf nlPdf(v)} + v.nlAoM()</NOBR>,
          <I>plus</I> <NOBR>K() * log(v.norm()).</NOBR>   E.g.,
          <NOBR>K = D-1 = 3-1 = 2,</NOBR> norm&nbsp;=&nbsp;8, implies
          <I>plus</I> 2&times;3 = 6&nbsp;bits. This adjusts for the norm. */
      public double nlPr(Value v)
       { Vector vv = (Vector) v;
         return nlPdf(v) + v.nlAoM() + K() * Math.log(vv.norm());
       }//nlPr(v)

   // public double nlLH(Value ss), from R_D.M, note nlAoM & normalising!

    }//Direction.M abstract class, the Model
   // ......................................

   /** The UnParameterised Uniform Model of Directions in
       <B>R</B><SUP>D</SUP>. It just needs the problem-defining
       parameter, D, to be set. The fully parameterised Model is
       {@link Uniform#Mdl}, of class {@link Uniform.M Uniform.M}. */
   public static class Uniform extends Direction           // Direction.Uniform
    { /** The dimension of <B>R</B><SUP>D</SUP>; see {@link #D() D()}. */
      public final int D;

      /** Given dimension, D, construct a Uniform Direction UPModel. */
      public Uniform(Value D)
       { super(D);
         this.D = ((Value.Int) D) .n();
       }//constructor

      /** The dimension of <B>R</B><SUP>D</SUP>; return {@link #D D}. */
      public int D() { return D; }

      /** sp2Model(0, m2, ()) returns a {@link M M Model}. */
      public M sp2Model(double m1, double m2, Value t /* = triv */)
       { return new M(m1, m2, t); }

      /** The default sufficient statistics <NOBR>ss = stats(ds)</NOBR> of
          a data-set ds; <NOBR>ss = ds</NOBR> itself. See {@link M#nlLH}.
          More on stats {@link UPModel#stats here}. */
      public Value stats(Vector ds, int lo, int hi) { return ds.slice(lo,hi); }

      public Value stats(boolean add, Value ss0, Value ss1)
       { return ((Vector) ss0).combine(add, (Slice) ss1); }

      /** The trivial Estimator that is a Uniform Direction's. */
      public Estimator estimator(final Value t /* = triv */)
       { return Mdl.asEstimator(t); }

      /** <I>The</I> (trivially) fully parameterised
          Uniform Model of Directions. */
      public final M Mdl = new M(0, 0, triv);
      // ..............................

      /** {@link #Mdl} should be sufficient for many purposes, but here is
          M, the class of fully parameterised Uniform Direction Models.
          The UnParameterised Model is {@link Uniform Uniform}. */
      public class M extends Direction.M                 // Direction.Uniform.M
       { /** Requires msg1=0, sp=triv. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, triv); zeroTriv(msg1, sp); }//constructor

         /** The negative log probability density,
             that is {@link #logArea logArea}. */
         public double nlPdf(Value v)
          { Vector vv = (Vector) v;
            if( vv.nElts() != D() )
               error( ".nlPdf(v=" + v + "), v\'s dimension !=" + D() );
            return logArea();  // i.e. 1/area of sphere of r=1.
          }//nlPdf(v)

         // public double nlPr(v), from Direction.M

         // public double nlLH(Value ss), use the default nlLH, for now?
         /** Assumes that {@link Uniform#stats(Vector,int,int) statistics}
             are the data-set itself. */
         public double nlLH(Value ss)
          { return sumNlPr((Vector) ss); }

         /** Generate a random Direction, a random unit Vector in
             <B>R</B><SUP>D</SUP>. To do this, sample D components from
             the {@link MML#N01 N<SUB>0,1</SUB>} Normal distribution
             and normalise the result.  */
         public void random(double [] x)
          { double nrm = 0;
            for( int i = 0; i < D(); i ++ )
             { x[i] = N01.random_x();  nrm += x[i]*x[i]; }
            nrm = Math.sqrt(nrm);
            for( int i = 0; i < D(); i ++ ) x[i] /= nrm;
          }//random(x)

      };//Direction.Uniform.M class, the Model

    }//Direction.Uniform class, the UPModel

 }//Direction (abstract) class, the UPModel

// (c) Lloyd Allison 2012, Fri 27 Jan 2017. Licence to 'Direction' is Gnu GPL.
