//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;
import la.maths.Matrix;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised Dirichlet Model (probability distribution)
    for data from a {@link #K() K}-Simplex. It is often applied to
    the proportions (relative abundancies) of categories.
    <I>(An estimator is not yet implemented.)</I>  */
public class Dirichlet extends Simplex                             // Dirichlet
 { private final int k; // degrees of freedom

   /** K, the degrees of freedom in the <I>data</I>, that is
       one less than the {@link #D() dimension} of the data. */
   public Dirichlet(Value K)
    { super(K); this.k = ((Int) K).n(); }//constructor

   /** The degrees of freedom of the K-Simplex sub-space
       <NOBR> &sube; [0, 1]<SUP>K+1</SUP>. </NOBR>
       Also see {@link #D()}.  */
   public int K() { return k; }                                          // K()

   /** Return a "fully parameterised" Dirichlet {@link Dirichlet.M Model}. */
   public M sp2Model(double m1, double m2, Value alpha)
    { return new M(m1, m2, alpha); }

   /** The default sufficient statistics <NOBR>ss = stats(ds) = ds</NOBR>;
       maybe we can do better in the future?
       (???TODO: Looks like maybe [&sum; log d<SUB>i,0</SUB>, ...]
        would be the go, but beware any datum d=[0,...]???)
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo,hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   // TODO: NO Dirichlet Estimator as yet. Note that for max. LH
   // Ronning (1989) and Narayanan (1991) use Newton-Raphson methods.
   // .........................................................................

   /** The fully parameterised Dirichlet Model (probability distribution);
       the UnParameterised Model is {@link Dirichlet here}.
       The business of M is in {@link #nlPdf(Value) nlPdf(d)}.  */
   public class M extends Simplex.M                              // Dirichlet.M
    {
      /** M's statistical parameter; <NOBR> &alpha;<SUB>i</SUB> &gt; 0, </NOBR>
          <NOBR> 0 &le; i &lt; {@link #D() D}. */
      public final Vector alpha;

      private double [] alpha_1 = new double [D()];

      /** The (log(B)) normalising constant. */
      public final double logB;

      /** Note, the length of the statistical parameter {@link #alpha &alpha;}
          is <NOBR> {@link #D() D} = {@link #K() K} + 1. </NOBR>  */
      public M(double msg1, double msg2, Value alpha)
       { super(msg1, msg2, alpha);
         this.alpha = (Vector) alpha;
         if( this.alpha.nElts() != D() )
            error( ", alpha[" + this.alpha.nElts() + "]" );
         double sumA = 0, lnB = 0;
         for( int i = 0; i < D(); i ++ )  // ???
          { double A_i = this.alpha.elt(i).x();
            if( A_i <= 0 ) error( ", alpha[" + i + "]=" + A_i + "<=0" );
            sumA += A_i;
            alpha_1[i] = A_i - 1;
            lnB += logGamma(A_i);
          }
         logB = lnB - logGamma(sumA);
         // i.e., B = (PROD_i Gamma(A_i)) / Gamma(SUM_i A_i)
       }//constructor

      // NB. nlPr(d) is inherited and uses nlPdf(d) below

      /** <NOBR> pdf(d) = (&prod;<SUB>i</SUB>
                 d<SUB>i</SUB><SUP>&alpha;<SUB>i</SUB>-1</SUP>)
               / B(&alpha;) </NOBR>
          where datum 'd' is a Vector in [0,1]<SUP>D</SUP> such that
          <NOBR> ||d||<SUB>1</SUB> = 1. </NOBR>
          Hence 'd' has <NOBR>{@link #K() K} = {@link #D() D} - 1</NOBR>
          degrees of freedom, which has implications for d's
          {@link Vector#nlAoM nlAoM} (as for {@link Direction Direction}).
          Note, if <NOBR> {@link #alpha &alpha;}<SUB>i</SUB> &gt; 1 </NOBR>
          then a datum 'd' s.t. <NOBR> d<SUB>i</SUB> = 0 </NOBR> has
          pdf(d)=0 and nlPr(d)=&infin;. Also see {@link #nlPr nlPr(d)}. */
      public double nlPdf(Value d)                                  // nlPdf(d)
       { check(d);  // i.e., Simplex.check(d), dimension?, norm_1, etc.
         Vector dv = (Vector) d;
         double sum = logB;               // normalising const
         for( int i = 0; i < D(); i ++ )  // ???
          { final double a_1_i = alpha_1[i], term;
            if( a_1_i > 0 )
               // note, pr([...,0,...])=0, term=-inf, sum=inf, pr=0, ok???
               term = a_1_i * Math.log( dv.elt(i).x() );
            else if( a_1_i == 0 ) term = 0;  // x^0=1 forall x, even 0^0=1
            else // a_1_i < 0, alpha(i) < 1
             { double d_i = dv.elt(i).x();
               if( d_i == 0 )  // avoid a -inf -nlPdf, ie. a +inf pdf, ok???
                  d_i = Math.exp( - d.nlAoM() / K() ) * 0.5;  // NB. K, not D
               term = a_1_i * Math.log(d_i);  // avoid (< 0)*log 0 = +inf,
             }
            sum -= term;  // NB. '-='
          }
         return sum;
       }//nlPdf(v)

      /** Assumes that {@link #stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)                                  // nlLH(ss)
       { return sumNlPr((Vector) ss); }

      /** Cache for use by {@link #random() random()}; when filled
          <NOBR> gamma<SUB>i</SUB> =
                 {@link Gamma Gamma}(alpha<SUB>i</SUB>, 1). </NOBR> */
      private GammaUPM.M [] gamma = null;

      /** Generate a Dirichlet-distributed random variate, a Vector of
          {@link #D() D} elements. It makes use of {@link GammaUPM.M#random()}.
          (See [<A HREF="http://en.wikipedia.org/wiki/Dirichlet_distribution"
                   TARGET="EX">wikip</A>][10/2014].)  */
      public Vector random()                                        // random()
       { if( gamma == null )  // populate cache
          { gamma = new GammaUPM.M [D()];  // gamma[i] ~ Gamma(alpha[i], 1)
            for( int i = 0; i < gamma.length; i ++ )
               gamma[i] = Gamma.sp2Model(0, 0, pair(alpha.elt(i), one));
          }//if
         double sum = 0;
         double [] y = new double [D()];
         for( int i = 0; i < y.length; i ++ )
          { final double r = gamma[i].random_x();
            y[i] = r;
            sum += r;
          }//for i
         for( int i = 0; i <= k; i ++ ) y[i] /= sum;  // normalise (L_1)
         return Vector.doubles(y);
       }//random()

    }//Dirichlet.M class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** A very few, very rudimentary tests; also see {@link Test}. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run Dirichlet --");

      Dirichlet   DirU = new Dirichlet(two); // K=2, D=3, 3D-vectors
      Dirichlet.M
      Unif = (Dirichlet.M) DirU.apply(Vector.doubles(new double[]{1, 1, 1 })),
      DirH = (Dirichlet.M) DirU.apply(Vector.doubles(new double[]{.5,.5,.5})),
      Dir2 = (Dirichlet.M) DirU.apply(Vector.doubles(new double[]{2, 2, 2 }));

      Matrix ds_3D = Matrix.doubles(new double[][]{
         {0.98, 0.01, 0.01}, {0.01, 0.98, 0.01}, {0.01, 0.01, 0.98},
         {0.49, 0.49, 0.02}, {0.49, 0.02, 0.49}, {0.02, 0.49, 0.49},
         {1.0/3, 1.0/3, 1.0/3} }, /*AoM= */ 0.01 );
   // NB. a '0.0' component gives pdf=0, nlPr=inf, if the corr alpha_i > 1.

      mml.Test.exercise(Unif, ds_3D);  // all pdfs should = 2
      mml.Test.exercise(DirH, ds_3D);  // boundary probable, centre not
      mml.Test.exercise(Dir2, ds_3D);  // centre probable, boundary not

      System.out.println("   see mml.Test");

      System.out.println("-- done Dirichlet --");
    }//main

 }//Dirichlet class, the UnParameterised Model

// (c) Lloyd Allison, Oct 2014, 28 Oct 2014. Licence to 'Dirichlet' is Gnu GPL.
