//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;
import static la.util.Util.*;
import static la.maths.Maths.*;
import la.util.Series;
import la.la.Value;
import la.la.Value.Tuple;
import la.maths.Vector;

import graph.*;  // the business at hand
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The general <I>abstract</I> UnParameterised Model of the structure
    of {@link graph.Graph Graphs} having explicit {@link #gType() gType}
    and {@link #upmV() upmV} where the former is the Type (population)
    of Graphs and the latter is the Model of the number of Vertices, |V|.
    The (abstract) fully parameterised Model is {@link Graphs.M}.
    Also see {@link IndependentEdges}.
    Note that Vertex- or Edge-labels, if any, are not considered.
    (Also&nbsp;see package [{@link graph.README graph</A>}].)   */
public abstract class Graphs extends UPModel                        // * Graphs
 {
   /** Public face is {@link #gType() gType()}.
       Some constructor must set gType.  */
   protected graph.Type gType;

   /** The Graph-{@link graph.Type Type} being modelled, particularly
       {@link graph.Directed Directed} versus
       {@link graph.Undirected Undirected} Graphs.
       gType&nbsp;defines the population of Graphs being modelled. */
   public graph.Type gType() { return gType; }                       // gType()

   /** Public face of 'upmV' is {@link #upmV() upmV()}.
       Some constructor must set upmV.  */
   protected Discretes upmV;

   /** The UnParameterised Model of&nbsp;{@link graph.Graph#vSize |V|}.
       Note, graphs require |V|&ge;1, e.g., for
       {@link IndependentEdges.M#random() random()}.
       Also see {@link M#mdlV() mdlV()}. */
   public Discretes upmV() { return upmV; }                           // upmV()

    /** Return {@link #upmV}'s stats on |V| for ds[lo,hi).
        Calls upon stats(ds,lo,hi) of {@link #upmV}. Probably the
        same as {M#vStats M.vStats(.,.,.)} but could differ.  */
    protected Value vStats(Vector ds, int lo, int hi)
     { int [] Vs = new int[ds.nElts()];
       for( int i = lo; i < hi; i ++ )
          Vs[i] = ((Graph) ds.elt(i)) .vSize();
       return upmV.stats( Vector.ints(Vs) .weightLike(ds.slice(lo,hi)) );
     }//vStats(ds,lo,hi)

   /** Estimate a Model of |V|, from 'ds', a data-set of Graphs.
       Requires parameter 'ps' for {@link #upmV upmV}'s
       {@link #estimator estimator(ps)}. Also see {@link M#mdlV()}.  */
   public Discretes.M ds2mdlV(Value ps, Vector ds)            // ds2mdlV(ps,ds)
    { return (Discretes.M)
         upmV.estimator(ps).ss2Model( vStats(ds, 0, ds.nElts()) );
    }//ds2mdlV(s)

   /** Statistics are the data-set ds[lo,hi) itself.  */
   public Vector stats(Vector ds, int lo, int hi)
    { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Vector.Slice) ss1); }

   /** Constructor only for internal use;
       it does not set {@link #gType gType} or {@link #upmV upmV}. */
   protected Graphs(Value dp)
    { super(dp);
      gType = null;  upmV = null;  // it's abstract
    }//constructor
   // .........................................................................

   /** Fully parameterised Graphs Models; also see
       UnParameterised {@link Graphs}. An&nbsp;M has an
       explicit Model, {@link #mdlV() mdlV}, over |V|. */
   public abstract class M extends UPModel.M                      // * Graphs.M
     {
       /** Public face of 'mdlV' is {@link #mdlV() mdlV()}.
           Some constructor must set mdlV.  */
       protected Discretes.M mdlV;

       /** The fully parameterised Model of the number of Vertices,
           {@link graph.Graph#vSize |V|}.  Note that we require
           |V|&ge;1 of Graphs, e.g., for
           {@link IndependentEdges.M#random random()}.
           Also see {@link #upmV()}. */
       public Discretes.M mdlV() { return mdlV; }                     // mdlV()

       public M(double msg1, double msg2, Value sp)
        { super(msg1, msg2, sp); }//constructor

       /* Note, stats(Vector,int,int), stats(boolean,Value,Value)
          inherited from UPModel.M,  i.e., from Graphs. */

       /** Sum of {@link #nlPr nlPr}s &mdash; assumes that
           statistics 'ss' is the data-set of Graphs itself. */
       public double nlLH(Value ss)                        // Graphs.M.nlLH(ss)
          { return sumNlPr((Vector) ss); }

       /** Not implemented for Graphs.M; a subclass may do so. */
       public Graph random()
        { throw UOE( ".random() is not yet(?) supported, sorry" ); }

     }// Graphs.M class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** Extends {@link Graphs} to Models where there is also an
       explicit Model {@link #upmE upmE} on the existence of Edges.
       {@link GERfixed} and {@link GERadaptive} are useful special cases.  */
   public static abstract class IndependentEdges extends Graphs            // *
    {
      /** Public face is {@link #upmE() upmE()}.
          Some constructor must set upmE.  */
      protected Discretes.Bounded upmE;

      /** The UnParameterised {@link MultiState 2-state} Model over the
          <I>existence</I> (0/1, false/true) of possible Edges.
          Also see {@link M#mdlE() mdlE()}. */
      public Discretes.Bounded upmE() { return upmE; }                // upmE()

      /** Constructor only for use by {@link GERfixed} and the like;
          it does not set {@link #gType gType}, {@link #upmV upmV} or
          {@link #upmE upmE}. */
      protected IndependentEdges(Value dp)
       { super(dp);
         /* see Graphs */ upmE = null; // it's abstract
       }//constructor

      /** Estimate a Discretes.Bounded.M, of Edge <I>existence</I>
          (0/1, false/true) from 'ds', a data-set of Graphs.
          Requires parameter 'ps' for {@link #upmE upmE}'s
          {@link #estimator estimator(ps)}. Also see {@link M#mdlE()}.  */
      public abstract Discretes.Bounded.M ds2mdlE(Value ps, Vector ds);

      public Est estimator(final Value ps)
       { final Tuple psT = ps.isTuple(2);
         return new Est(ps)
          { public M ss2Model(Value /*stats*/ ss)
             { Vector ssV = (Vector) ss;  // i.e., ss is the data-set
               Model mdlV = ds2mdlV(psT.fst(), ssV);  // of |V|
               Model mdlE = ds2mdlE(psT.snd(), ssV);  // of Edges
               return IndependentEdges.this.sp2Model(
                         mdlV.msg1() + mdlE.msg1(),   // msg1
                         mdlV.msg2() + mdlE.msg2(),   // msg2
                         pair(mdlV.statParams(), mdlE.statParams()) );
             }//s2Model(ss)
          };//new Est(ps)
       }//estimator(ps)

      public abstract M sp2Model(double msg1, double msg2, Value sp);      // *
      // ......................................................................

      /** The general abstract fully parameterised Graph Model having
          both explicit {@link #mdlV mdlV} and {@link #mdlE mdlE}.
          {@link GERfixed.M} and {@link GERadaptive.M} are useful
          special cases of&nbsp;M.  */
      public abstract class M extends Graphs.M          // * IndependentEdges.M
       {
         /** The Model of Edge existence.
             Public face is {@link #mdlE() mdlE()}.
             Some constructor must set mdlE.  */
         protected Discretes.Bounded.M mdlE;

         /** The fully parameterised Model of the existence of Edges.  It
             embodies the edge density of Graphs. Also see {@link #upmE()}. */
         public Discretes.Bounded.M mdlE() { return mdlE; }           // mdlE()

         /** Constructor only for use by {@link GERfixed.M} and the like. */
         protected M(double msg1, double msg2, Value sp)                   // *
          { super(msg1, msg2, sp);
            mdlV = null;
            mdlE = null;
          }//constructor

         /** The negative log probability of a Graph datum&nbsp;'g';
             it costs (i)&nbsp;|V| and (ii)&nbsp;the adjacency matrix.
             Also see {@link #nlLH nlLH(ss)}. */
         public double nlPr(Value g)                                 // nlPr(g)
          { assert mdlV != null : " mdlV==null";
            Graph gg = (Graph) g;
            if( gType.isDirected != gg.isDirected() ) // paranoia
                error( ".nlPr(" + gg + ":" + gg.type() );
            final int V = gg.vSize();
            final double [] es = gg.eStats();
            return mdlV.nlPr_n(V)  // |V| + (0s, 1s) ...
                 + mdlE().nlLH( Vector.doubles( gg.eStats() ) );  // !
          }//nlPr()

         /** Generate a random (unlabelled) Graph, Directed or
             Undirected according to {@link #gType gType}.
             Choose |V|={@link #mdlV}.random() then fill in the
             adjacency matrix at {@link #mdlE}.random().  */
         public Graph random()                                      // random()
          { assert mdlV != null : " mdlV==null";
            final int V = mdlV.random().n();
            if( V < 1 ) error( ".random(), |V|=" + V );  // !
            Series.Discrete eSer = mdlE().randomSeries();  // Edge existence
            if( gType.isDirected )
             { int [] [] A = new int [V] [V];
               for( int v0 = 0; v0 < V; v0 ++ )
                { for( int v1 = 0; v1 < V; v1 ++ )
                   { A[v0][v1] = eSer.elt_n(); // to be or not to be?
                     eSer.advance();
                   }
                  if( ! gType.selfLoops ) A[v0][v0] = 0;
                }//for v0
               return Directed.dense(gType.selfLoops, A);
             }
            else  // Graph is Undirected
             { final int Vd = V - (gType.selfLoops ? 0 : 1);  // V or V-1
               int [] [] A = new int [Vd] [];
               for( int v0 = 0; v0 < Vd; v0 ++ )         // upper right...
                { A[v0] = new int[Vd - v0];              // ...triangular
                  for( int v1 = 0; v1 < Vd-v0; v1 ++ )
                   { A[v0][v1] = eSer.elt_n();
                     eSer.advance();
                   }
                }//for v0
               return Undirected.dense(gType.selfLoops, A);
             }//fi
          }//random()

       }//IndependentEdges.M class, the fully parameterised Model

    }//IndependentEdges class, the UPModel
   // -------------------------------------------------------------------------

   /** Common special case of the {@link IndependentEdges} Model where
       {@link #upmE() upmE()} is {@link MultiState MultiState} over
       [0,&nbsp;1].  In particular, the statistical parameters of
       {@link M#mdlE()} control the probability of a possible Edge existing.
       The fully parameterised Model is {@link M}.
       ({@link GERadaptive GERadaptive} is closely related.)
       Also&nbsp;see<BR>
       E. N. Gilbert, <I>Random graphs</I>, Annals of Math. Stats.
          (Inst. Math. Stats.) 30, pp.1141-1144, 1959, and<BR>
       P. Erdos &amp; A. Renyi, <I>On random graphs</I>,
          Publ. Math. Debrecen, 6, pp.290-297, 1959.  */
   public static class GERfixed extends IndependentEdges            // GERfixed
    {
      /** Problem defining parameters
          <NOBR> dp = &lang;{@link #gType gType},
                            {@link  #upmV()  upmV()}&rang;. </NOBR>
          Note, it is hard-wired that {@link #upmE() upmE()} is
          {@link MultiState MultiState} over [0,1]. */
      public GERfixed(Value dp)
       { super(dp);
         Tuple dpT = dp.isTuple(2);
         gType = (graph.Type) dpT.fst();          // population
         upmV  = (Discretes)  dpT.snd();          // for |V|
         upmE  = new MultiState(pair(zero, one)); // Edges
       }//constructor

      public Discretes.Bounded.M ds2mdlE(Value ps, Vector ds)
       { double [] es = new double [] {0,0};
         for( int i = 0; i < ds.nElts(); i ++ )
          { final double wti = ds.wt(i);
            double [] esi = ((Graph) ds.elt(i)).eStats();
            es[0] += wti * esi[0];  // GERfixed is that all
            es[1] += wti * esi[1];  // Graphs have the same 'p'
          }
         return (Discretes.Bounded.M)
            upmE.estimator(ps).ss2Model( Vector.doubles(es) );
       }//ds2mdlE(ps,ds)

      /** {@link IndependentEdges#estimator}(pair(ps,triv)),
          the triv for MultiState's estimator.  */
      public Est estimator(Value ps)
       { return super.estimator(pair(ps, triv)); }

      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }
      // ......................................................................

      /** The fully parameterised Gilbert, Erdos-Renyi Model of Graphs.
          The UnParameterised Model is {@link GERfixed GERfixed}.
          ({@link GERadaptive.M GERadaptive.M} is closely related.  */
      public class M extends IndependentEdges.M                   // GERfixed.M
       {
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            Tuple spT = sp.isTuple(2);
            mdlV = upmV.apply(spT.fst());
            mdlE = (Discretes.Bounded.M) upmE.sp2Model(0, 0, spT.snd());
          }//constructor

         // NB. nlLH(ss)     as per Graphs.M.nlLH(ss)

       }//GERfixed.M class, the fully parameterised Model

    }//Graphs.GERfixed class, the UnParameterised Model
   // -------------------------------------------------------------------------

   /** Special case of the {@link IndependentEdges} Model where
       {@link #upmE() upmE()} is {@link Adaptive Adaptive} over [0,&nbsp;1].
       Therefore, every possible Edge in <I>a</I> Graph has the same
       probability of existing, but that probability is expected to
       change from Graph to Graph, even in the one data-set.
       The fully parameterised model is {@link M}.
       Also see {@link GERfixed}. */
   public static class GERadaptive extends IndependentEdges      // GERadaptive
    { /** Problem defining parameters
          <NOBR> dp = &lang;{@link #gType gType},
                            {@link  #upmV()  upmV()},
                 [&alpha;<SUB>0</SUB>, &alpha;<SUB>1</SUB>]&rang;. </NOBR>
          Note, it is hard-wired that {@link #upmE() upmE()} (for Edges)
          is the {@link Adaptive Adaptive} Model over [0,1], and
          [&alpha;<SUB>0</SUB>,&nbsp;&alpha;<SUB>1</SUB>] are
          the "offsets" for it; this also controls Edge-density of
          {@link M#random() random()} Graphs.  */
      public GERadaptive(Value dp)
       { super(dp);
         Tuple dpT = dp.isTuple(3);
         gType = (graph.Type) dpT.fst();
         upmV  = (Discretes)  dpT.snd();
         upmE  = new Adaptive(pair( pair(zero, one),  // (bounds)
                                    dpT.thrd() ));    // alpha
       }//constructor

      public Discretes.Bounded.M ds2mdlE(Value ps /* = triv */, Vector ds)
       { double eCost = 0;
         Adaptive.M mdlE = ((Adaptive) upmE). Mdl;
         for( int i = 0; i < ds.nElts(); i ++ )  // only msg part_2 changes
            eCost += mdlE.nlLH(Vector.doubles( ((Graph) ds.elt(i)).eStats() ));
         return (Discretes.Bounded.M) mdlE.asGiven( eCost );
       }//ds2mdlE(ps,ds)

      /** {@link IndependentEdges#estimator}(pair(ps,triv)),
          the triv for Adaptive's estimator.  */
      public Est estimator(Value ps)
       { return super.estimator(pair(ps, triv)); }

      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }
      // ......................................................................

      /** The fully parameterised Gilbert, Erdos-Renyi <I>Adaptive</I>
          Model of Graphs. The UnParameterised Model is {@link GERadaptive}.
          Also see {@link GERfixed.M GERfixed.M}.  */
      public class M extends IndependentEdges.M                // GERadaptive.M
       {
         /** Statistical parameter(s) 'sp' is for
             {@link #mdlV mdlV}&nbsp;=&nbsp;{@link #upmV}(sp).
             ({@link #mdlE} is {@link Adaptive#Mdl}.)  */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            mdlV = upmV.apply(sp);
            mdlE = ((Adaptive) upmE). Mdl;  // c.f. GERfixed
          }//constructor

         // NB. nlLH(ss)     as per Graphs.M.nlLH(ss)

       }//GERadaptive.M class, the fully parameterised Model

    }//Graphs.GERadaptive class, the UnParameterised Model
   // -------------------------------------------------------------------------

   /** An UnParameterised Model of Graphs where the degree distribution is
       skewed, few Vertices having high {@link graph.Graph#degree degree} and
       many having low degree.  It is an attempt at <I>something like</I>
       (or including) a scale free, preferential attachment Model of Graphs.
       The fully parameterised Model is {@link Skewed.M}.
       For similar Models of Graphs, I can think of plausible ways to
       calculate {@link M#nlPr nlPr(G)} (as here), and ways to to generate
       {@link M#random() random} Graphs, but not to do both in ways
       that exactly correspond.  */
   public static class Skewed extends Graphs                          // Skewed
    { /** Problem definition parameters
          <NOBR> dp = &lang;{@link #gType}, {@link #upmV}&rang;</NOBR>. */
      public Skewed(Value dp)
       { super(dp);
         Tuple dpT = dp.isTuple(2);
         gType = (graph.Type) dpT.fst();
         upmV  = (Discretes)  dpT.snd();
       }//constructor

      /** Sufficient statistics are elements [lo,&nbsp;hi)
          of the data-set 'ds' itself. */
      public Vector stats(Vector ds, int lo, int hi)         // stats(ds,lo,hi)
       { return ds.slice(lo, hi); }

      public Vector stats(boolean add, Value ss0, Value ss1)       // stats(,,)
       { return ((Vector) ss0).combine(add, (Vector.Slice) ss1); }

      /** Only {@link M#mdlV} is actually estimated. */
      public Est estimator(final Value ps)
       { return new Est(ps)
          { public M ss2Model(Value ss)
             { Vector ds = (Vector) ss;
               Discretes.M mdlV = ds2mdlV(ps, ds);
               Value sp = mdlV.statParams();
               M m = new M(0, 0, sp);
               return new M(mdlV.msg1(), m.nlLH(ds), sp);
             }
       }; }//estimator(ps)

      /** Return a fully parameterised {@link Skewed.M}. */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }
      // ......................................................................

      /** The fully parameterised Model of Skewed Graphs.
          Also see the UnParameterised {@link Skewed} Model. */
      public class M extends Graphs.M                               // Skewed.M
       { /** Statistical parameters
             sp&nbsp;=&nbsp;{@link #mdlV mdlV}<SUB>sp</SUB>. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            mdlV = upmV.sp2Model(0, 0, sp);
          }//constructor

         /** The negative log probability of Graph 'g'.
             Only a few Vertices are expected to have high
             {@link Graph#degree degree}. The adjacency matrix is
             encoded for high-degree vertices first, then low,
             which puts tight constraints on the latter. */
         public double nlPr(Value g)                                 // nlPr(g)
          { Graph.Renumbered G = (Graph.Renumbered) ((Graph) g) .byDegree();
            final int V = G.vSize();
            final boolean SL = gType.selfLoops, isDir = gType.isDirected;
            double edgeCost = 0,
                   perm = logFactorial(V); // may be able to recover some
            int nLoops = 0, deg_v = 0, ptr = -1, deg_ptr = 2*V+10;

            int edgeLimit = isDir ? (SL ? 2*V : 2*V-2)  // this for ...
                                  : (SL ? V+1 :   V-1); // ... Vertex v0

            for( int v = 0; v < V; v ++ )  // each Vertex v (byDegree)
             { deg_v = 0;
               int nEdges = 0;  // # v's (non loop) Edges not yet encoded

               if( isDir ) // i.e., A[,] is square
                { Series.Int vsp = G.directPredecessors(v), // {<?,v>}
                             vss = G.directSuccessors(v);   // {<v,?>}

                  // some of v's Edges have already been transmitted...
                  while( vsp.hasSome() && vsp.elt_n() < v ) // <u,v>, u < v
                   { edgeLimit -- ; vsp.advance(); }        // above diag
                  while( vss.hasSome() && vss.elt_n() < v ) // <v,u>, u < v
                   { edgeLimit -- ; vss.advance(); }        // left of diag

                  if( vsp.hasSome() && vsp.elt_n() == v )   // <v,v>
                   { nLoops ++ ; edgeLimit -= 2;  // & see deg_v below
                     assert vss.elt_n() == v : ".nlPr !";
                     vsp.advance();  vss.advance();
                   }

                  // the following not yet transmitted...
                  while( vsp.hasSome() )                    // <u,v>, u > v
                   { nEdges ++ ; vsp.advance(); }           // below diag
                  while( vss.hasSome() )                    // <v,u>, u > v
                   { nEdges ++ ; vss.advance(); }           // right of diag

                  deg_v += vsp.position() + vss.position(); // !
                }
               else // unDirected, i.e., A[,] is symmetric (~ upper right)
                { Series.Int vs = G.joinedTo(v);

                  while( vs.hasSome()  && vs.elt_n() < v )  // (u,v), u < v
                   { edgeLimit -- ; vs.advance(); }         // above diag

                  if( vs.hasSome() && vs.elt_n() == v )     // (v,v)
                   { deg_v ++ ; nLoops ++ ; edgeLimit -= 2; vs.advance(); }

                  while( vs.hasSome() )                     // (u,v), u > v
                   { nEdges ++ ; vs.advance(); }            // right of diag

                  deg_v += vs.position();                   // !
                }

               if( edgeLimit > 0 ) // v's (non-loop) Edge cost...
                  edgeCost += Math.log(edgeLimit+1) // nEdges in [0,edgeLimit]
                           + log_nCk(isDir ? 2*(V-v-1) : V-v-1, nEdges);

               if( deg_v < deg_ptr )  // end of a "run" of equal degree
                { final int runLen = v-ptr;  // >= 1,  perm bits back...
                  if( runLen > 1 ) perm -= logFactorial(runLen);
                  deg_ptr = deg_v; ptr = v;  // begin next run
                }

               if( nEdges > edgeLimit ) // paranoia
                  error( ".nlPr(g), v=" + v + ", nEdges="+nEdges
                       + " > edgeLimit="+edgeLimit);

               edgeLimit = deg_v;       // Note, degree(v+1) <= degree(v)
             }//for v

            if( ptr < V-1 )                      // ? final run of
               perm -= logFactorial(V-ptr); // equal degree

            double loopCost = SL ? Math.log(V+1)+log_nCk(V,nLoops) : 0;

            return mdlV().nlPr(new Value.Int(V)) + perm + loopCost + edgeCost;
          }//nlPr(g)

         // NB. nlLH(ss) as per Graphs.M.nlLH(ss)

         /** PROTOTYPE, Generate a random (unlabelled) Graph, Directed or
             Undirected according to {@link #gType gType}.
             The frequency of degree 'd' in the <I>target</I>-degrees
             ~&nbsp;1&nbsp;/&nbsp;(d+minDeg) where minDeg&nbsp;=&nbsp;0
             (or =1 to taste). */
         public Graph random()                                      // random()
          { assert mdlV != null : " mdlV==null";
            final int V = mdlV.random().n();
            if( V < 1 ) error( ".random(), |V|=" + V );  // !
            int [] vs  = new int [V],       // Vertices, sorted '>=' on ...
                   deg = new int [V];       // ... degree
            mml.Permutation.random(vs);     // random vs[.]
            for( int i = 0; i < V; i ++ ) deg[i] = 0;
            int degMin = 1 /* 0 or 1 ? */, degMax;

            if( gType.isDirected )
             { final int [] [] A = new int [V] [V];
               for( int i = 0; i < V; i ++ )
                  for( int j = 0; j < V; j ++ )
                     A[i][j] = 0;           // A[.][.] = 0
               final int [] buffer = new int [2*V];
               degMax = 2*(V-1);            // selfLoops ???

               double [] cummulative = new double[degMax+1];  // for pr ~ 1/i
               double sum = 0, term = 1;
               for( int i = degMin; i < cummulative.length; i ++ )
                { sum += 1.0 / term; cummulative[i] = sum; term ++ ; }
               // have cummulative[] probs of ~ 1/n
               double [] sample = new double[V];
               for( int i = 0; i < V; i ++ )  // sample[] is uniform ...
                  sample[i] = sum * MML.RNG.nextDouble();  // in [0, sum)
               java.util.Arrays.sort(sample); // sample now increasing

               int [] degTgt = new int[V];    // target degrees
               int tgt = degMax - 1;
               for( int i = 0; i < V; i ++ )  // each Vertex i
                { double s_i = sample[V-1-i]; // i ~ sample[...]
                  while( tgt >= degMin && s_i < cummulative[tgt] )
                     tgt -- ;                 // i.e., linear search :-(
                  degTgt[i] = tgt + 1; // s_i <  cummulative[tgt+1] &
                }                      // s_i >= cummulative[tgt]
               // frequency of degree 'd' in degTgt[] is ~ 1/(d+degMin)
               // and degTgt[.] is decreasing (<=)

               for( int trye = 0; trye < 2; trye ++ )
                { // try to reach each Vertex's target degree
                  for( int i = 0; i < V; i ++ )
                   { int v0 = vs[i],           // v0
                         poss = 2*(V-i-1);     // possible edge positions
                     int degTgt0 = degTgt[i];
                     if( deg[v0] >= degTgt0 ) continue;

                     for( int j = 0; j < degTgt0-deg[v0]; j ++ )
                        buffer[j] = 1;         // extra Edges and ...
                     for( int j = degTgt0-deg[v0]; j < poss; j ++ )
                      { buffer[j] = 0;         // ... non-Edges
                        final int s = (int) Math.floor(
                                      MML.RNG.nextDouble() * (j+1) );
                        // 0 <= s <= j, like random() Permutation
                        final int temp = buffer[s];  buffer[s] = buffer[j];
                        buffer[j] = temp;      // swap
                      }//for j

                     A[v0][v0] = 0;            // ??? loops
                     for( int j = 0; j < V-i-1; j ++ )
                      { final int ij1 = i+j+1;
                        final int v1 = vs[ij1],    // NB. v1 != v0
                                  k  = MML.RNG.nextDouble() < 0.5 ? 0 : 1;
                        if( deg[v1] >= degTgt[ij1] ) continue; // !
                        final int v0v1 = buffer[2*j+  k];
                        if( A[v0][v1] == 0 && v0v1 == 1 )
                         { A[v0][v1] = v0v1;
                           deg[v0]  += v0v1;
                           deg[v1]  += v0v1;
                         }
                        if( deg[v1] >= degTgt[ij1] ) continue; // !
                        final int v1v0 = buffer[2*j+1-k];
                        if( A[v1][v0] == 0 && v1v0 == 1 )
                         { A[v1][v0] = v1v0;
                           deg[v0]  += v1v0;
                           deg[v1]  += v1v0;
                         }
                      }//for j

                     // re-sort vs[i+1, V) on ascending deg[vs[.]]
                     for( int j = i+2; j < V; j ++ )
                      { final int w = vs[j];  final int deg_w = deg[w];
                        int k;                 // ? probably a faster way ?
                        for( k = j-1; deg_w > deg[vs[k]]; k -- )
                           vs[k+1] = vs[k];
                        // k >= i and deg_w <= deg[vs[k]]
                        vs[k+1] = w;
                      }//for j

                   }//for i (and Vertex v0)
                }//for trye

               return Directed.dense(gType.selfLoops, A);
             }
            else// Undirected
               error( " not yet implemented for Undirected ");
            return null; // ???
          }//random()

      }//Skewed.M class, the fully parameterised Model

    }//Graphs.Skewed class, the UnParameterised Model
   // -------------------------------------------------------------------------

   /** UnParameterised Models based on the notion of frequent sub-graphs
       (motifs, patterns) and that can calculate the message-length
       {@link #msgMotifs contribution} of a set of motifs.
       Also see {@link Graphs.Motifs.M}. */
   public static abstract class Motifs extends Graphs               // * Motifs
    {
      /** Whether (verbosity&gt;0) or not (verbosity==0) to Print
          tracing information. Presently the values 0,1,2 print
          increasing amounts of tracing information.  */
      public int verbosity = 1;

      /** The (prior) Model of the number of {@link M#motifs motifs}
          (motifs, patterns) in a {@link M} Model.
          Anything sensible will prefer "few" to "many".
          NB.&nbsp;must be set by constructor of non-abstract subclass. */
      protected Discretes.M mdlNmotifs;                           // mdlNmotifs

      /** The Model used for each individual {@link M#motifs motifs}
          (subGraph, pattern) in a {@link M}-Model.  NB.&nbsp;must be
          set by the constructor of non-abstract subclass. */
      protected Model mdlMotif;                                     // mdlMotif

      /** Constructor only for use by subclasses of this class;
          it does not set {@link #gType gType} or {@link #upmV upmV}. */
      protected Motifs(Value dp)
       { super(dp);
         mdlNmotifs = null;  mdlMotif = null;  // it's abstract
       }//constructor

      /** Return the cost, in nits, of stating 'N' motifs
          m[0], ..., m[N-1] (including stating N) as part of some
          fully parameterised {@link Motifs.M Model}.   */
      public double msgMotifs(Graph [] ms)                     // msgMotifs(ms)
       { double ans = mdlNmotifs.nlPr_n(ms.length);  // i.e., 'N'
         for( int i = 0; i < ms.length; i ++ )
            ans += mdlMotif.nlPr(ms[i]);
         return ans;
       }//msgMotifs(ms)

      /** Return the cost, in nits, of stating 'N' motifs
          m<SUB>0</SUB>, ..., m<SUB>N-1</SUB> (including stating N).
          Also see {@link Graphs.Motifs}. */
      public double msgMotifs(Vector vm)                       // msgMotifs(vm)
       { double ans = mdlNmotifs.nlPr_n(vm.nElts());  // i.e., 'N'
         for( int i = 0; i < vm.nElts(); i ++ )
            ans += mdlMotif.nlPr(vm.elt(i));
         return ans;
       }//msgMotifs(vm)


      /** Greedy search for a subset of some given 'motifs' that minimises
          the message length under this Motif Model of Graphs.
          It requires motifs, a Model of |V|, and statistics ss=stats(Gs).
          Greedy may be used select from a given list of proposed motifs or
          used within a Motif Model's {@link #estimator estimator}. */
      public M greedy( Vector motifs,    // candidate motifs.        greedy(,,)
                       Discretes.M mdlV, // mdl of |V|
                       Value    ss )     // stats of Graph data-set
       { if(verbosity > 0 )
            System.out.println("   " + this + ".greedy()" );

         final Value mdlV_sp = mdlV . statParams();
         final int   nMotifs = motifs.nElts();

         int [] bestPick = new int [0];  // empty current bestPick
         boolean [] chosen = new boolean [nMotifs],
                    ignore = new boolean [nMotifs];
         for( int i = 0; i < nMotifs; i ++ )
          { chosen[i] = false;  // i.e., none
            ignore[i] = false;
          }
         // to begin, the 'best' model has zero motifs, {}, ...
         Graph [] none = new Graph [] {};
         Motifs.M best = sp2Model(0,0, pair(mdlV_sp, Vector.values(none)));
         double bestRes_Ms   = msgMotifs(none),  // cost of (as yet no) Motifs
                bestRes_data = best.nlLH(ss);
         double bestRes = bestRes_Ms + bestRes_data; // minimum cost known
         boolean improved = true;

         if(verbosity > 0 )
            System.out.println( "   greedy start: "
             + x2String(bestRes/MML.log2) + " bits." );

         while( improved && bestPick.length < nMotifs )
          { Motifs.M bestAug = null;  // best augmented Model to be
            double   bestAugRes = Double.MAX_VALUE,  // "infinity"
                     bestAugRes_Ms   = 0,
                     bestAugRes_data = 0;
            int []   bestAugPick     = null;
            int choice = -1;
            improved  = false;

            for( int i = 0; i < motifs.nElts(); i ++ ) // try to augment the...
             { if( ! chosen[i] && ! ignore[i] )       // ...bestPick with 'i'
                { int [] augPick = new int [bestPick.length+1];
                  int j;
                  for( j = 0; j < bestPick.length && bestPick[j] < i; j ++ )
                     augPick[j] = bestPick[j];
                  augPick[j] = i;   // insert 'i'
                  for( ; j < bestPick.length; j ++ )
                     augPick[j+1] = bestPick[j];
                  Vector augPickV = motifs.elts(augPick);
                  Motifs.M aug = sp2Model(0, 0, pair(mdlV_sp, augPickV));
                  double augRes_Ms   = msgMotifs(augPickV),
                         augRes_data = aug.nlLH(ss);
                  double augRes = augRes_Ms + augRes_data;  // cost, in nits

                  if(verbosity > 0 )
                     System.out.println("   greedy " + Vector.ints(bestPick)
                      + "->" + Vector.ints(augPick) + " "
                      + x2String(augRes/MML.log2) + " bits = "
                      + x2String(augRes_Ms/MML.log2) + "+"
                      + x2String(augRes_data/MML.log2) + ";"
                    );

                  if( augRes > bestRes * 1.005 ) ignore[i] = true;  // ???
                  if( augRes < bestAugRes )  // looks good
                   { bestAug = aug; bestAugPick = augPick;
                     bestAugRes = augRes;
                     bestAugRes_Ms = augRes_Ms; bestAugRes_data = augRes_data;
                     choice = i;
                   }//if
                }//if
             }//for i
            // NB. Do not mess with 'best*' inside the 'for i' loop above!
            if( bestAugRes < bestRes )  // new best model
             { best = bestAug; bestPick = bestAugPick;
               bestRes = bestAugRes;
               bestRes_Ms = bestAugRes_Ms; bestRes_data = bestAugRes_data;
               chosen[choice] = true;
               improved = true;
               if(verbosity > 0 )
                  System.out.println( "   " + this + ".greedy added "
                   + choice + " improved: " + (bestRes/MML.log2)+" bits." );
             }//if
            else
              if(verbosity > 0 )
                 System.out.println( "   " + this
                  + ".greedy added nothing (no improvements) " );
          }//while improved...

         if(verbosity > 0 )
            System.out.println( "   greedy picked " + Vector.ints(bestPick)
             + ": " + best + " " + (bestRes/MML.log2) + " bits.\n" );

         Vector motifsPick = motifs.elts(bestPick);
         double msg1 = mdlV.msg1() + bestRes_Ms,
                msg2 = bestRes_data;
         return sp2Model(msg1, msg2, pair(mdlV_sp, motifsPick));
       }//greedy(...)

      /** Given candidate 'motifs' and ss=stats(Gs), estimate mdlV and then
          return {@link #greedy greedy(motifs,&nbsp;mdlV,&nbsp;ss)}.  */
      public M greedy( Value ps,         // for upmV.estimator(ps)
                       Vector motifs,    // candidate motifs,         greedy(,)
                       Value    ss )     // stats of Graph data-set
       { return greedy(motifs, ds2mdlV(ps, (Vector) ss), ss ); }

      public abstract Motifs.M sp2Model(double msg1, double msg2, Value sp);
      // ......................................................................

      /** Fully parameterised Models based around the notion of
          frequent sub-graphs (patterns motifs). */
      public abstract class M extends Graphs.M                    // * Motifs.M
       { /** Return the Motifs on which this Model is based. */
         public abstract Graph [] motifs();                                // *

         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); }//constructor

         /** Motifs.{@link Motifs#msgMotifs(Graph[]) msgMotifs(ms)}. */
         public double msgMotifs(Graph [] ms)
          { return Motifs.this.msgMotifs(ms); }

         /** Motifs.{@link Motifs#msgMotifs(Vector)  msgMotifs(vm)}. */
         public double msgMotifs(Vector vm)
          { return Motifs.this.msgMotifs(vm); }

       }//Graphs.Motifs.M class, the fully parameterised Model

    }//Graphs.Motifs class, the UnParameterised Model
   // -------------------------------------------------------------------------

 }//Graphs class, the UnParameterised Model

// (c)Lloyd Allison, Aug 2014, Sun 10 Dec 2017. Licence to 'Graphs' is Gnu GPL.
