//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;   // standard functions
import la.la.Value;
import static la.maths.Maths.*;
import la.maths.Vector;
import la.maths.Vector.Slice;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** An UnParameterised Mixture Model capable of producing a fully
    parameterised Mixture {@link M Model}; in this case, the
    problem-{@link #defnParams defining} parameter is a {@link UPModel},
    {@link #upm}, and the {@link M#statParams statistical}-parameters
    are the weights, plus the statParams, of the clusters &mdash; the classes
    in the Snob sense of "class". Here, each component is the same kind of
    Model &mdash; a fully parameterised upm-Model.  */
public class Mixture extends UPModel                                 // Mixture
 { /** The {@link UPModel} of a class- (component-) sub-Model,
       <NOBR><TT>upm : statParams &rarr; Model</TT>,</NOBR> that
       is to be "mixed" is the single {@link M#defnParams()}.  */
   public final UPModel upm;

   public Mixture(UPModel upm)
    { super(upm); this.upm = upm; }//constructor

   /** sp2Model(msg1, msg2, (wts,sps)) where wts is the weights of the
       classes (clusters) and sps is the statParams of the classes, to
       be given to {@link #upm upm}.  Note that an MML message starts
       with the # of components, N, here N=|wts| but it must be encoded. */
   public M sp2Model(double m1, double m2, Value wp)
    { return new M(m1, m2, wp); }

   /** Mixture sufficient statistics, ss&nbsp;=&nbsp;ds.[lo,hi) itself.
       There does not seem to be any alternative <I>in general</I>.
       More on stats [{@link UPModel#stats here}]. */
   public Vector stats(Vector ds, int lo, int hi) { return ds.slice(lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return ((Vector) ss0).combine(add, (Slice) ss1); }

   /** Return an {@link Est estimator} for 'this' Mixture;
       note, its parameter will be upm's parameter. */
   public Estimator estimator(Value ps)
    { return new Est(ps); }
   // .........................................................................

   /** The {@link Mixture Unparameterised} Mixture's Estimator,
       capable of {@link #ss2Model estimating} a fully
       {@link Mixture.M parameterised} Mixture Model. */
   public class Est extends UPModel.Est                          // Mixture.Est
    {
      /** The estimator for an individual class (cluster, component)
          of the Mixture {@link Mixture.M Model}. */
      public final Estimator classEst;

      /** Note that {@link #classEst classEst} =
          {@link #upm upm}.estimator(ps). */
      public Est(Value ps)
       { super(ps);
         classEst = upm.estimator(ps);
       }//constructor

      /** Given ss (=ds), estimate a Mixture {@link Mixture.M Model}.
          ss2Model&nbsp;carries out a heuristic search, calling
          {@link #EM EM}, {@link #split split}, {@link #kill kill}
          and {@link #join join} until convergence.  */
      public M ss2Model(Value ss)
       { Vector ds = (Vector) ss; // actually ss=ds, the data-set
         // First, estimate the trivial 1-class mixture ...
         Model c0 = classEst.ds2Model(ds);            // c0, a single class
         final double msg1 = WallaceInt0.nlPr_n(0) + c0.msg1();
         M imprv = (M) sp2Model(msg1, c0.msg2(),      // a "mixture" of c0
                   pair(Vector.doubles(new double[]{ 1.0 }),
                        Vector.values( new Value[]{ c0.statParams() } ))),
           ans = null;
         for( int iter = 1; ans != imprv; iter ++ ) // while improving...
          { ans = imprv;
// System.out.println( "  ans: " + ans.msg() + " nits " + ans );
            int nC = ans.mixer.upb_n() + 1;    // nC, # of classes
            for( int c = 0; c < nC; c ++ )
             { M m = split(ans, c, ds);             // split c?
// System.out.println( "   s" + c + ": " + m.msg() + " " + m );
               if( m.msg() < imprv.msg() ) imprv = m;
             }//for c
            if( imprv != ans && (iter % 2) != 0 ) continue;
         // else, has not improved, or 2 divides iter ...
            if( nC > 2 ) // note, nC=2 equiv to join 0 to 1
               for( int c = 0; c < nC; c ++ )
                { M m = kill(ans, c, ds);           // kill c?
// System.out.println( "   k" + c + ": " + m.msg() + " " + m );
                  if( m.msg() < imprv.msg() ) imprv = m;
                }//for c
            if( nC > 1 )
               for( int c = 0; c < nC-1; c ++ )
                  for( int c2 = c+1; c2 < nC; c2 ++ )
                   { M m = join(ans, c, c2, ds);    // join c and c2?
// System.out.println( " j" + c + " " + c2 + ": " + m.msg() + " " + m );
                     if( m.msg() < imprv.msg() ) imprv = m;
                   }//for c2
          }//for iter
         return ans;
       }//ss2Model(ss)

      /** The core Expectation Maximization algorithm. Alternately
          estimate memberhips from a Model, and a Model from memberships.
          To start, 'mx' or 'mshp', but not both, may be null. */
      public M EM(M mx, double [] [] mshp, Vector ds)      // EM(mx,mshp,ds)
       { int maxIter = 25;  // roughly!
         int inner = Math.min(5, maxIter);
         if( mx == null )                       // start from mshp
            mx = mshp2Model(mshp, ds);          // mshp -> mx
         else /* starts from mx */ if( mshp == null ) // allocate space
            mshp = new double [mx.mixer.upb_n()+1] [ds.nElts()];
         double msg2 = mx.nlLH(ds),   // NB. nlLH(.), msg2() is not set
                bestDec = Double.NEGATIVE_INFINITY;
         for( int iter = 1; iter <= maxIter; /*NB*/ )
          { double msg2a = msg2;
            for( int j = 0; j < inner; j ++ )   // EM algorithm...
             { Model2mshp(mx, mshp, ds);        //    mx -> mshp
               mx = mshp2Model(mshp, ds);       //    mshp -> mx
               iter ++ ;
             }//for j                           // repeat.
            msg2 = mx.nlLH(ds);                 // Now and then...
            double decrease = (msg2a-msg2)/inner;  // improvement rate
            if( decrease > bestDec ) bestDec = decrease;
            else if( 5.0 * decrease < bestDec && decrease < 0.0001 * msg2a )
               break;  // flattened out, stop
          }//for iter
         return (M) mx.asGiven(mx.msg1(), msg2);
       }//EM(mx,mshp,ds)

      /** {@link #EM Find} a good Mixture Model starting from the given
          memberships. The number of classes (clusters) is not changed. */
      public M EM(double [] [] mshp, Vector ds)
       { return EM(null, mshp, ds); }

      /** {@link #EM Find} a good Mixture Model starting from the given
          Model, mx. The number of classes (clusters) is not changed. */
      public M EM(M mx, Vector ds)
       { return EM(mx, null, ds); }

      /** Try to {@link #EM Find} a good Mixture Model with
          'nClass' classes (clusters).
          It starts from {@link #randomMshp random} memberships. */
      public M makeNclasses(int nClass, Vector ds)
       { return EM(null, randomMshp(nClass, ds.nElts()), ds); }

     /** Remove class 'c', redistribute its members,
         and {@link #EM adjust}. */
      public M kill(M mx, int c, Vector ds)
       { final int nCls_1 = mx.mixer.upb_n(), nData = ds.nElts();
         assert 0<=c && c<=nCls_1 : "kill: c="+c + ", nCls="+(nCls_1+1);
         double [] [] mshp =  Model2mshp(mx, ds);
         final double [] mshp_c = mshp[c];         // c's members
         double [] [] mshp2 = new double [nCls_1] [];
         for( int cl = 0; cl < nCls_1; cl ++ )
            mshp2[cl] = mshp[cl];                  // copy memberships,
         if( c < nCls_1 ) mshp2[c] = mshp[nCls_1]; // overwrite ex class,
         for( int cl = 0; cl < nCls_1; cl ++ )     // adjust remaining
          { final double [] mshp2_cl = mshp2[cl];  // memberships ...
            for( int r = 0; r < nData; r ++ )
             { final double loss = mshp_c[r];
               if( loss > 0.94 )  // little left to go on ...
                    mshp2_cl[r] = 1.0 / nCls_1;    // ... evenly, or
               else mshp2_cl[r] /= 1-loss;         // ... up
          }  }
         return EM(null, mshp2, ds);    // adjust all
       }//kill(mx,c,ds)

      /** Split class 'c', replace it with two sub-classes,
          and {@link #EM adjust}. */
      public M split(M mx, int c, Vector ds)
       { final int nCls = mx.mixer.upb_n()+1, nData = ds.nElts();
         assert 0<=c && c<nCls : "split: c="+c + ", nCls="+nCls;
         double [] [] mshp =  Model2mshp(mx, ds);
         Vector ds_c = ds.weight(mshp[c]); // weight as c's members
         M mx_s = makeNclasses(2, ds_c);   // split c, and adjust
         double [] [] mshp_s = Model2mshp(mx_s, ds_c); // two subclasses
         final double [] mshp_c = mshp[c],
                         mshp_s_0 = mshp_s[0], mshp_s_1 = mshp_s[1];
         for( int r = 0; r < nData; r ++ ) // for each datum r,
          { final double mcr = mshp_c[r];  // original membership
            mshp_s_0[r] *= mcr;            // divided between child 0
            mshp_s_1[r] *= mcr;            // and child 1
          }//for r
         double [] [] mshp2 = new double [nCls + 1] [];
         for( int cl = 0; cl < nCls; cl ++ )
            mshp2[cl] = mshp[cl];       // copy memberships,
         mshp2[c]      = mshp[nCls-1];  // overwrite the ex class,
         mshp2[nCls-1] = mshp_s_0;      // add child 0
         mshp2[nCls]   = mshp_s_1;      // and child 1,
         return EM(null, mshp2, ds);    // adjust all
       }//split(mx,c,ds)

      /** Merge classes c1 and c2 into one class,
          and {@link #EM adjust}. */
      public M join(M mx, int c1, int c2, Vector ds)
       { final int nCls = mx.mixer.upb_n()+1, nData = ds.nElts();
         assert 0<=c1 && c1<nCls && 0<=c2 && c2<nCls && c1!=c2
              : "join: c1="+c1 + ", c2="+c2 + ", nCls="+nCls;
         double [] [] mshp =  Model2mshp(mx, ds);
         final double [] [] mshp2 = new double [nCls - 1] [];
         if( c1 > c2 ) { int t = c1; c1 = c2; c2 = t; } // c1 < c2
// bug fixed 5/11/14  final double [] m_c1 = mshp[c2], m_c2 = mshp[c2];  // ...
         final double [] m_c1 = mshp[c1], m_c2 = mshp[c2];
         for( int r = 0; r < nData; r ++ )
            m_c1[r] += m_c2[r];                 // add c2 into c1
         for( int c = 0; c < nCls-1 /*NB -1*/; c ++ )
            mshp2[c] = mshp[c];                 // copy memberships
         if( c2 < nCls-1 ) mshp2[c2] = mshp[nCls-1]; // delete c2
         return EM(null, mshp2, ds);            // adjust all
       }//join(mx,c1,c2,ds)

      /** Given (possibly fractional) class memberships, mshp,
          (re-)estimate the Mixture Model (but <B>note</B> that msg2()
          is zero).  Memberships are stored by columns for use as
          {@link Vector#weight(double[]) weights}. */
      public M mshp2Model(double [] [] mshp, Vector ds)
       { final int nCls = mshp.length;
         Model mixer =
            new MultiState(pair(zero, new Value.Int(nCls-1))) // [0, nCls)
               .estimator(triv)
                  .ss2Model(Vector.doubles(mshp2abndc(mshp, ds)));
         double msg1 = WallaceInt0.nlPr_n(nCls-1) // # classes
                     + mixer.msg1();  // NB. msg1(), don't care for mshps!
         final Value [] ps = new Value [nCls];    // class params
         for( int c = 0; c < nCls; c ++ )         // each class (column)
          { Model clss = classEst.apply( ds.weight(mshp[c]) );  // NB.
            msg1 += clss.msg1();
            ps[c] = clss.statParams();
          }//for c
         return (M) sp2Model(msg1, 0, pair(mixer.statParams(),
                                           Vector.values(ps)) );
       }//mshp2Model(,)

      /** Calculate class abundances from class memberships, mshp.
          Note, 'ds' is used to provide the data weights. */
      public double [] mshp2abndc(double [] [] mshp, Vector ds)
       { final int nData = ds.nElts(),  nCls = mshp.length;
         final double [] abndc = new double [nCls];
         for( int c = 0; c < nCls; c ++ )         // for each class, c
          { abndc[c] = 0;
            final double [] mshp_c = mshp[c];
            for( int r = 0; r < nData; r ++ )     // for each datum, r
               abndc[c] += mshp_c[r] * ds.wt(r);  // NB., *wt(r)
          }//for c
         return abndc;
       }//mshp2abndc(mshp)

      /** Given a Mixture Model, mx, return the (fractional)
          class memberships, mshp, of the data. */
      public void Model2mshp(M mx, double [] [] mshp, Vector ds)
       { final int nElts = ds.nElts(),
                   nCls  = mx.mixer().upb_n() + 1;  // = |[lwb,upb]|
         final double [] nlPr = new double[nCls],
                         odds = new double[nCls];
         for( int r = 0; r < nElts; r ++ )    // for each datum...
          { Value ds_r = ds.elt(r);
            for( int c = 0; c < nCls; c ++ )  // for each class, c
               nlPr[c] = mx.nlPr(c, ds_r);    // class_c & ds_r
            nlPrs2odds(nlPr, odds);      // exp-, normalise
            for( int c = 0; c < nCls; c ++ )
               mshp[c][r] = odds[c];
          }//for r
       }//Model2mshp(mx,mshp,ds)

      /** See {@link #Model2mshp}. This is another parameterisation. */
      public double [] [] Model2mshp(M mx, Vector ds)
       { double[] [] mshp = new double[mx.mixer().upb_n() + 1] [ds.nElts()];
         Model2mshp(mx, mshp, ds);
         return mshp;
       }

      /** Return a matrix of random class memberships. It has
          'nClass' classes (clusters, columns) and 'nData' rows,
          stored by columns for use as
          {@link Vector#weight(double[]) weights}. */
      public double [] [] randomMshp(int nClass, int nData)
       { double [] [] mshp = new double [nClass] [nData];
         for( int r = 0; r < nData; r ++ )      // for each datum
          { double sum = 0;
            for( int c = 0; c < nClass; c ++ )  // for each class
             { double rn = RNG.nextDouble();    // in [0,1)
               mshp[c][r] = rn;  sum += rn;
             }//for c
            for( int c = 0; c < nClass; c ++ )
               mshp[c][r] /= sum;               // normalise
          }//for r
         return mshp;
       }//randomMshp(nC,nD)

    }//Mixture.Est class
   // .........................................................................

   /** A fully parameterised (non-abstract) Mixture Model
       with {@link #statParams() statistical}-parameters being the
       weights and parameters of the classes (clusters, components).
       Also see the {@link Mixture UnParameterised} Mixture. */
   public class M extends UPModel.M                                // Mixture.M
    {
      private final MultiState.M mixer; // see mixer() below
      private final     Model [] ms;    // see ms() below

      /** Return the MultiState that weights the class Models. */
      public MultiState.M mixer() { return mixer; }

      /** Return the sub-Models, one per class (cluster). */
      public Model []        ms() { return ms;    }

      /** Given statistical parameter sp&nbsp;=&nbsp;(wts,&nbsp;sps),
          where 'wts' is the relative abundances and 'sps' the classes'
          statistical parameters, construct a Mixture Model. */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Tuple wts_sps = sp.isTuple(2);
         final Vector wts = (Vector) wts_sps.fst();  int N = wts.nElts();
         final Vector sps = (Vector) wts_sps.snd();
         if( sps.nElts() != N )
            error( " |wts|=" + N + " != |sps|=" + sps.nElts() );
         mixer = new MultiState(
                    Value.pair(zero, new Value.Int(N-1))) // bounds
                    .sp2Model(0.0, 0.0, wts);
         ms = new Model[N];
         for( int i = 0; i < N; i ++ )  // the sub-Models, per class
            ms[i] = upm.sp2Model(0.0, 0.0, sps.elt(i));
       }//constructor

      /** The negative log probability of datum d under 'this' Mixture is
          <NOBR>log( &sum;{ mixer().pr(i) &times; ms()[i].pr(d) } ) </NOBR>
          <NOBR> = logSum{ mixer().nlPr(i) + ms()[i].nlPr(d) } </NOBR>.
          Also see {@link la.maths.Maths#logSum}.  */
      public double nlPr(Value d)                                 // nlPr(d)
       { final double [] x = new double[ms().length];
         for( int i = 0; i < ms().length; i ++ )
            x[i] = mixer().nlPr_n(i) + ms()[i].nlPr(d);
         return logSum(x);           // NB. logSum, i.e.,
         // typically slightly worse than the best class alone
       }//nlPr(d)

      /** The negative log probability of datum 'd' according to class 'i'
          <I>alone</I>. That is, nlPr(i)&nbsp;+&nbsp;nlPr(d|i).  */
      public double nlPr(int i, Value d)
       { return mixer().nlPr_n(i) + ms()[i].nlPr(d); }

      /** Assumes that {@link Mixture#stats(Vector,int,int) statistics}
          are the data-set itself. */
      public double nlLH(Value ss)
       { return sumNlPr((Vector) ss); } // best one can do?

      /** Produce (sample) a random Value from 'this' Mixture Model.
          Select a random sub-Model {@link #ms() ms()[r]} and generate
          a {@link Model#random() random} Value from it, if possible. */
      public Value random()                                      // random()
       { try{ return ms()[ mixer().random_n() ].random();
            }catch(Exception e) { throw UOE( ".random() " + e ); }
       }//random()

      public M asGiven(double msg2)
       { return /* an M */ asGiven(0, msg2); }

      public M asGiven(double msg1, double msg2)
       { return /* an M */ sp2Model(msg1, msg2, statParams()); }


      /** 'this' Mixture.M is not a {@link Multivariate.M} <I>but</I>
          this.asMultivariateM() is &ndash; provided that {@link #upm} is
          a Multivariate. In particular, this allows such a mixture model
          access to {@link Multivariate.M#functionModel}.  */
      public Multivariate.M asMultivariateM()
       { /** Trivial derivation from Mixture and Mixture.M just to satisfy
             Multivariate and Multivariate.M over width()!  */
         class MV extends Multivariate
          { public MV(Value dp) { super(dp); }
            public int width()  { return ((Multivariate) upm).width(); }
            public Value stats(Vector ds, int lo, int hi)
             { return Mixture.this.stats(ds, lo, hi); }
            public Value stats(boolean add, Value ss0, Value ss1)
             { return Mixture.this.stats(add, ss0, ss1); }
            public MVM sp2Model(double msg1, double msg2, Value sp)
             { return new MVM(msg1, msg2, sp); }
            // ..........................................
            class MVM extends Multivariate.M 
             { public MVM(double msg1, double msg2, Value sp)
                { super(msg1, msg2, sp); }
               public double nlPr(Value d)  { return Mixture.M.this.nlPr(d);  }
               public double nlLH(Value ss) { return Mixture.M.this.nlLH(ss); }
               public Value  random()       { return Mixture.M.this.random(); }
               public String toString()
                { return Mixture.M.this.toString() + ".asMultivariateM()"; }
             }//MVM class
          }//MV class

         if( ! (upm instanceof Multivariate) )  // see width() above
            error( " upm is not Multivariate" );
         return new MV(triv).new MVM(0, 0, triv);
       }//asMultivariateM()


      /** Return a short description of the Mixture Model. */
      public String toString()
       { return "(" + Mixture.this + " " + statParams() + ")"; }

    }//Mixture.M class, the Model

 }//Mixture class, the UPModel

// (c) Lloyd Allison, 2012, 5 March 2018.  Licence to 'Mixture' is Gnu GPL.
