//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Type;
import la.maths.Vector;
import static la.maths.Maths.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The (abstract) class of UnParameterised Models over Multivariate data
    (Tuples).  The fully parameterised Model is {@link Multivariate.M}.
    Also see {@link Independent Independent}, {@link Dependent Dependent}, and
    Mixture.M.{@link Mixture.M#asMultivariateM() asMultivariateM()}. */
public abstract class Multivariate extends UPModel            // * Multivariate
 {
   public Multivariate(Value upms)
    { super(upms); }//constructor

   /** Return {@link #width}, the number of variables (columns). */
   public abstract int width();                                 // * width()
   // ................................

   /** The (abstract) class of fully parameterised Models over
       Multivariate data (i.e., Tuples).  Also see {@link Multivariate},
       {@link Independent.M} and {@link Dependent.M}. */
   public abstract class M extends UPModel.M               // Multivariate.M
    { /** Given two-part message lengths, msg1 and msg2, and a
          Tuple of statistical parameters, sps, construct a Model
          of Tuples from the upms. */
      public M(double msg1, double msg2, Value sps)
       { super(msg1, msg2, sps); }//constructor

      /** Return the enclosing Multivariate's
          {@link Multivariate#width width()}. */
      public int width() { return Multivariate.this.width(); }


      /** Return a FunctionModel of Cs&rarr;col where col is of Bounded
          Discrete colType and is one of 'this' MultiVariate's columns
          (variables), and Cs is a <EM>Tuple</EM> of Values for all the
          <EM>other</EM> columns (variables). */           // functionModel(,,)
      public FunctionModel functionModel(int col, Type colType)
       { final Type.Discrete cT = (Type.Discrete) colType;
         final int lwb_n = cT.lwb_n(),   upb_n = cT.upb_n();
         if( col < 0 || col >= width() )
            error( ", col=" + col + " is < 0 or >= " + width() );

         /** msg1 = msg2 = 0 provided that we have "paid" this Multivariate. */
         FunctionModel fm = new FunctionModel(0, 0, triv)
          { /** Sufficient statistics are the data-set. */
            public Value stats(Vector ds, int lo, int hi)
             { return ds.slice(lo, hi); }

            /** See stats(ds,lo,hi) and nlLH(ss). */
            public Vector stats(boolean add, Value ss0, Value ss1)
             { return ((Vector) ss0).combine(add, (Vector.Slice) ss1); }

            /** stats(ds,...) are the data-set. */
            public double nlLH(Value ss)
             { return sumNlPr((Vector) ss); }

            /** Also see NaiveBayes.M.condModel(id). */
            public MultiState.M condModel( Value given )
             { final Value.Tuple gT = (Value.Tuple) given;
               if( gT.nElts() != width() - 1 )
                  error( "given.nElts()=" + given.nElts() );
               final Value [] vs = new Value [width()];
               for( int i = 0; i < col; i ++ )
                  vs[i] = gT.elt(i);
               vs[col] = null;  // t.b.a. ...
               for( int i = col+1; i <= gT.nElts(); i ++ )
                  vs[i] = gT.elt(i-1);
               final Value.Tuple tu = Value.tuple(vs);
               final double [] nlPrs = new double [upb_n- lwb_n + 1];
               for( int i = lwb_n; i <= upb_n; i ++ ) //possibilities
                { vs[col] = cT.int2value(i);
                  nlPrs[i-lwb_n] = M.this.nlPr(tu);
                }
               return new MultiState(cT.bounds())
                        . sp2Model(0, 0, Vector.doubles( nlPrs2odds(nlPrs) ));
             }//condModel(given)

            /** Say who 'fm' is.  */
            public String toString()
             { return M.this.toString()
                    + ".functionModel(" + col + "," +  cT + ")";
             }//toString()
          };//fm

         return fm;
       }//functionModel(...)


    }//Multivariate.M (abstract) class, the Model

 }//Multivariate (abstract) class, the UPModel

// (c) Lloyd Allison, 2012, 23 Jan 2017.  Licence to 'Multivariate' is Gnu GPL.
