//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** An UnParameterised Normal Model for cases where {@link #mu &mu;}
    is the problem-defining parameter (common knowledge, given) and
    &sigma; alone is the {@link M#statParams statistical} parameter
    of the fully parameterised {@link NormalMu.M} to be
    {@link NormalMu#estimator estimated}.  We have the ready-made
    {@link MML#NormalMu0} for the <NOBR>&mu; = 0</NOBR> case.
    Also see {@link NormalUPM} and {@link NormalUPM.M}. */
public class NormalMu extends NormalUPM                             // NormalMu
 {
   /** The given, problem-defining parameter, &mu;, as a Cts. */
   protected final Cts muC;
   /** The given, problem-defining parameter, &mu;, as a double. */
   public final double mu;

   /** {@link #muC &mu;} is a given, the single
       problem-{@link #defnParams() defining} parameter. */
   public NormalMu(Value mu)
    { super();  // call the "special" constructor of NormalUPM
      this.muC = (Cts) mu;  this.mu = muC.x();
    }//constructor

   /** Return Real {@link #muC &mu;}, the problem defining parameter. */
   public Cts defnParams()
    { return muC; }

   /** apply(&sigma;), return N<SUB>{@link #mu &mu;},&sigma;</SUB>. */
   public M apply(Value sigma)
    { return sp2Model(0, 0, sigma); }   // an M, not just a Model

   /** Given msg1, msg2 and the (one) statistical parameter, &sigma;,
       return a fully parameterised Normal {@link M Model}. */
   public M sp2Model(double msg1, double msg2, Value sigma)
    { return new M(msg1,msg2, sigma); }

   /** Return an Estimator of a fully parameterised Normal {@link M Model}.
     Note, <NOBR>ps = [&sigma;<SUB>min</SUB>, &sigma;<SUB>max</SUB>]</NOBR>,
       the range for the {@link NormalMu.M#sigma std dev.,&nbsp;&sigma;}.
       Also see NormalUPM.{@link NormalUPM#estimator estimator(.)}. */
   public Continuous.Est estimator(final Value ps)
    { return new Est(ps); }
   // -----------------------------

   /** The standard Estimator for a Normal Model with specified &mu;.
       Also see {@link NormalUPM.Est}.  */
   public class Est extends Continuous.Est                   // NormalMu.Est
    { /** &sigma; &mem; [siMin, siMax]. */
      final double siMin, siMax;

      /** ps = (siMin, siMax). */
      public Est(Value ps)
       { super(ps);
         Tuple bds = ps.isTuple(2); // h(si) ~ 1/si, in bounds...
         siMin = bds.fst().x(); siMax = bds.snd().x();
         assert 0 < siMin && siMin < siMax : errMsg( " ps" );
       }//constructor

      /** Given ss = {@link #stats stats(ds)} for a data-set, ds,
          estimate a Model, i.e., &sigma;. */
      public M ss2Model(Value ss)
       { double R_si = Math.log(siMax) - Math.log(siMin), // h(s) ~ 1/s
                k1   = latticeConstant(1);
         Vector ssV = (Vector) ss;
         double N = ((Value.Cts) ssV.elt(0)) .x(),  // # data
              sum = ((Value.Cts) ssV.elt(1)) .x(),  // sum x[i]
            sumSq = ((Value.Cts) ssV.elt(2)) .x(),  // sum sqr(x[i])
            nlAoM = ((Value.Cts) ssV.elt(3)) .x();  // of data-set
         // Note, divisor is N, not N-1, because mu is a given! ...
         double sigma = N >= 1
                   ? Math.sqrt((sumSq-mu*sum) / N)         // est sigma
                   : Math.sqrt(siMin * siMax); // ?default
         sigma = Math.max(siMin, Math.min(siMax, sigma));  // & bounds
         M mdl = (M) sp2Model(0, 0, real(sigma));
         // h(si) = 1/(si R_si), where R_si the range of log si,
         // Fisher, F =  2 N / si^2    -- wrt sigma
         // msg1 = - log(h(si)) + 1/2 log F + 2/2 log k1
         //      = log(si R_si) + 1/2 log(2N/si^2) + log k1
         //            ^^NB.
         double msg1 = Math.log(R_si) + Math.log(N) + 0.5*log2
                     + Math.log(k1),
                msg2 = mdl.nlLH(ss) + 0.5;
         assert msg1 >= 0 && msg2 >= 0 : errMsg( " msg lengths" );
         return (M) mdl.asGiven(msg1, msg2);
       }//ss2Model(ss)
    }//Est class

   /** Return an Estimator of a fully parameterised Normal {@link M Model};
       note, <NOBR>t = triv</NOBR>!  The Estimator calls
       {@link NormalUPM.M#estSigma} to estimate &sigma;.  If you require
       more control, e.g., of the prior, implement another Estimator.
       Also see {@link #estimator estimator((lo,hi))}. */
   public Continuous.Est estimatorB(Value t /* = triv */)
    { if( t != triv ) error( ".estimator(" + t + ")" );
      return new Est(t)
       { public M ss2Model(Value ss)
          { final Cts sigma = NormalUPM
              .estSigma(ss, latticeConstant(1), mu, /* sigma1 */ 1.0);
            final double msg1=new NearInverse(triv).new M(0.1).nlPr(sigma),
                         msg2=new M(0.0, 0.0, sigma).nlLH(ss) + 0.5;
            return new M(msg1, msg2, sigma);
          }//ss2Model(ss)
       };//new Estimator(())
    }//estimatorB(t)

   /** M, a fully parameterised {@link NormalUPM.M Normal} Model with
       problem-{@link #defnParams defining} parameter {@link #mu &mu;},
       and {@link #statParams statistical} parameter &sigma;, as
       produced by the UnParameterised {@link NormalMu}.
       Also see {@link NormalUPM.M}.  */
   public class M extends NormalUPM.M                          // NormalMu.M
    { /** Given 1st and 2nd part message lengths, msg1 and msg2,
          and statistical parameter &sigma;  ({@link #mu &mu;} is the
          problem-{@link #defnParams defining} parameter), construct
          a Normal Model, N<SUB>&mu;,&sigma;</SUB>, of Cts data. */
      public M(double msg1, double msg2, Value sigma)
       { super(msg1, msg2, NormalMu.this.mu, sigma); // calls special constr
       }//constructor

      // NB. Gets stats(ds) from NormalMu (and ultimately NormalUPM),
      //     and ditto toString()

    }//NormalMu.M class, the Model

 }//NormalMu class, the UPModel

// (c) Lloyd Allison, 2012, Tues 3 December 2013. Licence is Gnu GPL.
