	//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Lexical;
import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The class of UnParameterised Normal Models (Gaussian distributions).
    For most purposes {@link MML#Normal} and {@link NormalUPM.M} should
    be enough, with little need to call upon NormalUPM directly? */
public class NormalUPM extends Continuous                          // NormalUPM
 {
   /** Constructor for UnParameterised Normal Models; note t=triv!
       There is little need to call it -- use the previously
       prepared {@link MML#Normal Normal}. */
   public NormalUPM(Value t)
     { super(t);
       if( t != triv ) error( " constructor(" + t + ")" );
     }//constructor, also see ...

   /** A constructor for any subclass of NormalUPM that deals with its
       own problem-defining parameter(s), e.g., {@link NormalMu NormalMu}.
       Note '(&nbsp;)', no problem defining parameter seen here. */
   protected NormalUPM()
     { super( null ); }//constructor

   /** apply((&mu;, &sigma;)), return
       a fully parameterised Normal {@link M M Model}. */
   public M apply(Value ms)
    { return sp2Model(0, 0, ms); }   // an M, not just a Model

   /** Given (msg1, msg2, (&mu;, &sigma;)) return
       a fully parameterised Normal {@link M M Model}. */
   public M sp2Model(double m1, double m2, Value mu_sig)
    { return new M(m1, m2, mu_sig); }

   /** Return the sufficient statistics of data-set ds;
       calls <TT>{@link NormalUPM.M#ss}(ds)</TT>.
       More on stats {@link UPModel#stats here}. */
   public Vector stats(Vector ds, int lo, int hi)
    { return ss(ds, lo, hi); }

   public Vector stats(boolean add, Value ss0, Value ss1)
    { return (Vector)
         ((Vector) ss0).bOp(add?Lexical.plus:Lexical.minus, (Vector) ss1);
    }

   /** Given a data-set, ds, return its sufficient statistics, ss,
       that is <NOBR>[# of items, sum, sum of squares, nlAoM}]</NOBR>.
       Static&nbsp;ss(ds) is called by {@link #stats stats(ds)}, and
       elsewhere.  (Beware of possible numerical stability problems
       if the data-set ds is huge.)  */
   public static Vector ss(Vector ds, int lo, int hi)
    { double N = 0, sum = 0, sumSq = 0;
      for( int i = lo; i < hi; i ++ )
       { final Value ei = ds.elt(i);
         final double wi = ds.wt(i), eix = ei.x();
         N += wi;  sum += wi * eix;  sumSq += wi * eix * eix;
         // If you are worried about the weights, an easy
         // case to think about is where wt(i)=2, say.
       }
      return Vector.doubles(new double[]{N, sum, sumSq, ds.wtNlAoM(lo,hi)});
    }//ss(ds)

   /** Return an Estimator for the Normal distribution where parameter
       <NOBR>ps = (&mu;<SUB>min</SUB>, &mu;<SUB>max</SUB>,
       &sigma;<SUB>min</SUB>, &sigma;<SUB>max</SUB>)</NOBR> is for the
       Uniform prior on &mu;, and 1/&sigma; on &sigma;, within the bounds.
       It tries to behave well in most situations. */
   public Continuous.Est estimator(final Value ps /* pair */)
    { return new Est(ps); }
   // ----------------------------

   /** The standard Estimator for a Normal Model with unknown
       &mu; and&nbsp;&sigma;. There is also a convenience function,
       {@link NormalUPM#estimator}.  Also see {@link NormalMu.Est}.  */
   public class Est extends Continuous.Est                  // NormalUPM.Est
    { /** &mu;<SUB>min</SUB>, &mu;<SUB>max</SUB> and
          &sigma<SUB>min</SUB>, &sigma<SUB>az/SUB>
          are bounds on &mu; and &sigma;. */
      final double muMin, muMax, siMin, siMax;

      /** ps = &lang; &mu;<SUB>min</SUB>, &mu;<SUB>max</SUB>,
                      &sigma;<SUB>min</SUB>, &sigma;<SUB>max</SUB> &rang;,
          uniform prior on &mu; and 1/&sigma; on &sigma;, in bounds.  */
      public Est(Value ps)
       { super(ps);
         Tuple bds = ps.isTuple(4); // mu uniform, sigma~1/sigma, in bounds
         muMin = bds.fst().x();  muMax = bds.snd().x();
         siMin = bds.thrd().x(); siMax = bds.frth().x();
         assert muMin < muMax && 0 < siMin && siMin < siMax
              : errMsg( ", ps" );
        }//constructor

      /** Given statistics, ss = {@link #stats stats(ds)}, of a
          data-set, ds, estimate an M, i.e., &mu; and &sigma;. */
      public M ss2Model(Value ss)
       { double R_mu = muMax - muMin,        // h(mu) uniform over R_mu
                R_si = Math.log(siMax) - Math.log(siMin), // h(s) ~ 1/s
                k2   = latticeConstant(2);
         final Vector ssV = (Vector) ss;
         double N = ((Value.Cts) ssV.elt(0)) .x(),  // # data
              sum = ((Value.Cts) ssV.elt(1)) .x(),  // sum x[i]
            sumSq = ((Value.Cts) ssV.elt(2)) .x(),  // sum sqr(x[i])
            nlAoM = ((Value.Cts) ssV.elt(3)) .x();  // of data-set
         double mu = N > 0 ? sum / N                  // est mu ~ mean
                   : (muMin + muMax) * 0.5;
         mu = Math.max(muMin, Math.min(muMax, mu));        // & bounds
         double sigma = N > 1
                   ? Math.sqrt((sumSq-mu*sum) / (N-1))    // est sigma
                   : Math.sqrt(siMin * siMax); // ?default
         sigma = Math.max(siMin, Math.min(siMax, sigma));  // & bounds
         M mdl = (M) sp2Model(0, 0, pair(real(mu), real(sigma)));
         // h(mu,si) = 1(R_mu si R_si), where R_si the range of log si,
         // Fisher, F = 2 N^2 / si^4    -- wrt mu,sigma
         // msg1 = -log(h(mu,si)) + 1/2 log F + 2/2 log k2
         //      = log(R_mu si R_si) + 1/2 log(2 N^2/si^4) + log k2
         //                 ^^NB.
         N = Math.max(N, 1); // for the log...
         double msg1 = Math.log(R_mu * R_si)
                     + 0.5*log2 + Math.log(N) - Math.log(sigma)
                     + Math.log(k2),
                msg2 = mdl.nlLH(ss) + 1;  // NB. +1 = 2*0.5
         assert msg1 >= 0 && msg2 >= 0 : "msg1=" + msg1 + ", msg2=" + msg2;
         return (M) mdl.asGiven(msg1, msg2);
       }//ss2Model(ss)
    }//Est class


   /** Return an Estimator for the Normal distribution where parameter
       <NOBR>bounds = (&mu;<SUB>min</SUB>, &mu;<SUB>max</SUB>)</NOBR>
       is for the Uniform prior on &mu;; &sigma; ~
       {@link NearInverse nearly&nbsp;1/&sigma;} and is unbounded.
       This estimator requires minimal user input (parameter 'bounds').
       It should not "blow up" even on zero data.  It tries to behave
       well in most situations. On the other hand if you do need more
       control, implement another estimator. The Estimator calls
       {@link NormalUPM.M#estSigma} to estimate &sigma;.
       Also see {@link #estimator estimator((,,,))}.  */
   public Continuous.Est estimatorB(final Value bounds /* pair */)
    { return new Est(bounds)
       { final Tuple bds = bounds.isTuple(2);
         final double muMin = bds.fst().x(), muMax = bds.snd().x();
          { assert muMin < muMax : errMsg( " ps" ); }
         /** The MML-estimator for &mu; and &sigma; of a
             {@link M Normal}-Model from the sufficient statistics
             <NOBR><TT>ss = {@link #stats stats(ds)}</TT></NOBR>
             of a data-set, ds.
             It calls {@link NormalUPM.M#estSigma} for &sigma;.
             <I>???The message length calculations
             need triple checking!!!</I>  */
         public M ss2Model(Value ss)
          { final double R_mu = muMax - muMin;
            final Vector ssV = (Vector) ss;
            double N = ((Value.Cts) ssV.elt(0)) .x(),  // # data
                 sum = ((Value.Cts) ssV.elt(1)) .x(),  // sum x[i]
            // sumSq = ((Value.Cts) ssV.elt(2)) .x()   // sum sqr(x[i])
               nlAoM = ((Value.Cts) ssV.elt(3)) .x();  // of data-set
            double mu = N <= 0 ? (muMin + muMax) / 2             // !
                      : Math.max(muMin, Math.min(muMax, sum/N)); // bounds
            // Fisher = 2 N^2 / sigma^4    -- wrt mu,sigma
            //  nlAoM = 0.5 log(Fisher) + (2/2)log k2
            //    AoM = exp(- nlAoM) = sqrt( sigma^4 / (2N^2 . k2) )
            final double k2 = latticeConstant(2);
            Value.Cts sigmaEst =
               estSigma(ss, k2, mu, /* sigma1 */ (muMin+muMax)/4);
            final double muAoM =
               Math.min(R_mu, sigmaEst.AoM()*Math.sqrt(2.0));
            final Value.Cts muEst = Value.cts(mu, muAoM);
            final double msg1 = Math.log(R_mu  / muAoM)
                    + Math.max(0,
                         new NearInverse(triv).new M(0.1).nlPr(sigmaEst)),
                  msg2 = new M(0, 0, Value.pair(muEst, sigmaEst)).nlLH(ss)
                       + 0.5;
            assert msg1 >= 0 && msg2 >= 0 : errMsg( " msg lengths" );
            return new M(msg1, msg2, Value.pair(muEst, sigmaEst));
          }//ss2Model(ss)
       };//new Estimator
    }//estimatorB(bounds);

   /** estSigma is called upon by
       NormalUPM.{@link NormalUPM#estimatorB estimatorB} and also by
       NormalMu.{@link NormalMu#estimatorB estimatorB} to estimate &sigma;.
       That is why estSigma is here and <TT>static</TT>.  Note, k is
       the appropriate lattice constant for two or one parameters.
       The returned value may not be less than 2&times; the per-datum AoM.
       If the number of data is &le;1, sigma1 is returned.  */
   public static Value.Cts estSigma(Value ss, double k, double mu,
                                    double sigma1)
    { Vector ssV = (Vector) ss;
      double N = ((Value.Cts) ssV.elt(0)) .x(),  // # data
         sumSq = ((Value.Cts) ssV.elt(2)) .x(),  // sum sqr(x[i])
         sigma, sigAoM, sigmaMin;
      if( N <= 0 ) sigmaMin = sigma1;
      else // N > 0
       { double nlAoM = ((Value.Cts) ssV.elt(3)) .x(); // of data-set
         sigmaMin = 2 * Math.exp( - nlAoM/N );    // 2* per datum AoM
       }
      // (sum_i (mu - x_i)^2)/N = (sum_i {mu^2 - 2 mu x_i + x_i^2})/N
      // = (N mu^2 - 2 N mu^2 + sum_i x_i^2)/N
      // = (sum_i x_i^2)/N - mu^2,  so ...
      double v = sumSq/N - mu*mu;  // the sample variance
      if( N <= 1 ) // tiny chance of 2 half weight data
       { sigma = Math.max(sigmaMin, sigma1);  sigAoM = sigma; }
      else // N > 1
       { sigma = Math.max(sigmaMin, Math.sqrt( v*N/(N-1) )); // NB. N-1
         sigAoM = 2 * sigma / Math.sqrt(N*k*2.0);
       }
      return Value.cts(sigma, sigAoM);
    }//estSigma()

   public String toString()
    { return this == Normal ? "Normal" : super.toString(); }
   // ..........................................

   /** NormalUPM.M, a fully parameterised Normal Model
       (Gaussian probability distribution), N<SUB>&mu;,&sigma;</SUB>.
       The statistical parameter(s) is a pair, that is a 2-tuple,
       &lang;&mu;,&nbsp;&sigma;&rang;.  Also see
       {@link NormalMu NormalMu}.{@link NormalMu.M <B>M</B>} for cases
       where &mu; is a <EM>given</EM>, and more on MML and the Normal:
       [<A HREF="http://www.allisons.org/ll/MML/Continuous/Normal/"
           TARGET="LL">www</A>].  */
   public class M extends Continuous.M                        // NormalUPM.M
    { /** The mean, &mu;, and standard deviation, &sigma;. */
      public final double mu, sigma;

      /** Useful, log(2&pi;)/2+log(&sigma;), and 2&sigma;<SUP>2</SUP>. */
      private double log_root_2_pi_sigma, two_sigma_2;

      /** Given the two-part message lengths, and
          statistical parameter <NOBR>(&mu; &sigma;)</NOBR>,
          construct N<SUB>&mu;&sigma;</SUB>. */
      public M(double msg1, double msg2, Value muSigma)
       { super(msg1, msg2, muSigma);
         Value.Tuple MSt = muSigma.isTuple(2);
         this.mu = MSt.fst().x();this.sigma = MSt.snd().x();
         setAndCheck();
       }//constructor

      /** A constructor with one statistical parameter, &sigma;,
          for use by {@link NormalMu.M}. */
      protected M(double msg1, double msg2, double mu, Value sigma)
       { super(msg1, msg2, sigma);
         this.mu = mu; this.sigma = ((Cts) sigma) .x();
         setAndCheck();
       }//constructor for NormalMu.M

      /** Must be called by some ultimate constructor
          to set M's private variables! */
      private void setAndCheck()                            // setAndCheck()
       { if( sigma <= 0 ) error( ", sigma <= 0" );
         log_root_2_pi_sigma = 0.5*(log2 + logPI) + Math.log(sigma);
      // log_root_2_pi_sigma = 0.5*Math.log(2 * Math.PI) + Math.log(sigma);
         two_sigma_2 = 2 * sigma * sigma;
       }//check()

      /** Return &nu; = {@link #sigma sigma}<SUP>2</SUP>. */
      public double variance()
       { return sigma * sigma; }

      /** The negative log pdf of a datum, x (double),
          for use by {@link #nlPdf nlPdf}. */
      public double nlPdf_x(double x)
       { final double x_mu = x - mu;
         return log_root_2_pi_sigma + x_mu * x_mu / two_sigma_2;
       }//nlPdf(d)

      /** Given ss = {@link NormalUPM#stats stats(ds)}, of a data-set, ds,
          return nlLH(ss), that is the negative log likelihood of ds. */
      public double nlLH(Value ss)
       { Vector ssV = (Vector) ss;
         final double N = ((Value.Cts) ssV.elt(0)) .x(),
                    sum = ((Value.Cts) ssV.elt(1)) .x(),
                  sumSq = ((Value.Cts) ssV.elt(2)) .x(),
                  nlAoM = ((Value.Cts) ssV.elt(3)) .x(), // of whole dataSet
                     mx = mu;
         return N * log_root_2_pi_sigma
              + (sumSq - 2*mx*sum + N*mx*mx) / two_sigma_2
              + nlAoM;  // of the whole dataSet
       }//nlLH(ss)

      /** Generate a <TT>double</TT> from N<SUB>&mu;,&sigma;</SUB>
          for use in {@link #random() random()}.  It calls Java's
          Random.nextGaussian() which "uses the polar method
          of G.E.P.Box, M.E.Muller and G.Marsaglia."  */
      public double random_x()
       { return mu + sigma * RNG.nextGaussian() /* ~N01 */ ; }

      // Generate a <TT>double</TT> from N(&mu;,&sigma;) for use in
      // {@link #random() random()}, presently by the
      // Box Muller method (Marsaglia modification)
      // [<A HREF="http://en.wikipedia.org/wiki/Gaussian_random_variable"
      //     TARGET="EX">wikip</A>].
      // p ublic double random_x()
      // { double U, V, R;
      //   while( true )
      //    { U = RNG.nextDouble()*2.0 - 1.0; // in [-1, 1) and in
      //      V = RNG.nextDouble()*2.0 - 1.0; // [-1, 1) make [-1,1)x[-1,1)
      //      R = U*U + V*V;     // radius**2
      //      if( R < 1 ) break; // i.e., break pi/4 of the time
      //    } // ... ? So, the probability of R = 0 (U=V=0) is zero ?! ...
      //   return mu.x() + sigma.x() * U * Math.sqrt(-2.0 * Math.log(R) / R);
      //   // ( R = 0  =>  log(R)/R = -infinity/zero = infinity  :-)
      // }//random_x()

    }//NormalUPM.M class, the Model

 }//NormalUPM class, the UPModel

// (c) Lloyd Allison, 2012, Tues 3 December 2013. Licence is Gnu GPL.
