//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.maths.Vector;

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The abstract class of UnParameterised Models over
    Permutations <NOBR>{0, ..., {@link #N() N}-1}</NOBR>. */
public abstract class Permutation extends UPModel              // * Permutation
 {
   public Permutation(Value dp) { super(dp); }//constructor

   /** Permutations of <NOBR>{0, ..., N()-1}</NOBR>. */
   public abstract int N();                                            // * N()
   // .....................................

   public abstract class M extends UPModel.M                 // * Permutation.M
    {
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Generate a Uniform random Permutation of
          <NOBR>{0, ..., N-1}</NOBR>. */
      public Vector random()
       { int [] p = new int[N()];
         random(p);
         return Vector.ints(p);
       }//random()

      /** Fill array p[] with a Uniform random Permutation of
          <NOBR>{0, ..., N-1}</NOBR>. */
      public abstract void random(int [] p);                   // * random(p[])

    }//Permutation.M (abstract) class, the fully parameterised Model
   // -------------------------------------------------------------------------

   /** The UnParameterised Uniform Model of Permutations.
       {@link #Mdl Mdl}:{@link Uniform.M M} is <I>the</I>
       fully parameterised Uniform Model. */
   public static class Uniform extends Permutation       // Permutation.Uniform
    {
      /** The problem-defining parameter is 'N' (here as an int)
          for Permutations of <NOBR>{0, ..., N-1}</NOBR>. */
      protected final int N;

      /** log(N!), the cost of stating a Permutation
          of <NOBR>{0, ..., N-1}</NOBR> uniformly. */
      protected final double log_fact_N;

      /** See {@link #N N}. */
      public Uniform(Value N)
       { super(N); this.N = ((Value.Int) N).n();
         log_fact_N = la.maths.Maths.logFactorial(this.N);
       }//constructor

      public int N() { return N; }

      /** sp2Model(0, m2, ()) returns a {@link M Model}. */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Given a data-set ds[lo,hi) of Permutations return
          sufficient statistics ss=ds.wts[lo,hi).
          Basically, how many are there. Can be fractional. */
      public Real stats(Vector ds, int lo, int hi)
       { return real(ds.wts(lo, hi)); }

      /** Combine sufficient statisticses 'ss0' and 'ss1'. */
      public Real stats(boolean add, Value ss0, Value ss1)
       { final double ss0x = ((Real) ss0).x(), ss1x = ((Real) ss1).x();
         return real(add ? ss0x + ss1x : ss0x - ss1x);  // combine (+/-)
       }//stats(add,ss0,ss1)

      /** The trivial Estimator that is a Uniform {@link #Mdl Model}'s. */
      public Estimator estimator(final Value t /* = triv */)
       { return Mdl.asEstimator(t); }

      /** <I>The</I> (trivially) fully parameterised Uniform Model
          of Permutations; also see its class,&nbsp;{@link M M}. */
      public final M Mdl = new M(0, 0, triv);        // Permutation.Uniform.Mdl
      // .............................

      /** {@link #Mdl Mdl} should be sufficient for most purposes, but
          here is the class of "fully parameterised" Uniform Models of
          Permutations. The class of UnParameterised Uniform Models
          is {@link Uniform}. */
      public class M extends Permutation.M             // Permutation.Uniform.M
       {
         /** Requires msg1=0 and sp=triv (trivial "statistical"
             parameters); also see {@link #Mdl Mdl}. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp); zeroTriv(msg1, sp); }//constructor

         /** The negative log probability of Permutation p,
             that is log&nbsp;N! */
         public double nlPr(Value p)
          { assert ((Vector) p).nElts() == N : "|p|=" + p.nElts();
            return log_fact_N;
          }//nlPr(p)

         /** The negative log likelhood of Permutations data-set 'ds'
             where ss={@link #stats stats(ds)}. */
         public double nlLH(Value ss) { return ((Cts) ss).x() * log_fact_N; }

         /** Fill array p[] with a Uniform random Permutation of
             <NOBR>{0, 1, ..., {@link #N N}-1}</NOBR>. */
         public void random(int [] p)
          { assert p.length == N : this + ".random(p[" + p.length + "])";
            Permutation.random(p);
          }//random(p)

       }//Permutation.Uniform.M, the fully parameterised Model

    }//Permutation.Uniform class, the UnParameterised Model
   // .........................................................................

   /** Fill array p[] with a <U>Uniform</U> random Permutation of
       <NOBR>{0, 1, ..., |p|-1}</NOBR>.  Note,&nbsp;static. */
   public static void random(int [] p)                             // random(p)
    { final int N = p.length;
      p[0] = 0;
      for( int i = 1; i < N; i ++ )
       { p[i] = i;  // 1, 2, ..., n-1
         final int s = (int) Math.floor(RNG.nextDouble() * (i+1));
         assert s <= i : "Permutation.random1(p), s=" + s + ", i=" + i;
         final int temp = p[s];  p[s] = p[i];  // ...
         p[i] = temp;                          // p[s] <-> p[i]
       }
    }//random(p)

 }//Permutation (abstract) class, the UnParameterised Model

// (c) L.Allison, 2014, Fri 20 Feb 2015. Licence to 'Permutation' is Gnu GPL.
