//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Function;
import la.maths.Vector;
import la.maths.Matrix;
import java.util.BitSet; // see ForestSearch
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** R_D, the (abstract) class of UnParameterised Models of D-dimensional
    Vectors of Cts, that is of <B>R</B><SUP>D</SUP>.
    Fully&nbsp;parameterised R_D Models start at {@link R_D.M}.  */
public abstract class R_D extends ByPdf                                // * R_D
 { /** Given problem-defining parameter(s), dp, construct an
       UnParameterised Model of Vectors in <B>R</B><SUP>D</SUP>. */
   public R_D(Value dp)
    { super(dp); }//constructor

   /** D(), the dimension of the data-space <B>R</B><SUP>D</SUP>.
       R_D Models are of fixed-length Vectors of Cts, nElts()==D().
       D()&nbsp;is logically part of the type of 'this'. You cannot
       count on learning D() from a data-set - it might be empty. */
   public abstract int D();                                            // * D()

   /** Calculate all per-column Normal-{@link NormalUPM#stats stats}.
       This may be of use in various R_D models including generalisations
       of, and variations on, {@link Forest}.  NB.&nbsp;static.
       Also see {@link #ds2L1stats} and {@link #ds2NorL1stats}.  */
   public static Vector ds2Nstats(Vector ds)                   // ds2Nstats(ds)
    { int D = ds.nCols();
      Value [] ssN = new Value [D];
      for( int i = 0; i < D; i ++ )
         ssN[i] = MML.Normal.stats(ds.col(i)); // Normal dist'n for 'i'
      return Vector.values(ssN);
    }//ds2Nstats(ds)

   /** Calculate all per-two-column Linear1-{@link Linear1#stats stats}.
       This may be of use in various Linear models including generalisations
       of, and variations on, {@link Forest}.  NB.&nbsp;static.
       Also see {@link #ds2Nstats} and {@link #ds2NorL1stats}.  */
   public static Matrix ds2L1stats(Vector ds)                  //ds2L1stats(ds)
    { final int D = ds.nCols();
      Value [] [] ssL = new Value [D] [D];
      final Vector dummy = Vector.doubles( new double [] { -99.0 } );
      for( int i = 0; i < D; i ++ )
         for( int j = 0; j < D; j ++ )
            if( i == j ) ssL[i][j] = dummy;
            else // Linear1 for 'i -> j'
               ssL[i][j] = MML.Linear1.stats( ds.col(i).zip(ds.col(j)) );
      return Matrix.values(ssL);
    }//ds2L1stats(ds)

   /** Calculate Normal-{@link NormalUPM#stats stats} for each
       "parent-less" column and Linear1-{@link Linear1#stats stats}
       for each column that has a 'parent' column. If parent[i]&lt;0
       then column 'i' has no parent. Result is a pair, a Vector for
       the Normals and a Vector for the Linear1s. Used by {@link Forest}
       and might be useful in other Linear models. NB.&nbsp;static.
       Also see {@link #ds2Nstats} and {@link #ds2L1stats}.  */
   public static Tuple ds2NorL1stats(Vector ds, int[] parent)  // ds2NorL1stats
    { final int D = parent.length;
      if( D != ds.nCols() ) la.util.Util.error(
                    "ds.nCols()=" + ds.nCols() + "!=" + "parent.length=" + D );
      int nChildren = 0, nN = 0, nL = 0;
      for( int i = 0; i < D; i ++ ) if( parent[i] >= 0 ) nChildren ++ ;
      Value [] ssN = new Value [D-nChildren],
               ssL = new Value [  nChildren];
      for( int i = 0; i < D; i ++ ) // each column 'i'
       { final int pi = parent[i];
         if( pi >= 0 ) // pi -> i
          { ssL[nL] = MML.Linear1.stats( ds.col(pi).zip(ds.col(i)) ); nL++ ; }
         else // i is parent-less
          { ssN[nN] = MML.Normal.stats(ds.col(i)); nN ++ ; }
       }//for i
      return pair(Vector.values(ssN), Vector.values(ssL));
    }//ds2NorL1stats(ds,parent)
   // ................................

   /** (abstract) M, fully parameterised Models of <B>R</B><SUP>D</SUP>.
       A subclass of R_D.M gets {@link R_D.M#D()} and {@link M#stats},
       "for free", from its enclosing UPModel.  */
   public abstract class M extends ByPdf.M                           // * R_D.M
    { public M(double msg1, double msg2, Value dp)
       { super(msg1, msg2, dp); }//constructor

      /** R_D.this.{@link R_D#D() D()}, the dimension, D, of
          <B>R</B><SUP>D</SUP>. */
      public int D() { return R_D.this.D(); }                            // D()

   // public abstract   pdf(d)   from ByPdf.M                              // *
   // public abstract nlPdf(d)   from ByPdf.M                              // *

      /** Return a random Vector in
          <B>R</B><SUP>{@link #D() D()}</SUP>
          (<I>calls {@link #random_x()}</I>).  */
      public Vector random()                                        // random()
       { return Vector.doubles(random_x()); }

      /** Return the components of a random Vector
          (<I>calls {@link #random(double[])}</I>).
          Called by (default) {@link #random()}.  */
      public double [] random_x()
       { final double [] x = new double[ D() ];
         random(x);
         return x;
       }//random_x()

      /** Optional - this default throws an Exception, may be overridden.
          If overridden, leave the components of a random Vector in x[].
          Called by (default) {@link #random_x()}.  */
      public void random(double [] x)
       { throw UOE( ".random(double[]) is not yet(?) supported, sorry" ); }

    }//R_D.M (abstract) class, the Model(s)
   // -------------------------------------------------------------------------

   /** An UnParameterised Model of Vectors in <B>R</B><SUP>D</SUP> made from
       UnParameterised Models of norms (lengths) and of Directions.
       The fully parameterised Model is {@link NrmDir.M} */
   public static class NrmDir extends R_D                         // R_D.NrmDir
    {
      /** The dimension, D, of the data-space, <B>R</B><SUP>D</SUP>, i.e.,
          {@link #dirnUPM dirnUPM}'s {@link Direction#D() D()}. */
      public final int D;

      /** {@link #normUPM normUPM} is the UnParameterised
          Continuous Model of norm (length, magnitude, ...). */
      public final UPModel normUPM;

      /** dirnUPM is the UnParameterised Model of
          {@link Direction Direction} in <B>R</B><SUP>D</SUP>. */
      public final Direction dirnUPM;

      /** Given dp&nbsp;=&nbsp;(normUPM,&nbsp;dirnUPM),
          construct the UnParameterised Model. */
      public NrmDir(Value dp)
       { super(dp);
         Value.Tuple dpT = dp.isTuple(2);
         normUPM = (UPModel)   dpT.fst();  // of norms
         dirnUPM = (Direction) dpT.snd();  // of Directions
         D = dirnUPM.D();                  // Dimension of R^D
       }//constructor

      /** Return {@link #D D}, the dimension of
          the data-space, <B>R</B><SUP>D</SUP>. */
      public int D() { return D; }

      /** Given 1st &amp; 2nd part message lengths, m1 &amp; m2, and
          statistical parameters, sp&nbsp;=&nbsp;(norm's,&nbsp;dirn's),
          return a fully parameterised Model. */
      public Model sp2Model(double m1, double m2, Value sp)
        { return new M(m1, m2, sp.isTuple(2)); }

      /** The sufficient statistics, ss&nbsp;=&nbsp;stats(ds), of a data-set,
          ds.  ss&nbsp;=&nbsp;(norm&nbsp;stats,&nbsp;dirn&nbsp;stats).
          Calls upon stats of {@link #normUPM} and {@link #dirnUPM},
          and may or may not equal {@link M#stats(Vector,int,int)}.
          More on stats {@link UPModel#stats here}. */
      public Value stats(Vector ds, int lo, int hi)
       { if( ds.nElts() > 0 && ds.nCols() != D() )
            error( ".stats(ds), nCols=" + ds.nCols() + ", D=" + D() );
         Vector nrms = ds.map( new Function()
             { public Value apply(Value v)
                { return ((Vector) v).norm_Cts(); }
             }/*new Fn*/     );       // NB. map result is weightLike(ds) !
         Vector nvs  = ds.map( new Function() // ???can we get around this?
             { public Vector apply(Value v)
                { return ((Vector) v).normalised(); }
             }/*new Fn*/     );       // NB. map result is weightLike(ds) !
         return Value.pair( normUPM.stats(nrms,lo,hi),
                            dirnUPM.stats(nvs, lo,hi) );
       }//stats(ds,lo,hi)

      /** Calls upon stats of {@link #normUPM} and {@link #dirnUPM}, and
          may or may not equal {@link M#stats(boolean,Value,Value)}. */
      public Value stats(boolean add, Value ss0, Value ss1)
       { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
         return Value.pair( normUPM.stats(add, ss0t.fst(), ss1t.fst()),
                            dirnUPM.stats(add, ss0t.snd(), ss1t.snd()) );
       }//stats(add,ss0,ss1)

      /** Return an Estimator; its parameter
          ps&nbsp;=&nbsp;(ps<SUB>n</SUB>,&nbsp;ps<SUB>d</SUB>) where
          ps<SUB>n</SUB> is for {@link #normUPM normUPM}'s estimator and
          ps<SUB>d</SUB> is for {@link #dirnUPM dirnUPM}'s.  */
      public Est estimator(Value ps)
       { final Value.Tuple psT = ps.isTuple(2);
         return new Est(ps)
          { public M ss2Model(Value ss)
             { Value.Tuple ssT = ss.isTuple(2);
               Model
                  normM = normUPM.estimator(psT.fst()).ss2Model(ssT.fst()),
                  dirnM = dirnUPM.estimator(psT.snd()).ss2Model(ssT.snd());
               return new M(normM.msg1() + dirnM.msg1(),
                            normM.msg2() + dirnM.msg2(),
                      Value.pair(normM.statParams(), dirnM.statParams()));
             }//ss2Model(ss)
       }; }//estimator(ps)
      // .................................

      /** A fully parameterised Model of Vectors
          in <B>R</B><SUP>D</SUP>, made from a Model of
          {@link Vector#norm norms} (Vector lengths) and a Model of
          {@link Direction Directions}.  To be an efficient code the
          {@link #normMdl former} would assume and make use of
          v.norm()&nbsp;&ge;&nbsp;0, for a Vector&nbsp;v. */
      public class M extends R_D.M                              // R_D.NrmDir.M
       { /** A fully parameterised  Model of Vector
             {@link Vector#norm norms} (lengths).
             Note that v.norm()&nbsp;&ge;&nbsp;0, of course. */
         public final Continuous.M normMdl;

         /** A fully parameterised  Model of Directions, that is of
             <B>R</B><SUP>D</SUP> Vector Directions. */
         public final Direction.M dirnMdl;

         /** Two-part message lengths, msg1 and msg2, and statistical
             parameters sps&nbsp;=&nbsp;(sp<SUB>n</SUB>,&nbsp;sp<SUB>d</SUB>)
             where sp<SUB>n</SUB> is {@link #normMdl normMdls}'s sp
             and   sp<SUB>d</SUB> is {@link #dirnMdl dirnMdls}'s.  */
         public M(double msg1, double msg2, Value sps)
          { super(msg1, msg2, sps);
            final Value.Tuple spT = sps.isTuple(2);
            normMdl = (Continuous.M) normUPM.apply(spT.fst());
            dirnMdl = (Direction.M)  dirnUPM.apply(spT.snd());
          }//constructor

         /** The negative log probability density of a Vector datum, v, in
             <B>R</B><SUP>D</SUP>.  Note, clearly v is not normalised in
             general, so the Direction {@link Direction.M#nlPdf density}
             is adjusted to a "radius" of v.norm(). */
         public double nlPdf(Value v)
          { final Vector  vv = (Vector)v;
            if( vv.nElts() != D() )
               error( ", v.nElts()=" + vv.nElts() );
            final double nrm = vv.norm();
            return normMdl.nlPdf_x(nrm)        // norm + ...
                 + dirnMdl.nlPdf(v)            // dirn adjusted...
                 + (D() - 1) * Math.log(nrm);  // ...for norm!
          }//nlPdf(v)

      // public double nlPr(v), from R_D.M and ultimately from ByPdf.M

         /** The statistics, ss&nbsp;=&nbsp;stats(ds), of a data-set, ds.
             ss&nbsp;=&nbsp;(norm&nbsp;stats,&nbsp;dirn&nbsp;stats).
             Calls upon stats of {@link #normMdl} and {@link #dirnMdl}, and
             may or may not equal {@link NrmDir#stats(Vector,int,int)}.
             Also see {@link M#nlLH M.nlLH(ss)}.
             More on stats {@link Model#stats here}. */
         public Value stats(Vector ds, int lo, int hi)
          { if( ds.nElts() > 0 && ds.nCols() != D() )
               error( ".stats(ds), nCols=" + ds.nCols() + ", D=" + D() );
            Vector nrms = ds.map( new Function()
                { public Value apply(Value v)
                   { return ((Vector) v).norm_Cts(); }
                }/*new Fn*/     );       // NB. map result is weightLike(ds) !
            Vector nvs  = ds.map( new Function() // ???can we get around this?
                { public Vector apply(Value v)
                   { return ((Vector) v).normalised(); }
                }/*new Fn*/     );       // NB. map result is weightLike(ds) !
            return Value.pair( normMdl.stats(nrms,lo,hi),
                               dirnMdl.stats(nvs, lo,hi) );
          }//stats(ds,lo,hi)

         /** Calls upon stats of {@link #normMdl} and {@link #dirnMdl},
             and may or may not equal
             {@link NrmDir#stats(boolean,Value,Value)}. */
         public Value stats(boolean add, Value ss0, Value ss1)
          { Value.Tuple ss0t = ss0.isTuple(2), ss1t = ss1.isTuple(2);
            return Value.pair( normMdl.stats(add, ss0t.fst(), ss1t.fst()),
                               dirnMdl.stats(add, ss0t.snd(), ss1t.snd()) );
          }//stats(add,ss0,ss1)

         /** Given sufficient statistics,
             ss&nbsp;=&nbsp;{@link #stats stats(ds)}, of a
             data-set, ds, return the negative log likelihood of ds. */
         public double nlLH(Value ss)
          { Value.Tuple ssT = ss.isTuple(2); // a pair(for_norm, for_dirn)
            return normMdl.nlLH(ssT.fst()) + dirnMdl.nlLH(ssT.snd());
          }//nlLH(ss)

         /** Generate a random <B>R</B><SUP>D</SUP>-Vector from 'this' Model.
             It needs the sub-Models to implement random. */
         public Vector random()
          { return ((Vector)dirnMdl.random()) .scaled(normMdl.random_x()); }

       }//R_D.NrmDir.M class, the Model

    }//R_D.NrmDir class, the UPModel
   // -------------------------------------------------------------------------

   /** A (UnParameterised) model in which a column may be
       {@link Linear1 linearly} dependent on a single
       {@link #parent parent} column.
       The fully parameterised model is {@link Forest.M}.
       Also see {@link ForestSearch}.  */
   public static class Forest extends R_D                         // R_D.Forest
    {
      /** If parent[i] &ge; 0 then column 'i' has that parent column. */
      public final int [] parent;

      /** The number of columns that have a {@link #parent parent}. */
      public final int nChildren;

      /** The number of columns, the dimension, D, of <B>R</B><SUP>D</SUP>. */
      public final int D;

      /** The dimension, {@link #D}, of <B>R</B><SUP>D</SUP>. */
      public final int D() { return D; }

      /** Problem definition parameter 'dp' gives {@link #parent parent[]},
          must specify a forest. If parent[i]&lt;0 then column 'i'
          has no parent.  */
      public Forest(Value dp)
       { super(dp);
         Vector dpV = (Vector) dp;
         D = dpV.nElts();
         parent = new int [D];
         int nC = 0;
         for( int i = 0; i < D; i ++ )
          { int pi = dpV.elt(i).n();  parent[i] = pi; if( pi >= 0 ) nC ++ ; }
         nChildren = nC;
       }//constructor

      /** Statistical parameters, a pair &ndash; a Vector for the
          {@link M#Nrml Normals} and another for the {@link M#Lin1 Linear1s}.
          Uses {@link #ds2NorL1stats(Vector,int[])}.  */
      public Tuple stats(Vector ds, int lo, int hi)
       { return ds2NorL1stats(ds.slice(lo, hi), parent); }

      public Tuple stats(boolean add, Value ss0, Value ss1)
       {return (Tuple)ss0.bOp(add?la.la.Lexical.plus:la.la.Lexical.minus,ss1);}

      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Estimator for a fully {@link M parameterised} Forest model. */
      public Est estimator(Value ps)
       { Tuple psT = ps.isTuple(6); // repack for Normals and Linear1s ...
         final Tuple psN = quad(psT.fst(), psT.snd(),  psT.thrd(),psT.frth());
         final Tuple psL = quad(psT.elt(4),psT.elt(5), psT.thrd(),psT.frth());
         return new Est(ps)
          { public M ss2Model(Value ss)
             { Tuple ssT = ss.isTuple(2);
               Vector ssN = (Vector) ssT.fst(), ssL = (Vector)ssT.snd();
               double msg1 = 0, msg2 = 0;
               Value [] spN = new Value [D-nChildren],
                        spL = new Value [    nChildren];
               for( int i = 0; i < D-nChildren; i ++ ) // the Normals...
                { Model Nrml = MML.Normal.estimator(psN).ss2Model(ssN.elt(i));
                  spN[i] = Nrml.statParams();
                  msg1 += Nrml.msg1;  msg2 += Nrml.msg2();
                }
               for( int i = 0; i < nChildren; i ++ )  // the Linear1s...
                { Model Lin1 = MML.Linear1.estimator(psL).ss2Model(ssL.elt(i));
                  spL[i] = Lin1.statParams();
                  msg1 += Lin1.msg1;  msg2 += Lin1.msg2();
                }
               return Forest.this.sp2Model(msg1, msg2,
                         pair(Vector.values(spN), Vector.values(spL)));
             }//ss2Model(ss)
          };//new Est(ps)
       }//estimator(ps)
      // ...............................

      /** Fully parameterised {@link Forest}-model.  Each column
          may have at most one {@link #parent parent} column.  */
      public class M extends R_D.M                              // R_D.Forest.M
                     implements HasPdf
       {
         /** There is one Normal model in Nrml[] per
             {@link #parent parent}-less column, in order.
             Nrml.length = {@link Forest#D D}-{@link #nChildren nChildren}.  */
         public final NormalUPM.M  [] Nrml = new NormalUPM.M  [D-nChildren];
         /** There is one Linear1 model n Lin1[] per
             {@link #parent parent}-ed column, in order.
             Lin1.length = {@link #nChildren nChildren}.  */
         public final Linear1.M [] Lin1 = new Linear1.M [nChildren];

         /** 'sp' is a pair, a Vector of parameters for {@link #Nrml Normal}-
             and a Vector of parameters for {@link #Lin1 Linear1}-models. */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            Tuple spT = sp.isTuple(2);
            Vector spN = (Vector) spT.fst(), spL = (Vector) spT.snd();
            for( int i = 0; i < D - nChildren; i ++ )
               Nrml[i] = MML.Normal.new M(0, 0, spN.elt(i));
            for( int i = 0; i < nChildren; i ++ )
               Lin1[i] = MML.Linear1.new M(0, 0, spL.elt(i));
          }//constructor

         /** Negative log probability density of a
             {@link #D() D}-element datum v. */
         public double nlPdf(Value v)
          { Structured vv = (Structured) v;
          // Vector vv = (Vector) v;
            if( vv.nElts() != D ) error( ", v.nElts()=" + vv.nElts() );
            double ans = 0;  int nN = 0, nL = 0;
            for( int i = 0; i < D; i ++ )
             { final int pi = parent[i];
               if( pi < 0 )
                { ans += Nrml[nN].nlPdf(vv.elt(i)); nN ++ ; }
               else
                { ans += Lin1[nL].nlPdf(pair(vv.elt(pi), vv.elt(i))); nL ++ ; }
             }
            return ans;
          }//npPdf(v)

         /** Negative log likelihood for data-set ds where
             ss={@link #stats(Vector,int,int) stats(ds)}. */
         public double nlLH(Value ss)
          { Tuple ssT = ss.isTuple(2);
            Vector ssN = (Vector)ssT.fst(), ssL = (Vector) ssT.snd();
            double ans = 0;
            for( int i = 0; i < D-nChildren; i ++ )
               ans += Nrml[i].nlLH(ssN.elt(i));
            for( int i = 0; i < nChildren; i ++ )
               ans += Lin1[i].nlLH(ssL.elt(i));
            return ans;
          }//nlLH(v)

         /** TODO random(xs) not yet tested??? */
         public void random(double [] xs)
          { int ni = 0, li = 0;
            for( int i = 0; i < D(); i ++ )  // first, the parent-less
               if( parent[i] < 0 )
                { xs[i] = Nrml[ni].random_x(); ni ++ ; }
            for( int i = 0; i < D(); i ++ )  // second, the with-parent
             { int pi = parent[i];
               if( pi >= 0 )
                { xs[i] = Lin1[li].random(Value.real(xs[pi])).x(); li ++ ; }
             }
          }//random(xs)

       }//R_D.Forest.M class, the fully parameterised Model

    }//R_D.Forest class, the UnParametrised UPModel
   // -------------------------------------------------------------------------

   /** A generalisation of {@link R_D.Forest} that is able to estimate the
       model's <I>structure</I> &ndash; the child-{@link M#parent parent}
       relations &ndash; as well as the parameters of the component
       distributions.  Note that the "estimating"
       {@link #stats(la.maths.Vector,int,int) statistics} and the "likelihood"
       {@link M#stats(Vector,int,int) statistics} differ considerably.
       The fully parameterised model is {@link ForestSearch.M}.  */
   public static class ForestSearch extends R_D             // R_D.ForestSearch
    {
      /** A datum is a Vector of 'D' Continuous Values. */
      public final int D;
      /** Return {@link #D}.  */
      public int D() { return D; }

      /** 'dp' is the dimension, "width", {@link #D}. */
      public ForestSearch(Value dp)
        { super(dp); D = ((Int) dp).n(); }

      /** Sufficient statistics 'ss' are
          &langle;D&times; stats for Normals,
          D&times;(D-1) stats for Linear1s&rangle;.
          This is much more than {@link M#stats(Vector,int,int)}
          because the structure is not yet decided.
          (Uses {@link #ds2Nstats} and {@link #ds2L1stats}.)  */
      public Value stats(Vector ds, int lo, int hi)
       { Vector ds_lo_hi = ds.slice(lo, hi);
         return pair(ds2Nstats(ds_lo_hi), ds2L1stats(ds_lo_hi));
       }//stats(ds,lo,hi)

      /** Combine {@link #stats(Vector,int,int) statistics} ss0 and ss1,
          either positively (add=true) or negatively (add=false).  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return ss0.bOp(add?la.la.Lexical.plus:la.la.Lexical.minus, ss1); }

      /** Return a fully parameterised {@link M model}. */
      public M sp2Model(double msg1, double msg2, Value sp)
       { return new M(msg1, msg2, sp); }

      /** Parameter 'ps' is bounds on &mu;, &sigma;, and b. A greedy search
          is used to find parent-child relationships and component models.
          Note that {@link M#parent parent[]} must represent a Forest.
          Also see {@link NormalUPM#estimator} and
          {@link Linear1#estimator}. */
      public Est estimator(Value ps)
       { Tuple psT = ps.isTuple(6); // repack for Normals and Linear1s ...
         final Tuple psN = quad(psT.fst(),  psT.snd(),  psT.thrd(), psT.frth());
                         //         ^ mu bounds ^          ^ sigma bounds ^
         final Tuple psL = quad(psT.elt(4), psT.elt(5), psT.thrd(), psT.frth());
                         //         ^ b bounds  ^          ^ sigma bounds ^
         return new Est(ps)
          { public M ss2Model(Value ss)
             { Tuple  ssT = ss.isTuple(2);
               Vector ssN = (Vector) ssT.fst();
               Matrix ssL = (Matrix) ssT.snd();
               int [] parent = new int [D];
               BitSet [] descendants = new BitSet [D];
               Model    [] NM = new Model [D];      // Normals
               Model [] [] LM = new Model [D] [D];  // Linear1s, i -> j
               Estimator NE = MML.Normal.estimator(psN),
                         LE = MML.Linear1.estimator(psL);
               for( int i = 0; i < D; i ++ )        // compute Models ...
                { parent[i] = -1;  // i.e., initially no parents
                  descendants[i] = new BitSet(D); descendants[i].set(i); // {i}
                  NM[i] = NE . ss2Model(ssN.elt(i));          // Normal for 'i'
                  for( int j = 0; j < D; j ++ )
                     if( i == j ) LM[i][j] = null;
                     else // i != j
                        LM[i][j] = LE.ss2Model(ssL.elt(i, j)); // Linear i->j
                }//for i

               int nChildren = 0;
               while( true )                                   // greedy search
                { double bestSaving = 0;  // so far
                  int p = -1, c = -1;     // parent -> child ?
                  for( int j = 0; j < D; j ++ )     // all parent-less *j*
                   { if( parent[j] >= 0 ) continue; // j already has a parent
                     double baseLine = NM[j].msg(); // Normal for j
                     for( int i = 0; i < D; i ++ )  // non-descendants *i* of j
                      { if( descendants[j].get(i) ) continue; // ?acyclic
                        double saving = baseLine - LM[i][j].msg(); // i -> j
                        if( saving > bestSaving )
                         { bestSaving = saving; p = i; c = j; } // p -> c saves
                      }//for i
                   }//for j
                  if( bestSaving <= 0 ) break; // no improvement, done, else...
                  parent[c] = p;  nChildren ++ ;
                  for( int i = 0; i < D; i ++ )       // make transitive...
                     if( descendants[i].get(p) )
                        descendants[i].or(descendants[c]);
                }//while. 'parent[]' represents a forest over the cols [0,D).

               double msg1 = D * Math.log(D+1) /* for parent[] */,  msg2 = 0;
               Value [] spN = new Value [D - nChildren],
                        spL = new Value [nChildren];
               int nN = 0, nL = 0; // step thru Nrmls and Linears
               for( int i = 0; i < D; i ++ )
                  if( parent[i] < 0 )  // none, use Normal for 'i'
                   { msg1 += NM[i].msg1();  msg2 += NM[i].msg2();
                     spN[nN] = NM[i].statParams(); nN ++ ;
                   }
                  else // >= 0         // use a Linear1 parent[i] -> i
                   { Model lm = LM [parent[i]] [i];
                     msg1 += lm.msg1();  msg2 += lm.msg2();
                     spL[nL] = lm.statParams(); nL ++ ;
                   }
               Tuple sp = triple( Vector.ints(parent),  // the forest,
                                  Vector.values(spN),   // sp of Normals,
                                  Vector.values(spL) ); // sp of Linears
               return ForestSearch.this.sp2Model(msg1, msg2, sp);
             }//ss2Model
          };//new Estimator
       }//estimator(ps)
      // ...................................................

      /** The fully parameterised {@link ForestSearch}-model.  Note
          {@link ForestSearch.M#stats(Vector,int,int)} in particular.  */
      public class M extends R_D.M                            // ForestSearch.M
       { /** If parent[i]&ge;0 then variable i has that parent. */
         public final int []          parent = new int [D];
         /** The number of variables that have a parent. */
         public final int             nChildren;
         /** The models for the parent-less variables.   */
         public final NormalUPM.M  [] Nrml;
         /** The models for the with-parent variables.   */
         public final Linear1.M [] Lin1;

         /** 'sp' contains [p0, p2, ..., p(D-1)] where p_i is the parent
             of 'i', a negative value indicates that p_i has "no parent",
             parameters &langle;&mu;,&sigma;&rangle; of the various
             parent-less N_(&mu;,&sigma;), and &langle;a,b,s&rangle; of
             the parented Linear1_(a,b,s).  */
         public M(double msg1, double msg2, Value sp)
          { super(msg1, msg2, sp);
            Tuple  spT = sp.isTuple(3);
            Vector pV  = (Vector) spT.fst();
            Vector spN = (Vector) spT.snd(), spL = (Vector) spT.thrd();
            int nC = 0;
            for( int i = 0; i < D; i ++ )
             { parent[i] = pV.elt(i).n(); if( parent[i] >= 0 ) nC ++ ; }
            nChildren = nC;
            if( pV.nElts() != D || spN.nElts() != D-nC || spL.nElts() != nC )
               error( ", in statistical parameters " + sp );
            Nrml = new NormalUPM.M [D-nChildren];
            Lin1 = new Linear1.M[  nChildren];
            for( int i = 0; i < D - nChildren; i ++ )
               Nrml[i] = MML.Normal.new M(0, 0, spN.elt(i));
            for( int i = 0; i < nChildren; i ++ )
               Lin1[i] = MML.Linear1.new M(0, 0, spL.elt(i));
          }//constructor

         /** Note that these M.stats(...) differs from
             {@link ForestSearch#stats(Vector,int,int)} because
             {@link #parent parent[]} is <I>known</I> in&nbsp;M.
             Uses {@link #ds2NorL1stats}.  */
         public Tuple stats(Vector ds, int lo, int hi)            // stats(...)
          { return ds2NorL1stats(ds.slice(lo, hi), parent); }  // !

         /** Combine (+/-) two
             {@link ForestSearch.M#stats(Vector,int,int) statistics},
             ss0 and ss1. */
         public Tuple stats(boolean add, Value ss0, Value ss1)
          {return (Tuple)
              ss0.bOp(add?la.la.Lexical.plus:la.la.Lexical.minus,ss1);
          }//stats(add,ss0,ss1)

         public double nlPdf(Value v)                               // nlPdf(v)
          { Structured vv = (Structured) v;
            if( vv.nElts() != D ) error( ", v.nElts()=" + vv.nElts() );
            double ans = 0;  int nN = 0, nL = 0;
            for( int i = 0; i < D; i ++ )
             { final int pi = parent[i];
               if( pi < 0 )
                { ans += Nrml[nN].nlPdf(vv.elt(i)); nN ++ ; }
               else
                { ans += Lin1[nL].nlPdf(pair(vv.elt(pi), vv.elt(i))); nL ++ ; }
             }
            return ans;
          }//nlPdf(v)

         /** Negative log likelihood for data-set ds where
             ss={@link #stats(Vector,int,int) stats(ds)}. */
         public double nlLH(Value ss)                               // nlLH(ss)
          { Tuple ssT = ss.isTuple(2);
            Vector ssN = (Vector)ssT.fst(), ssL = (Vector) ssT.snd();
            double ans = 0;
            for( int i = 0; i < D-nChildren; i ++ )
               ans += Nrml[i].nlLH(ssN.elt(i));
            for( int i = 0; i < nChildren; i ++ )
               ans += Lin1[i].nlLH(ssL.elt(i));
            return ans;
          }//nlLH(v)

         /** TODO &ndash; not tested!  */
         public void random(double[] xs)
          { int ni = 0, li = 0;
            for( int i = 0; i < D(); i ++ )  // first, the parent-less
               if( parent[i] < 0 )
                { xs[i] = Nrml[ni].random_x(); ni ++ ; }
            for( int i = 0; i < D(); i ++ )  // second, the with-parent
             { int pi = parent[i];
               if( pi >= 0 )
                { xs[i] = Lin1[li].random(Value.real(xs[pi])).x(); li ++ ; }
             }
          }//random(xs)

       }//R_D.ForestSearch.M class, the fully parameterised Model

    }//R_D.ForestSearch class, the (UnParameterised) UPModel
   // -------------------------------------------------------------------------

 }//R_D (abstract) class, the UnParameterised UPModel(s)

// (c) Lloyd Allison, 2012, Mon 4 Dec 2017. Licence to 'R_D' is Gnu GPL.
