//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.la.Value;
import la.la.Function;
import la.maths.Vector;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** UPModel, the abstract class of UnParameterised statistical Models.
    Fully parameterised Models are {@link UPModel.M} and {@link Model};
    also see {@link Estimator}.  <I>Statistical</I> parameters are
    not yet set in a UPModel but it holds any common-knowledge,
    problem-{@link #defnParams defining} parameters, and returns a Model when
    it is {@link #apply applied} to statistical-parameter(s) sp (or calls
    <SPAN STYLE="white-space:nowrap">
       {@link #sp2Model sp2Model}(
       {@link Model#msg1() msg1},
       {@link Model#msg2() msg2},
       {@link Model#statParams() sp})),</SPAN>
    and hopefully can also {@link #estimator estimate} a
    fully parameterised {@link UPModel.M Model} from a given data-set.<BR>
    Useful examples include {@link MML#Normal Normal},
    {@link NormalMu NormalMu}, {@link MML#Poisson0 Poisson0},
    {@link MML#Geometric0 Geometric0}, and {@link Multivariate Multivariate}.
    Also see {@link UPFunctionModel UPFunctionModel} and
    {@link UPSeriesModel UPSeriesModel}.  */
public abstract class UPModel extends Function                     // * UPModel
 { /** The <I>problem-defining</I> parameter(s) of 'this'
       UnParameterised Model; see {@link #defnParams()}.  */
   protected final Value dp;

   /** Given common-knowledge problem-defining parameter(s), dp,
       create an UnParameterised Model.  */
   public UPModel(Value dp)
    { this.dp = dp; }//constructor

   /** Return the common-knowledge, problem-defining parameter(s) for the
       fully parameterised {@link M M-Model}s that 'this' UPModel can and
       will produce.  This implementation returns {@link #dp}.
       Also see {@link Model#statParams()}.  */
   public Value defnParams()                                  // defnParams()
    { return dp; }

   /** Given statistical parameters, sp, return a Model,
       <SPAN STYLE="white-space:nowrap">
         {@link #sp2Model sp2Model(0, 0, sp)},</SPAN>
       i.e., one that is not estimated, having zero part&nbsp;1 and
       part&nbsp;2 message lengths. (Also see {@link Function#apply}.)  */
   public Model apply(Value sp)                                    // apply(sp)
    { return sp2Model(/* m1 = */ 0.0, /* m2 = */ 0.0, sp); }

   /** Given two-part message lengths msg1 &amp; msg2, and statistical
       parameter sp, return a fully parameterised {@link M M-Model}.
       If 'this' UPModel produces a Model m, then
       <SPAN STYLE="white-space:nowrap">
          sp2Model(m.msg1(), m.msg2(), m.statParams())</SPAN>
       must be equivalent to&nbsp;m. Also see {@link #apply apply(sp)}.  */
   public abstract Model sp2Model(double msg1, double msg2, Value sp);     // *
      // *** NOTE we canNOT make this return new M(.,.,) as default
      //     because M is abstract.-(    javac doesn't like it.

   /** Return {@link #stats(Vector,int,int) stats(ds,0,ds.nElts())}
       (there is more on stats at the link).
       There is unlikely to be any need for a subclass (a UPModel)
       to override stats(ds).  */
   public Value stats(Vector ds)                                   // stats(ds)
    { return stats(ds, 0, ds.nElts()); }

   /** Return sufficient statistics 'ss' of elements [lo,&nbsp;hi),
       lo&nbsp;inclusive to hi&nbsp;exclusive, of data-set ds;
       also see {@link M#stats(Vector,int,int) M.stats(...)}.
       A simple option is to just return the elements &ndash;
       which is sometimes the best that we can do.
       Note, this.stats(...) must return statistics that are enough to
       {@link #estimator estimate} a fully parameterised {@link M M-model},
       whereas M.stats(...) must be enough to calculate
       {@link M#nlLH M.nlLH(.)}; often these are the same, but not always,
       for example, see {@link BestOf#stats BestOf.stats(...)}.
       In&nbsp;any case,
       <SPAN STYLE="white-space:nowrap">ss = this.stats(ds, lo, hi)</SPAN>
       must be compatible with {@link Estimator#ss2Model ss2Model(ss)}
       (as must be
       <SPAN STYLE="white-space:nowrap">
          ss' = {@link M#stats(Vector,int,int) M.stats(ds, lo, hi)}</SPAN>
       with {@link M#nlLH M.nlLH(ss')}).
       {@link Vector#wt(int) Weighted} data must also be considered.
       When defining a new UPModel stats(ds,lo,hi) and
       {@link #stats(boolean,Value,Value) stats(add,ss0,ss1)}
       are generally what must defined with regard to statistics.
       The statistics are not necessarily those used under
       maximum likelihood statistics, say, for example, nlAoM is
       included for continous Models.  */
   public abstract Value stats(Vector ds, int lo, int hi);                 // *

   /** Combine sufficient statisticses 'ss0' and 'ss1' additively
       (add=true), or remove ss1 from ss0 (add=false). Also see
       {@link #stats(Vector,int,int) stats(ds,lo,hi)}.  */
   public abstract Value stats(boolean add, Value ss0, Value ss1);         // *

   /** {@link #stats(boolean,Value,Value) Combine} statisticses ss0 and
       {@link #stats(Vector,int,int) stats(ds,lo,hi)}. This allows
       advantage to be taken of additive (incremental) statistics,
       for example, by {@link mml.Intervals}. There is unlikely to be a need
       for a subclass (a UPModel) to override stats(add,ss0,ds,lo,hi).  */
   public Value stats(boolean add, Value ss0, Vector ds, int lo, int hi)
    { return stats(add, ss0, stats(ds, lo, hi)); }

   /** Given parameter(s), 'ps' (possibly parameters of a prior), return an
       Estimator for 'this' UPModel. Estimator(ps) is optionally implemented
       and this default throws an Exception. It is desirable that
       estimator(ps) is implemented (overridden) but we may not
       have gotten around to it yet, or we may not be clever enough.  */
   public Estimator estimator(Value ps)                      // ? estimator(ps)
    { throw UOE( ".estimator("+ps+") has not yet been implemented." ); }

   /** Return a String representation of 'this' UnParameterised Model,
       including its problem-{@link #defnParams() defining} parameters.  */
   public String toString()                                       // toString()
    { return "(" + getClass().getName() + " " + defnParams() + ")"; }
   // .........................................................................

   /** Est, an Estimator that uses information from its enclosing
       {@link UPModel UnParameterised} Model. <I>Hopefully</I> it is
       instantiated in a particular UPModel.  */
   public abstract class Est extends Estimator                 // * UPModel.Est
    { /** Given parameter 'ps', possibly trivial, construct an
          Estimator of fully parameterised {@link M M-Models}.
          ps&nbsp;might, for example, control a prior on their
          statistical parameters.  */
      public Est(Value ps)
       { super(ps); }//constructor

      /** Return the enclosing UPModel's problem defining parameters,
          UPModel.this.{@link UPModel#defnParams() defnParams()}.  */
      public Value defnParams()
       { return UPModel.this.defnParams(); }

      /** Given part&nbsp;1 and part&nbsp;2 message lengths, m1 and m2, and
          statistical parameter(s), sp, return the fully parameterised Model
          UPModel.this.{@link UPModel#sp2Model sp2Model(m1,m2,sp)}.  */
      public final Model sp2Model(double m1, double m2, Value sp)
       { return UPModel.this.sp2Model(m1, m2, sp); }

   // public abstract Model ss2Model(Value ss), * from class Estimator

      /** Given a data-set ds and element bounds
          [lo,&nbsp;hi), return sufficient statistics by calling the
          enclosing {@link UPModel UPModel's}
          {@link UPModel#stats(Vector,int,int) stats(ds,lo,hi)},
          for use in {@link #ss2Model ss2Model(ss)}, say.  */
      public final Value stats(Vector ds, int lo, int hi)
       { return UPModel.this.stats(ds, lo, hi); }

      /** Return {@link UPModel UPModel's}
          {@link UPModel#stats(boolean,Value,Value) stats(add,ss0,ss1)}.  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return UPModel.this.stats(add, ss0, ss1); }

      /** Return a String representation of 'this' Estimator.  */
      public String toString()
       { return UPModel.this + ".estimator(" + params + ")"; }

    }//UPModel.Est (abstract) class
   // .........................................................................

   /** A fully parameterised Model, M, nested within an UnParameterised
       {@link UPModel UPModel}. A subclass of Model that extends
       UPModel.M gets {@link M#stats stats(ds...)} "for free" from
       UPModel.{@link UPModel#stats stats(ds...)} &ndash;
       <I>if appropriate</I> &ndash; similarly {@link M#toString()}.
       Also see {@link Model}.  */
   public abstract class M extends Model                         // * UPModel.M
    { /** Given two-part message lengths, msg1 and msg2, and
          statistical parameter(s), sp, construct an M-Model.  */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp); }//constructor

      /** Return sufficient statistics
          <SPAN STYLE="white-space:nowrap">ss = stats(ds, lo, hi)</SPAN>
          of elements [lo,&nbsp;hi), lo&nbsp;inclusive to hi&nbsp;exclusive,
          of data-set 'ds' using the
          {@link UPModel#stats(Vector,int,int) stats(ds,lo,hi)}
          of the enclosing {@link UPModel UPModel}.
          This is often appropriate, but not always, for example,
          see {@link BestOf.M#stats BestOf.M.stats(...)}.
          Note that ss=M.stats(ds...) must be enough to calculate
          {@link M#nlLH M.nlLH(ss)} but UPModel.stats(ds...) must be enough
          to {@link #estimator estimate} a fully parameterised M-Model.
          More on stats [{@link UPModel#stats(Vector,int,int) here</A>}].  */
      public Value stats(Vector ds, int lo, int hi)
       { return UPModel.this.stats(ds, lo, hi); /* which is abstract */ }

      /** By default, combine statisticses ss0 and ss1 using the
          {@link UPModel#stats(boolean,Value,Value) stats(add,ss0,ss1)}
          of the the enclosing {@link UPModel UPModel}.
          This is often appropriate, but not always.
          Note, ss=M.stats(ds...) must be enough to calculate
          {@link M#nlLH M.nlLH(ss)} but UPModel.stats(ds...) must be
          enough to estimate a fully parameterised model.
          More on stats [{@link UPModel#stats(Vector,int,int) here</A>}].  */
      public Value stats(boolean add, Value ss0, Value ss1)
       { return UPModel.this.stats(add, ss0, ss1); /* which is abstract */ }

   // public abstract double nlLH(Value ss);  // inherited from Model

      /** Calls {@link #asGiven(double,double) asGiven(0,msg2)}.  */
      public Model asGiven(double msg2) { return asGiven(0, msg2); }

      /** Enables setting the first- and second-part message lengths, msg1
          and msg2, after having estimated the statistical parameter(s) of
          a Model, say. Also see {@link #asGiven(double) asGiven(msg2)}.  */
      public Model asGiven(double msg1, double msg2)
       { return sp2Model(msg1, msg2, statParams()); }

      /** Return a String representation of 'this' fully parameterised Model.
          The result includes the enclosing
          {@link UPModel#toString() UnParameterised} Model, with its
          problem defining parameters, and 'this' M-Model's
          {@link #statParams() statistical} parameter(s).  */
      public String toString()
       { return "(" + UPModel.this + " " + statParams() + ")"; }

    }//UPModel.M (abstract) class, the Model(s)
   // -------------------------------------------------------------------------

   /** See {@link mml.Test}.  */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run UPModel --");
      System.out.println("   see mml.Test");
      System.out.println("-- done UPModel --");
    }//main

 }//UPModel (abstract) class

// (c) Lloyd Allison, 2012, Tue 15 Aug 2017. Licence to 'UPModel' is Gnu GPL.
