//       1         2         3         4         5         6         7
// 4567890123456789012345678901234567890123456789012345678901234567890123456789

package mml;

import la.util.Util;
import la.la.Lexical;
import la.la.Value;
import la.la.Function;
import static la.maths.Maths.*;
import la.maths.Vector;  // for Vector etc.
import la.maths.Matrix;  // used in main()

import static mml.MML.*;
// ----------------------------------------------------------------------------

/* Copyright (c) 2018 Lloyd Allison.

   Reference: 
     Lloyd Allison, "Coding Ockham's Razor", Springer, 2018.

   This file is free software: you can redistribute it and/or modify it
   under the terms of the GNU General Public License (GPL) as published
   by the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This file is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this file. If not, see http://www.gnu.org/licenses/
 */
// ----------------------------------------------------------------------------

/** The UnParameterised <NOBR>von Mises - Fisher</NOBR> Model of
    Directions in <B>R</B><SUP>D</SUP>. The fully parameterised Model
    is {@link vMF.M}.  A Direction is equivalent to a point on the
    surface of a unit-radius D-ball, so having D-1 degrees of freedom.
    For the special case of D=3, <B>R</B><SUP>3</SUP>, there is
    {@link MML#vMF3}.  (Also see
    [<A HREF="http://en.wikipedia.org/wiki/Von_Mises-Fisher_distribution"
        TARGET="EX">wikip</A>]['12].)  */
public class vMF extends Direction                                       // vMF
 { private final int D;
   private final double D2;

   /** D, the dimension of <B>R</B><SUP>D</SUP>. */
   public vMF(Value D)
    { super(D); this.D = ((Int) D).n(); D2 = this.D*0.5; }//constructor

   /** D(), the dimension of <B>R</B><SUP>D</SUP>. */
   public int D() { return D; }

   /** Given two-part message lengths, and statistical parameters
       <NOBR>sp = (&mu;, &kappa;),</NOBR> return a fully parameterised
       von Mises - Fisher {@link M Model}. */
   public M sp2Model(double msg1, double msg2, Value sp)
    { return new M(msg1, msg2, sp); }

   /** Statistics, ss = stats(ds) <NOBR> = (N, R, nlAoM)</NOBR>,
       of a data-set, ds.  All are {@link Vector#wt(int) weighted},
       N&nbsp;the number of data, Vector&nbsp;R the sum of the
       (normalised) Directions, and the data total (scaled) nlAoM.
       More on stats {@link UPModel#stats here}. */
   public Value stats(Vector ds, int lo, int hi)
    { if( ds.nCols() != D() )
         error( ".stats(ds), " + ds.nCols() + " wrong shape" );
      double [] xs = new double[D()];   // sum of Directions
      for( int c = 0; c < D(); c ++ ) xs[c] = 0;
      double nlAoM = 0;  // total nlAoM, weighted and adjusted
      for( int r = lo; r < hi; r ++ )        // for each row, r
       { final Vector v_r = (Vector) ds.elt(r);
         final double w_r = ds.wt(r),  nrm = v_r.norm();
         nlAoM += w_r*( v_r.nlAoM() + K()*Math.log(nrm) );  // total, adjusted
         for( int c = 0; c < D(); c ++ )            // for each column, c
            xs[c] += w_r * ds.elt(r, c).x() / nrm;  // NB. .../nrm !
       }//for r
      final double N = ds.wts(lo, hi);
      Vector R = Vector.doubles(xs);
      return triple(real(N), R, real(nlAoM));
    }//stats(ds)

   public Value stats(boolean add, Value ss0, Value ss1)
    { Value.Tuple ss0t = ss0.isTuple(3), ss1t = ss1.isTuple(3);
      return triple( real( add ? ss0t.fst().x() + ss1t.fst().x()
                               : ss0t.fst().x() - ss1t.fst().x() ),
                 ((Vector) ss0t.snd()).bOp(add?Lexical.plus:Lexical.minus,
                                           (Vector) ss1t.snd()),
                     real(add ? ss0t.thrd().x() + ss1t.thrd().x()
                              : ss0t.thrd().x() - ss1t.thrd().x()) );
    }
   // .........................................................................

   /** log(F(&kappa;)), log Fisher.  <NOBR> F = N<SUP>D</SUP>
          (&kappa; A<SUB>D</SUB>(&kappa;))<SUP>D-1</SUP>
          A'<SUB>D</SUB>(&kappa;); </NOBR> see {@link #AD AD}. Also see
          [<A HREF="http://www.allisons.org/ll/MML/Continuous/von-Mises/"
              TARGET="LL">vMF</A>]. */
   public double logF(double N, double kappa)
    { kappa = Math.max(kappa, 2 * epsilon);  // stop log(0)
      return D * Math.log(N)      // D-1 for mu, 1 for kappa
           + (D-1) * Math.log(kappa * AD.apply_x(kappa))
           + Math.log( AD.d_dx().apply_x(kappa) );
    }//logF(kappa)

   /** Function A<SUB>D</SUB>(&kappa;)
       <NOBR> = I(D/2,&kappa;) / I(D/2-1,&kappa;) </NOBR>
       <NOBR> = <B>E(v <B>.</B> &mu;)</B></NOBR>. &nbsp;
       <NOBR> Note, I<SUB>&alpha;</SUB>(.) is a </NOBR>
       modified Bessel function of the first kind.  */
   Function.Cts2Cts AD = new Function.Cts2Cts()
    { public double apply_x(double kappa)
       { return Bessel_I(D()/2, kappa) / Bessel_I(D()/2-1, kappa); }
         // if D > 2, kappa = 0, blows up with 0/0; also see logF
    };//AD(kappa)

   /** Length of the first part of the message. */
   public double msg1(double N, double kappa)
    { return Math.max( 0,  // paranoia
                logArea() - Math.log(prior_k(kappa))  // h(mu, kappa)
                + logF(N, kappa) * 0.5
                + D2 * Math.log(latticeConstant(D)) );
    }//msg1(N,kappa)

   /** Length of the second part of the message. */
   public double msg2(Vector mu, double kappa, Value ss)
    { return nlLH(mu, kappa, ss) + D2 /*rounding*/ ; }

   /** <I>Maybe change this(?)--</I> the prior's pdf on&nbsp;&kappa;
       is (1+&kappa;)<SUP>-2</SUP>, &kappa;&gt;0.
       (The prior on the mean Direction&nbsp;&mu; is Uniform.)  */
   public double prior_k(double kappa)
    { double k1 = kappa+1; return 1.0/(k1*k1); }
      // INT[0..inf] 1/(1+x)^2 = {-1/(1+x)}[0..inf] = 0 - -1 = 1


   /** Calls the {@link #estimatorMML MML&nbsp;estimator}. */
   public Est estimator(final Value t)
    { return estimatorMML(t); }

   /** The MML Estimator for &mu; and &kappa;.
       The {@link #estimatorMaxLH maximum likelihood} estimate of &mu; is
       accepted but its estimate of &kappa; is taken as a starting point
       (using an R guaranteed that <NOBR> ||R|| &lt; N). </NOBR>
       The prior on the mean Direction &mu; is Uniform.
       See {@link #logF log&nbsp;Fisher} and {@link #stats stats(ds)}.
       Note, the vMF has D() parameters -- &kappa;, and D()-1
       for Direction&nbsp;&mu;. */
   public Est estimatorMML(final Value t)                    // estimatorMML(t)
    { return new Est(t)
       { /** Given sufficient stats, ss, return an M. */
         public M ss2Model(final Value ss)
          { final Tuple NRA = ss.isTuple(3);
            final Vector  R = (Vector) NRA.snd();
            final double  N = NRA.fst().x(),
                      nlAoM = NRA.thrd().x();  // of R
            // If by chance the data is {v,v,...,v}, then
            // ||R||=N, and Rbar=1, causing div by 0 in max LH est.
            // So consider datas' AoM to cap ||R|| if necessary.
            final double nrm = R.norm(),
                         AoM = Math.min(N*0.5, Math.exp( - nlAoM / K())),
                       limit = Math.sqrt(N*N-AoM*AoM); // AoM < N/2 & limit < N
            final Tuple ss2 = nrm > limit  // want ||R'|| < N
               ? triple(NRA.fst(), R.scaled(limit/nrm), NRA.thrd())
               : NRA;
            // Note, maxLH est(ss) requires ||R|| < N & Rbar < 1.
            M m_ml = (M) estimatorMaxLH(triv).ss2Model(ss2); // maxLH est.
            // Accept m_ml.mu, but use m_ml.kappa as a starting point
            final Vector mu = m_ml.mu;              // maxLH and MML est.
            /** 'msg' is a Cts2Cts so that (i) can use N, mu and ss globally
                and (ii) msg can be optimised by Function.turning(.,.) */
            Function.Cts2Cts msg = new Function.Cts2Cts()
                { public double apply_x(double kappa)
                   { return msg1(N,kappa) + msg2(mu,kappa,ss); }
                };
            double kappa = Math.max(0,
                           Math.min(m_ml.kappa,    // beware AoM and Rbar=1
                           msg.turning(m_ml.kappa)));  // MML est.
            return new M(msg1(N,kappa), msg2(mu,kappa,ss),
                         pair(mu, real(kappa)));
          }//ss2Model(ss)

       };//new Est
    }//estimatorMML(t)


   /** The (rough and ready) maximum likelhood estimator for &mu;
       and &kappa;; see the {@link #estimator MML&nbsp;estimator}.
       <NOBR> <I>NB. It requires</I> ||R|| &lt; N,&nbsp; Rbar &lt; 1. </NOBR>
       The prior on the mean Direction &mu; is Uniform.
       Note, the vMF has D() parameters -- &kappa;, and D()-1
       for Direction&nbsp;&mu;.  Also see
       {@link #stats stats(ds)}, {@link #estimatorMML estimatorMML}. */
   public Est estimatorMaxLH(final Value t)
    { return new Est(t)
       { /** Given sufficient statistics, ss, estimate an M. */
         public M ss2Model(Value ss)
          { Tuple NRA = ss.isTuple(3);
            double N     = NRA.fst().x();  // NB wt(i) can be fraction
            Vector R     = (Vector) NRA.snd();
            double nrm   = R.norm();
            Vector mu    = R.scaled(1.0/nrm);
            double Rbar  = nrm / N;
            if( Rbar >= 1 ) error( "estimatorMaxLH(" + t + ").ss2Model("
                                 + ss +"), Rbar = " + Rbar );
            // note, rough Max L'hood.  Can do better for kappa...
            double kappa = 0,    /* vMF ~ uniform distn */
                   msg1  = log2; /* 1 bit, at least! */
            if( N >= 2) // enough data...
             { kappa = Rbar*(D - Rbar*Rbar) / (1.0 - Rbar*Rbar);  // Rbar < 1!
               msg1 = msg1(N, kappa);
             }
// System.out.print( " {N=" + N +
// ", AD=" + AD(kappa) +
// ", dAD/dk=" + AD.d_dx().apply_x(kappa) +
// ", logF=" + logF(N,kappa) + "} " );
            return new M(msg1, msg2(mu,kappa,ss), pair(mu, real(kappa)));
          }//ss2Model(ss)

    }; }//estimatorMaxLH(t)
   // .........................................................................

   /** Given &mu;, &kappa;, and statistics,
       <NOBR> ss = {@link vMF#stats stats(ds)} </NOBR> of a
       data-set, ds, return the negative log likelihood of ds.
       Used by {@link #estimator estimator} and {@link M#nlLH}. */
   public double nlLH(Vector mu, double kappa, Value ss)
    { Tuple NRA = ss.isTuple(3);
      double N  = NRA.fst().x(),      // (wt(i) can be fraction)
          nlAoM = NRA.thrd().x();     // neg log AoM of all data
      Vector R  = (Vector) NRA.snd(); // sum of (norm'd) data vecs
      return - kappa * mu .dot(R) - N * logCD(kappa) + nlAoM;
    }//nlLH(mu,kappa,ss)

   /** The {@link M Model's} <I>log</I> normalisation constant.
       <UL>
       <LI>C<SUB>D</SUB>(&kappa;) = &kappa;<SUP>(D/2-1)</SUP>
             / {(2&pi;)<SUP>D/2</SUP> I<SUB>D/2-1</SUB>(&kappa;)},
       <LI>C<SUB>3</SUB> = &kappa;
             / {2&pi;(e<SUP>&kappa;</SUP> - e<SUP>-&kappa;</SUP>)}.
       </UL>   I<SUB>&alpha;</SUB>(.) is a modified Bessel function
               of the first kind.  */
   public double logCD(double kappa)
    { final double D2_1 = D2 - 1;
      return kappa <= 0 ? - logArea()                    // Uniform!
           : D()==3 ? Math.log(kappa)-log2-logPI-kappa   // R^3 is special
                    - (kappa > 10 ? 0 : Math.log(1-Math.exp(-2*kappa)))
                    // note that    0 = log 1 = log(1-exp(-big))
           : D2_1 * Math.log(kappa) - D2 * (log2+logPI)  // any D()
             - Math.log(Bessel_I(D2_1, kappa));
    }//logCD(D, kappa)
   // .........................................................................

   /** The fully parameterised von&nbsp;Mises - Fisher Model of Directions
       in <B>R</B><SUP>D</SUP>. Its statistical parameters are mean
       {@link #mu() &mu;} and concentration {@link #kappa() &kappa;}.
       The UnParameterised Model is {@link vMF vMF}. */
   public class M extends Direction.M                               // vMF.M
    { /** The mean Direction, &mu;. */
      public final Vector    mu;         // mean is normalised ...
      /** The mean Direction {@link #mu mu} as double[]. */
      public final double [] mu_x;
      /** The concentration parameter, &kappa;. */
      public final double kappa, logCD;  // k conc., & normalising const

      /** The vMF's statistical parameters
          <NOBR>sp = ({@link #mu() &mu;}, {@link #kappa() &kappa;}</NOBR>).
          Note, to be safe, the constructor normalises the Vector &mu;. */
      public M(double msg1, double msg2, Value sp)
       { super(msg1, msg2, sp);
         Value.Tuple mK = sp.isTuple(2);
         mu   = ((Vector) mK.fst()) .normalised();
         mu_x = new double[ D() ];
         for( int i = 0; i < D(); i ++ ) mu_x[i] = mu.elt(i).x();
         kappa = mK.snd().x();
         if( kappa < 0 ) error( ", kappa=" + kappa + "<0" );
         logCD = logCD(kappa);
       }//constructor

      /** &mu;, the mean, a Direction in <B>R</B><SUP>D</SUP>. */
      public Vector mu()
       { return ((Vector) statParams().isTuple(2).fst()) .normalised(); }

      /** &kappa; &ge; 0, the concentration parameter.
          <NOBR>&kappa; = 0 gives</NOBR> the
          {@link Direction.Uniform.M Uniform} Model.
          For large &kappa;,
          <NOBR>&kappa; ~ 1/&sigma;<SUP>2</SUP></NOBR>, the variance. */
      public double kappa() { return kappa; }

      /** The negative log {@link #pdf pdf(v)} for datum Direction v,
          return <NOBR> - &kappa; &mu;<B>.</B>v
          - {@link #logCD log C<SUB>D</SUB>(&kappa;)}</NOBR>.
          Note&nbsp;that v need not be already normalised -- that
          is taken care of.  Also see {@link #nlPr nlPr(v)}. */
      public double nlPdf(Value v)
       { Vector vv = (Vector) v;
         if( v.nElts() != D() ) error( ".nlPdf(" + v + ")" );
         double muDotV = 0, nrm = 0;
         for( int c = 0; c < D(); c ++ )
          { double vc = v.elt(c).x();
            muDotV += mu_x[c]*vc;  nrm += vc*vc;
          }
         nrm = Math.sqrt(nrm);
         return - kappa * muDotV / nrm - logCD;  // ok for kappa >= 0
       }//nlPdf(v)                  note, returns -logCD if kappa = 0

      // public double nlPr(Value v), inherited from Direction.M

      /** Given statistics, ss = {@link vMF#stats stats(ds)} of a
          data-set, ds, return the negative log
          {@link vMF#nlLH likelihood} of ds. */
      public double nlLH(Value ss)
       { return vMF.this.nlLH(mu, kappa, ss); }

      /** See random_x(x) below. */
      private BetaUPM.M   beta = null;
      private Direction.M unif = null;

      /** Generate a random Direction, a random unit Vector in
          <B>R</B><SUP>D</SUP>.  Called by {@link #random_x()}.  See
          T.A.Wood, <I>Simulation of the von Mises Fisher distribution</I>,
          Communications in Statistics - Simulation and Computation,
          23(1), pp.157-164, 1994.  */
      public void random(double [] x)
       { double U, W, Z, xDotMu, norm, scale;
         final double k2 = 2.0*kappa;
         final int   d_1 = K() /* dimension-1 */;
         final double  b = (Math.sqrt(k2*k2 + d_1*d_1) - k2) / d_1;
         final double x0 = (1 - b) / (1 + b);
         final double c = kappa*x0 + d_1*Math.log(1 - x0*x0);
         if( beta == null )
          { beta = Beta.sp2Model(0, 0, pair(real(d_1/2.0), real(d_1/2.0)));
            // Note, when D()=3, d_1/2=1, & Beta(1,1) is Uniform_[0,1].
            unif = new Direction.Uniform(new Int(D())).sp2Model(0, 0, triv);
            // Use Uniform(new Int(d_1))... if x[d_1] is the mean
          }
         do{ // Note, Wood has x[d_1] is the mean; we have mu.
             do{ Z = beta.random_x();  U = RNG.nextDouble();
                 W = (1 - (1+b)*Z) / (1 - (1-b)*Z);
               }while( kappa*W + d_1*Math.log(1-x0*W) - c < Math.log(U) );
             scale = Math.sqrt(1 - W*W);  // for the random part c.f. W*mu
             unif.random( x );            // note, uniform in R^D()
             // for( int i = 0; i < d_1; i ++ ) x[i] *= scale;
             // x[d_1] = W;      -- mean direction, mu, not considered
             // return           -- if x[d_1] is the mean.
             xDotMu = 0;
             for( int i = 0; i < D(); i ++ ) xDotMu += x[i] * mu_x[i];
           }while( xDotMu > 0.999 ); // ...for the paranoid programmer
         norm = 0;
         for( int i = 0; i < D(); i ++ )  // flatten x in the mu direction
          { final double xi = x[i] - xDotMu * mu_x[i];
                 x[i] = xi;
                 norm += xi * xi;
          }
         norm = Math.sqrt(norm);
         // if( norm == 0 ) // for the paranoid, the probability of ...
          // { // ... taking this path is zero to machine accuracy
            // for( int i = 0; i < D(); i ++ ) x[i] = mu_x[i];  // x = mu
            // return;
          // }
         scale /= norm;                   // NB.
         for( int i = 0; i < D(); i ++ )  // Normalise, scale, and add W mu
            x[i] = x[i] * scale + W * mu_x[i];
       }//random(x)

    }//vMF.M class, the Model
// ----------------------------------------------------------------------------

   /** Run a few, very(!) simple tests. */
   public static void main(String[] argv)                               // main
    { System.out.println("-- run mml.vMF --");

      int N = 100;

      int D = 3;
      Vector x = Vector.doubles(new double[]{1, 0, 0}),
             y = Vector.doubles(new double[]{0, 1, 0}),
             z = Vector.doubles(new double[]{0, 0, 1});
      double k = 10;
      vMF  vmf = new vMF(new Int(D));               // UnParameterised Model
      vMF.M mx = vmf.new M(0, 0, pair(x, real(k))), // fully parameterised
            my = vmf.new M(0, 0, pair(y, real(k))); // ditto

      Mixture mixture = new Mixture(vmf);           // UnParameterised

// a known mixture of two vMF.M ...
      Model mix =  mixture.apply(                   // fully parameterised...
                     pair( Vector.doubles(new double[]{0.6, 0.4}),  // wts
                           Vector.values(new Value[]                // params
                            { mx.statParams(),     // class 1
                              my.statParams() }    // class 2
                         )      )       );

// generate a random data-set from the known mixture ...
      double [] [] ds_x = new double [N] [D];
      for( int i = 0; i < ds_x.length; i ++ )
       { Vector v = (Vector) mix.random();
         for( int j = 0; j < D; j ++ ) ds_x[i][j] = v.elt(j).x();
       }
      Matrix ds = Matrix.doubles(ds_x, 0.001);  // all have this AoM

// fit a single vMF.M to the given data-set ...
      mml.Test.exercise(vmf.estimator(triv), ds);
      System.out.println( "--" );

// fit a mixture to the given data-set ...
      mml.Test.exercise(mixture.estimator(triv), ds);

      System.out.println("-- done mml.vMF --");
    }//main

 }//vMF class, the UPModel

// (c) Lloyd Allison 2012, Thu 29 Jan 2016. Licence to 'vMF' is Gnu GPL.
